-- Various damage related stuff;
-- spawning projectiles, hurtboxes and actually hurting entities, too

freeslot("MT_PSO_HURTBOX", "sfx_cburst")
mobjinfo[MT_PSO_HURTBOX] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	flags = MF_NOGRAVITY|MF_NOCLIPHEIGHT,
	radius = 32<<FRACBITS,
	height = 64<<FRACBITS,
}

-- spawns a hurtbox
PSO.lib.spawnHurtBox = function(x, y, z, w, t, source, potency, dtype, duration, rtype, condition, rate)
	
	assert(source, "spawnHurtBox: No source to get stats from")
	
	if rtype == nil
		rtype = RT_STAGGER
	end
	
	local h = P_SpawnMobj(x, y, z, MT_PSO_HURTBOX)
	h.radius = w/2
	h.height = t
	--h.scale = source.scale
	
	if PSO.debug
		h.sprite = SPR_THOK
		h.frame = A
		h.color = SKINCOLOR_RED
		
		/*local coords = {
			{h.x - h.radius, h.y - h.radius},
			{h.x - h.radius, h.y + h.radius},
			{h.x + h.radius, h.y - h.radius},
			{h.x + h.radius, h.y + h.radius},
		}
		
		for i = 1, #coords
			local cx = coords[i][1]
			local cy = coords[i][2]
			local t1 = P_SpawnMobj(cx, cy, h.z, MT_THOK)
			t1.frame = A
			t1.color = SKINCOLOR_RED
			t1.scale = h.scale/2
			t1.tics = duration
			t1.fuse = duration
		end*/
	end	
	
	h.target = source
	h.tics = duration or TICRATE/2
	h.fuse = duration or TICRATE/2
	h.starttime = leveltime
	
	local info = source.player and source.player.pso or source.pso
	
	local mult = 100
	if source.player
		if source.player.pso.weapon
		and source.player.pso.weapon.pa_damage
			mult = source.player.pso.weapon.pa_damage
		end
	end
	
	h.pso = {
		
		satk = info.satk or 1,
		ratk = info.ratk or 1,
		tatk = info.tatk or 1,
		potency = (potency or 1)*mult/100,
		damagetype = dtype or DMT_STRIKE,
		reactiontype = rtype,
		status = condition or 0,
		statuschance = rate or 0,
	}
	
	-- share table if possible
	if info.whitelist
		h.pso.whitelist = info.whitelist
	end
	
	return h
end

local function maintain_whitelist(mo1, mo2)
	-- add this entity to our whitelist, we can't hurt it anymore after this:
	mo1.pso.whitelist = $ or {}
		
	for i = 1, #mo1.pso.whitelist
		if mo1.pso.whitelist[i] == mo2
			return false	-- we already got hurt by this!
		end
	end
	
	-- check if it was already hit by someone else and add it to our own table to make subsequent checks faster
	if mo1.pso.links
		for i = 1, #mo1.pso.links
			local hb = mo1.pso.links[i]
			hb.pso.whitelist = $ or {}
			for j = 1, #hb.pso.whitelist
				if hb.pso.whitelist[j] == mo2
					table.insert(mo1.pso.whitelist, mo2)
					return false
				end
			end
		end	
	end
		
	mo1.pso.whitelist[#mo1.pso.whitelist+1] = mo2	
		
	return true
end

local function ra_tacticalroll_trigger(p)
	if PSO.lib.hasSkill(p, "ra_tacticalroll")
					
		S_StartSoundAtVolume(p.mo, sfx_cburst, 127)
					
		for i = 1, 8
			local s = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_SUPERSPARK)
			s.destscale = 1
			s.momx = P_RandomRange(-32, 32)*FRACUNIT
			s.momy = P_RandomRange(-32, 32)*FRACUNIT
			s.momz = P_RandomRange(-32, 32)*FRACUNIT
			s.colorized = true
			s.color = SKINCOLOR_BLUE
		end
					
		p.pso.pp = min(p.pso.maxpp, p.pso.pp+35)
	end
end

-- collision handling for the hitbox
local function hurtbox_collide(mo1, mo2)
	
	if mo1 and mo1.valid and mo1.type == MT_PSO_HURTBOX
	and mo2 and (mo2.flags & (MF_SHOOTABLE) or mo2.player)
	and mo2.type ~= MT_DOOR_HITBOX	-- this sometimes happens for some reason and the results are catastrophic
	
		local owner = mo1.target
		if not owner or not owner.valid	-- nope.
			return false
		end

		if mo1.pso.locks == 0	-- specifically 0, don't check for nil/0
			return false
		end
		
		if mo1.continuous
		and mo2.lasthit
		and leveltime - mo2.lasthit < mo1.continuous
			return false
		end
		
		--print("OK1")
		-- don't care
		if owner == mo2
		or owner.player and mo2.player
		or (not owner.player and not mo2.player)
			return false
		end
		
		if mo2.player
			local m2info = mo2.player.pso
			if m2info.dodgeroll and maintain_whitelist(mo1, mo2)
				
				ra_tacticalroll_trigger(mo2.player)
				return false
			end
		end
		
		--print("OK2")
		if mo1.z + mo1.height < mo2.z return false end	-- (too low) why do we have to do this manually....
		if mo1.z > mo2.z + mo2.height return false end	-- (too high)

		-- locks are now handled in the mobjthinker
		/*if mo1.pso.locks
			mo1.pso.locks = $-1
		elseif mo1.pso.locks == 0	-- hit the maximum number of entities
			return false
		end*/
		
		-- whitelisted item (already hit)
		if not maintain_whitelist(mo1, mo2) return end
		
		
		if mo1.continuous
			mo2.lasthit = leveltime
		end
		
		-- ok we can get hurt
		mo1.pso.damagelist = $ or {}
		
		mo1.pso.damagelist[#mo1.pso.damagelist+1] = mo2
		--P_DamageMobj(mo2, mo1, owner)
		
		return false	-- we still have never collided, technically.
	end
end

addHook("MobjThinker", function(mo)
	
	local owner = mo.target
	local ownertarget
	
	if not owner or not owner.valid
		P_RemoveMobj(mo)
		return
	end
	
	if owner.player
		ownertarget = owner.player.pso.target
	end
	
	if mo.pso.damagelist
		
		if mo.pso.locks ~= nil
			-- if we have locks, sort the damagelist by relevancy
			if #mo.pso.damagelist > 2
				table.sort(mo.pso.damagelist,
					function(a, b)
						
						-- these two cases shouldn't happen but they do in rare instances i haven't figured out
						if a == nil or not a.valid		-- ????
							return false
						elseif b == nil or not b.valid	-- ???? (2)
							return true
						end
						
						if a == ownertarget
							return true
						end
						
						return R_PointToDist2(mo.x, mo.y, a.x, a.y) < R_PointToDist2(mo.x, mo.y, b.x, b.y)
					end
				)
			end
		end
		
		for i = 1, #mo.pso.damagelist
			if not mo.pso.damagelist[i]
			or not mo.pso.damagelist[i].valid
				continue
			end
			
			P_DamageMobj(mo.pso.damagelist[i], mo, owner)
			
			-- if we have locks, decrement the counter and stop iterating if we 
			if mo.pso.locks
				mo.pso.locks = $-1
				if not mo.pso.locks
					break
				end
			end
		end
		
		mo.pso.damagelist = nil
	end
end, MT_PSO_HURTBOX)

addHook("MobjCollide", hurtbox_collide, MT_PSO_HURTBOX)
addHook("MobjMoveCollide", hurtbox_collide, MT_PSO_HURTBOX)

-- spawns a missile object from mobj with the specified offsets from it
-- speed, angle etc can be specified as well as potency and damage type (strike, ranged or technique)
PSO.lib.spawnMissile = function(source, typ, angle, speed, potency, dtype, ox, oy, oz, rtype, condition, rate)
	
	assert(source, "spawnMissile: No source to get stats and position from")
	
	local x = source.x + FixedMul(ox or 0, source.scale) 
	local y = source.y + FixedMul(oy or 0, source.scale) 
	local z = source.z + FixedMul(oz or 0, source.scale) 
	local msl = P_SpawnMobj(x, y, z, typ)
	msl.scale = source.scale
	
	if rtype == nil
		rtype = RT_STAGGER
	end
	
	local info = source.player and source.player.pso or source.pso
	
	assert(info, "spawnMissile: No info on source")

	local mult = 100
	if source.player
		if source.player.pso.weapon
		and source.player.pso.weapon.pa_damage
			mult = source.player.pso.weapon.pa_damage
		end
	end
	
	msl.pso = {
		satk = info.satk or 1,
		ratk = info.ratk or 1,
		tatk = info.tatk or 1,
		potency = (potency or 1)*mult/100,
		damagetype = dtype or DMT_STRIKE,
		reactiontype = rtype,
		status = condition,
		statuschance = rate,
	}
	
	if angle ~= nil
		msl.angle = angle 
	else
		msl.angle = source.angle
	end
	
	if speed
		P_InstaThrust(msl, msl.angle, FixedMul(msl.scale, speed))
	end
	
	-- who it belongs to
	msl.target = source
	msl.friendly = source.player and true or false	-- in case things go apeshit.
	
	return msl
end

-- aims the missile's z momentum towards the target
-- call straight after spawning the missle
PSO.lib.aimMissile = function(missile, target)
	assert(target and target.valid, "PSO.lib.aimMissile: no valid target")
	
	-- only aim if we wouldn't hit to begin with, or if the target is on a slope
	if missile.z + missile.height < target.z
	or missile.z > target.z + target.height
	or target.standingslope
		
		local dist = max(1, R_PointToDist2(missile.x, missile.y, target.x, target.y))
		local speed = FixedHypot(missile.momx, missile.momy)
		
		local tics = FixedDiv(dist, speed or 1) /FRACUNIT
	

		missile.momz = ((target.z + target.height/3) - missile.z) / (tics or 1)
	end
end

-- lob the missile towards a target
PSO.lib.lobMissile = function(msl, target, bspeed)
	-- kind of do things like lobshot
	
	if bspeed
		P_InstaThrust(msl, R_PointToAngle2(msl.x, msl.y, target.x, target.y), bspeed)
	end
	
	local spd = R_PointToDist2(0, 0, msl.momx, msl.momy)
	local dist = R_PointToDist2(msl.x, msl.y, target.x, target.y)
					
	// horizontal = dist / airtime;
	local airtime = dist / spd
					
	local vertical = -(P_GetMobjGravity(msl)*airtime)/2
					
	msl.momz = $ + vertical
end

local chain_start = {
	200,
	160,
	145,
	135,
	130,
	125,
	120,
	115,
}

-- calculates damage
PSO.lib.damageCalc = function(mo, inflictor)
	
	-- the formula is different between players and enemies
	local dmt = inflictor.pso.damagetype or DMT_STRIKE	-- all damage is strike unless stated otherwise...
	local source = inflictor.target
	local minfo = mo.pso or mo.player.pso
	local sinfo
	if source
		sinfo = source.pso or source.player.pso
	end
	
	local extra_mult = 100
	
	local stat = {
		[DMT_STRIKE] = "satk",
		[DMT_RANGED] = "ratk",
		[DMT_TECHNIQUE] = "tatk",
	}
	
	local atk = inflictor.pso[stat[dmt]] or 1
	local def = minfo.def or 1
	local potency = inflictor.pso.potency or 1

	-- apply shifta?
	if inflictor.pso.shifta
	or sinfo and sinfo.shifta
		atk = $ * (100+PSO.info.SHIFTA)/100
	end
	
	-- apply deband?
	if minfo and minfo.deband
		def = $ * (100+PSO.info.DEBAND)/100
	end
	
	-- add player weapon attack if applicable
	if source and source.valid and source.player
		local weapon = PSO.lib.getPlayerActiveWeapon(source.player)
				
		if weapon
			local wdmt = PSO.info.weaponDamageType[weapon.weapontype]
			
			if wdmt == dmt
				atk = $ + weapon.atk
				--print("Add weapon atk")
			end
			
			-- increase potency on weaknesses!
			
			-- for attack element:
			if inflictor.pso.element == minfo.weakness
				potency = $ + potency*15/100
			end
			
			-- for weapon element
			if weapon.element == minfo.weakness
				local mult = 15 + (15*weapon.elementv/100)
				potency = $ + potency*mult/100
			end
		end
		
		-- hateful aggression (HU)
		if PSO.lib.hasSkill(source.player, "hu_hatefulaggression")
			local en = mo
			if en.pso.part
				en = mo.tracer
			end
			en.hateful = true	-- play effect when damaged
			
			if en.target == source
				extra_mult = $+5
			end
		end
		
	-- add player armor if applicable
	elseif mo and mo.player
		
		local armor = PSO.lib.getPlayerArmor(mo.player)
		
		if armor
			def = $ + armor.def
		end
	end
	local dmg = max(atk/10 +1, (atk - (def)) * potency/100 / 3) *extra_mult/100
	
	-- @TODO: damage multipliers (shifta/deband...)
	-- chain:
	if minfo.chain and minfo.chain > 1 and inflictor.pso.combobreaker
		
		local playercount = 0
		for player in players.iterate
			playercount = $+1
		end
		
		local start = chain_start[playercount] or chain_start[#chain_start]
		
		dmg = $ * (start + minfo.chain*2) / 100
	end
	
	-- lol variance
	dmg = $ * P_RandomRange(90, 100) / 100
	
	return max(1, dmg)
end

freeslot("sfx_lowhp1", "sfx_lowhp2", "sfx_lowhp3")
PSO.lib.damageObject = function(mo, inflictor, damage, skipchain)
	
	local info = mo.player and mo.player.pso or mo.pso
	
	-- do not damage players after a quest has been cleared
	if mo.player
	and (server.questdata and server.questdata.complete)
		return
	end
	
	-- STO-O-O-OOOP HE'S ALREADY DEAD :sob:
	if not info.hp
		return 
	end
	
	local prevhp = info.hp
	
	if mo.hateful
		local avgx = (mo.x/FRACUNIT + inflictor.x/FRACUNIT)/2 *FRACUNIT
		local avgy = (mo.y/FRACUNIT + inflictor.y/FRACUNIT)/2 *FRACUNIT
		local avgz = (mo.z/FRACUNIT + inflictor.z/FRACUNIT)/2 *FRACUNIT	
		for i = 1, 6
			local p = P_SpawnMobj(avgx, avgy, avgz, MT_THOK)
			p.color = SKINCOLOR_RED
			p.scale = $/2
			p.destscale = 1
			p.fuse = 12
			p.tics = 12
			
			p.momx = P_RandomRange(-16, 16)*FRACUNIT
			p.momy = P_RandomRange(-16, 16)*FRACUNIT
			p.momz = P_RandomRange(-16, 16)*FRACUNIT
		end
		mo.hateful = nil
	end
	
	info.hp = max(0, $ - damage)
	
	if mo.player
		-- visuals
		if mo.player == displayplayer
			P_StartQuake(FRACUNIT*20, 8)
			
			local sound_thresholds = {15, 30, 50}
			local sound_refs = {sfx_lowhp3, sfx_lowhp2, sfx_lowhp1}
			
			for i = 1, #sound_thresholds
				if prevhp*100/info.maxhp > sound_thresholds[i]
				and info.hp*100/info.maxhp <= sound_thresholds[i]
					S_StartSound(nil, sound_refs[i], mo.player)
					break
				end
			end
		end
		
		-- auto mate
		if info.hp >= 0
		and prevhp >= info.maxhp/2
		and info.hp < info.maxhp/2
		and PSO.lib.hasSkill(mo.player, "all_automate")
			local mates = {"monomate", "dimate", "trimate"}
			for i = 1, #mates
				local mate = mates[i]
				local index = PSO.lib.findPlayerItem(mo.player, mate)
				if index
					local it = mates[i]
					local itinfo = PSO.lib.itemInfo(it)
					
					-- if we made it here then we can use it!
					PSO.lib.playerLoseItem(mo.player, index, 1)	-- use the item up
								
					if itinfo.use
						itinfo.use(mo.player)
					end
					break
				end
			end
		end
		
		-- iron will
		if not info.hp	-- dead
		and not info.ironwill
		and PSO.lib.hasSkill(mo.player, "hu_ironwill")
			info.hp = 1
			info.ironwill = true	-- used
			
			local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			s.sprite = SPR_FORC
			s.frame = A|FF_ADD
			s.destscale = FRACUNIT*8
			s.scalespeed = FRACUNIT/7
			s.fuse = 16
			s.tics = s.fuse
			
			P_FlashPal(mo.player, 1, 3)
		end
		
		-- pp absorb
		if info.hp
		and PSO.lib.hasSkill(mo.player, "fo_ppabsorb")
			local ppa = damage/3
			info.pp = min(info.maxpp, info.pp+ppa)
		end
		
	end
	
	if info.part
	and info.partlinkdamage
	and mo.tracer and mo.tracer.valid
	and not skipchain	-- no more recursion past this shit.
		--print("recursive or garbo")
		PSO.lib.damageObject(mo.tracer, inflictor, damage, true)
	end
	
	-- players: count as damage taken
	if info.quest and damage and damage > 0	-- only count positive damage, just in case.
		info.quest.damagetaken = $ + damage
	end
	
	if not skipchain
	
		local source
		if inflictor and inflictor.valid
			source = inflictor.target or inflictor
		end

		PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z + mo.height/2, damage, 0, mo, 0, source)
	end
	
	-- if inflictor isn't a "combo breaker", build the chain combo!
	if inflictor and inflictor.valid	-- environmental damage doesn't count
	and not inflictor.pso.combobreaker
	and not mo.player	-- only works on enemies.
	and not info.chainbreak
	and not skipchain
		
		if not info.chain
			server.chaintrackers[#server.chaintrackers+1] = mo
		end
		
		info.chain = $ and $+1 or 1
		info.maxchaintime = max(TICRATE/2, TICRATE*5 - (info.chain-1)*3/2)
		info.chaintime = info.maxchaintime
		
		if info.chain == 1
			info.chaintriggers = {}
		end
		
	elseif info.chain and info.chain > 1 and inflictor and inflictor.valid and inflictor.pso.combobreaker
	and not info.chainbreak
	and not skipchain
		
		info.maxchainbreak = TICRATE + info.chain*3
		info.chainbreak = info.maxchainbreak
	end
	
	
	if inflictor and inflictor.valid
		
		local cbreak = info.chainbreak and inflictor.pso and inflictor.pso.combobreaker
		S_StartSoundAtVolume(mo, cbreak and sfx_cdfm14 or sfx_dmpain, cbreak and 160 or 127)
		
		if cbreak
			S_StartSoundAtVolume(mo, sfx_cburst, 127)
			
			-- get average coord
			local avgx = (mo.x/FRACUNIT + inflictor.x/FRACUNIT)/2 *FRACUNIT
			local avgy = (mo.y/FRACUNIT + inflictor.y/FRACUNIT)/2 *FRACUNIT
			local avgz = (mo.z/FRACUNIT + inflictor.z/FRACUNIT)/2 *FRACUNIT
			
			for i = 1, 8
				local s = P_SpawnMobj(avgx, avgy, avgz, MT_SUPERSPARK)
				s.destscale = 1
				s.momx = P_RandomRange(-32, 32)*FRACUNIT
				s.momy = P_RandomRange(-32, 32)*FRACUNIT
				s.momz = P_RandomRange(-32, 32)*FRACUNIT
				s.colorized = true
				s.color = SKINCOLOR_GOLD
			end
			
			-- chain trigger
			local source = inflictor.target or inflictor
			local p = source.player
			
			if p
				if PSO.lib.hasSkill(p, "ra_chaintrigger")
					
					if not info.chaintriggers[#p]
						info.chaintriggers[#p] = true
						
						-- now give PP back based on chain count and playercount
						-- the formula is: 4*chaincount - 5% per player (max 50%)
						local playercount = 0
						for player in players.iterate
							playercount = $+1
						end						
						
						local mult = max(50, 100 - (playercount-1)*5)
						local pp = 4*info.chain*mult/100
						
						p.pso.pp = min(p.pso.maxpp, $+pp)
					end
				end
			end
			
		end
		
		if inflictor.pso.ppregen
		and inflictor.target and inflictor.target.player
			inflictor.target.player.pso.pp = min(inflictor.target.player.pso.maxpp, $ + inflictor.pso.ppregen)
		end
		
		if not info.hp
			PSO.lib.killObject(mo, inflictor)
		else
			-- check if stagger etc must be inflicted.
			local drt = inflictor.pso.reactiontype or RT_NONE
			local weight
			
			-- weight takes away from drt
			-- so blowdown can become knockback or stagger depending on weight.
			if mo.player
				-- for players, "weight" depends on damage taken vs max hp.
				/*local percentage = damage*100/info.maxhp
				
				if percentage <= 1
					weight = 3	-- less than 1%: take no effect except blowdown as stagger
				
				elseif percentage <= 5	-- less than 5%: blowdown -> knockback, knockback -> stagger
					weight = 2
				
				else
					weight = 1
				end*/
				weight = 1	-- maybe later?
				
				if PSO.lib.hasSkill(mo.player, "hu_hunterphysique")
					weight = 2
				end
				
			elseif PSO.info.enemies[mo.type]
				-- enemies have weight directly defined on them.
				weight = PSO.info.enemies[mo.type].weight or 1
			end
			
			drt = max(RT_NONE, $ - (weight-1))
			
			if PSO.info.enemies[mo.type]
				if PSO.info.enemies[mo.type].damage
					local run, res = pcall(PSO.info.enemies[mo.type].damage, mo, inflictor, damage, drt)
					
					if not run
						print("\x82".."WARNING: ".."\x80".."damageObject: Enemy "..mo.type.." damage function: "..res)
					end
				end
			end
			
			-- statuses!
			if inflictor.pso.status
			and P_RandomRange(1, 100) <= inflictor.pso.statuschance
				PSO.lib.inflictStatus(mo, inflictor.pso.status)
			end
			
			if not info.superarmor
				
				-- if a player got hit by any stagger, stop action
				if drt > RT_NONE and mo.player
					local p = mo.player
					PSO.lib.resetPlayerAction(p)
				end
				
				-- now apply the effects
				if drt == RT_STAGGER
					--print("A")
					P_InstaThrust(mo, 0, 0)
					info.knockback = 0
					info.blowdown = 0
					info.stagger = 1
				
				elseif drt == RT_KNOCKBACK
					local strength = inflictor.pso.reactionstrength or 40
					info.stagger = 0
					info.blowdown = 0
					info.knockback = 1
					
					local source = inflictor.target
					if not source or not source.valid
						source = inflictor
					end
					
					local ang = inflictor.angle
					if inflictor.pso.reactionangle ~= nil
						ang = inflictor.pso.reactionangle
					elseif inflictor.pso.reactionradialangle
						ang = R_PointToAngle2(inflictor.x, inflictor.y, mo.x, mo.y)				
						/*if source.player
							ang = source.player.drawangle
						end*/
					end
					--mo.angle = ang + ANG1*180
					P_InstaThrust(mo, 0, 0)
					P_InstaThrust(mo, ang, strength*source.scale)
					
				elseif drt == RT_BLOWDOWN
					local strength = inflictor.pso.reactionstrength or 8
					info.stagger = 0
					info.knockback = 0
					info.blowdown = 1

					local source = inflictor.target
					if not source or not source.valid
						source = inflictor
					end

					local ang = inflictor.angle
					if inflictor.pso.reactionangle ~= nil
						ang = inflictor.pso.reactionangle
					elseif inflictor.pso.reactionradialangle
						ang = R_PointToAngle2(inflictor.x, inflictor.y, mo.x, mo.y)
					end		
					
					if not mo.player
						mo.angle = ang + ANG1*180
					else
						mo.player.drawangle = ang + ANG1*180
					end
						
					P_InstaThrust(mo, ang, strength*mo.scale)
					mo.momz = strength*mo.scale
				end
			end
		end
	else
		S_StartSoundAtVolume(mo, sfx_dmpain, 127)
		if not info.hp
			PSO.lib.killObject(mo, nil)
		end
	end
end

COM_AddCommand("psod_hurtme", function(p, arg)
	
	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	PSO.lib.damageObject(p.mo, nil, tonumber(arg))
end)

PSO.lib.killObject = function(mo, inflictor)
	
	if mo.flags & MF_ENEMY
		local num
		if mo.pso.exp and not mo.pso.part	-- parts don't give exp.
			num = PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z, mo.pso.exp, 1)
		end
		
		-- parts don't drop items.
		if not mo.pso.part
			PSO.lib.enemyDropItems(mo)
		end
		
		-- if the object had parts, kill them too.
		if mo.parts
			for i = 1, #mo.parts
				if mo.parts[i] and mo.parts[i].valid
					--print("kill part "..i)
					mo.parts[i].pso.hp = 0
					PSO.lib.killObject(mo.parts[i])
				end
			end
		end
		
		for p in players.iterate do
			local pmo = p.mo
			
			if not pmo or not pmo.valid continue end
			
			if R_PointToDist2(pmo.x, pmo.y, mo.x, mo.y) < PSO.info.expdist*pmo.scale
				PSO.lib.givePlayerExp(p, mo.pso.exp)
				-- count as a kill!
				if p.pso.quest
					p.pso.quest.kills[mo.type] = $ and $+1 or 1
				end			
			elseif p == consoleplayer
			and num
				num.extravalue2 = 1	-- don't display locally
			end
		end
		
		P_KillMobj(mo)
	
	elseif mo.player
		
		mo.player.pso.deadtimer = 1
		PSO.lib.removeStatus(mo)
		PSO.lib.closeMenu(mo.player)	-- close any actve menu
	end
end

-- heals an object, if possible.
PSO.lib.healObject = function(mo, inflictor, heal)
	local info = mo.player and mo.player.pso or mo.pso
	info.hp = min(info.maxhp, $ + heal)
	
	local source
	if inflictor and inflictor.valid
		source = inflictor.target or inflictor
	end
	
	PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z + mo.height/2, heal, 0, mo, 1, source)
end

PSO.lib.revivePlayer = function(p, pinflictor, amount)
	PSO.lib.healObject(p.mo, pinflictor, amount or p.pso.maxhp/2)
	p.pso.deadtimer = 0
	p.mo.state = S_PLAY_STND
	p.powers[pw_flashing] = TICRATE*3 -1
							
	PSO.lib.closeMenu(p, true)
	
	S_StartSound(p.mo, sfx_heal)
end

-- remove these trackers if we change maps
addHook("MapChange", do
	server.dmgnums = nil
	server.chaintrackers = nil
end)

-- handles damage numbers and chains
local function handleDamageNumbers()
	
	server.dmgnums = $ or {}
	server.chaintrackers = $ or {}
	
	local i = #server.dmgnums
	while i
		if not server.dmgnums[i]
		or not server.dmgnums[i].valid
			table.remove(server.dmgnums, i)
		else
			server.dmgnums[i].momz = $*90/100
		end
		i = $-1
	end

	i = #server.chaintrackers
	while i
		if not server.chaintrackers[i]
		or not server.chaintrackers[i].valid
		or not server.chaintrackers[i].pso.chain
			table.remove(server.chaintrackers, i)
		else
			local m = server.chaintrackers[i]
			if m and m.pso.chain
				
				if m.pso.chainbreak
					m.pso.chainbreak = $-1
				else
					m.pso.chaintime = $-1
				end	
					
				if not m.pso.chaintime
				or m.pso.chainbreak == 0
					m.pso.chain = nil
					m.pso.chaintime = nil	
					m.pso.chainbreak = nil
					table.remove(server.chaintrackers, i)
				end
			end
		end
		i = $-1
	end
	
end
PSO.lib.registerFrameFunc(handleDamageNumbers)


PSO.lib.spawnDamageNumber = function(x, y, z, num, xp, target, heal, source)
	
	local d = P_SpawnMobj(x, y, z, MT_THOK)
	d.state = S_INVISIBLE
	d.tics = TICRATE*2
	d.momz = FRACUNIT*8
	d.target = target
	d.source = source
	
	-- iterate thru damage numbers and move them away if they're going to overlap...
	for i = 1, #server.dmgnums
		local dn = server.dmgnums[i]
		
		if dn and dn.valid
			
			local dist = R_PointToDist2(d.x, d.y, dn.x, dn.y)
			if dist < FRACUNIT*32
			and abs(d.z - dn.z) < FRACUNIT*64
				local a = R_PointToAngle(d.x, d.y) + ANG1*90 + (i & 1)*ANG1*180	-- use local camera angle...
				local nx = dn.x + 32*cos(a)
				local ny = dn.y + 32*sin(a)
				P_SetOrigin(d, nx, ny, d.z)
				break
			end
		end
		
	end
	
	-- lol
	d.angle = num
	d.extravalue1 = xp or 0		-- hack
	d.extravalue2 = heal or 0	-- hack (1)
	
	server.dmgnums[#server.dmgnums+1] = d
	
	return d
end

addHook("ShouldDamage", function(mo, inflictor, source, damagetype)
	
	local minfo = mo.pso
	if not minfo and mo.player
		minfo = mo.player.pso
	end
	
	if (minfo) and inflictor and inflictor.valid and inflictor.pso
		
		if inflictor.target == mo
			return false
		end
		
		-- non-players and players don't damage among each other
		if (inflictor.target and inflictor.target.player and mo.player)
		or (inflictor.target and not inflictor.target.player and not mo.player)
			return false
		end
		
		if inflictor.flags & MF_MISSILE
			if not maintain_whitelist(inflictor, mo) return false end
		end
		
		-- players don't take damage from direct enemy contact.
		if mo.player
				
			if inflictor.flags & MF_ENEMY
				-- push me out
				--local ang = R_PointToAngle2(inflictor.x, inflictor.y, mo.x, mo.y)
				--P_InstaThrust(mo, ang, 16*FRACUNIT)
				return false
			end
			
			if inflictor.type ~= MT_PSO_HURTBOX
			and mo.player.pso.dodgeroll
				ra_tacticalroll_trigger(mo.player)
				return false
			end
			
			if mo.player.powers[pw_flashing]	-- hurt
			or mo.player.pso.jumppad	-- jumping
			or not mo.player.pso.type_initialized	-- new joiner?'
			or (server.questdata and server.questdata.complete)	-- quest complete
			--or minfo.dodgeroll	-- dodging
				return false
			end
		end
		
		-- entities with nil hp use their normal damage function/hooks whatever
		if minfo.hp == nil
			--print("Do damage")
			return true
		end
		
		-- used mostly by enemies
		if minfo.invincible
			return false
		end
		
		local dmg = PSO.lib.damageCalc(mo, inflictor)
		--PSO.lib.dprint("Should deal "..dmg.." damage")
		PSO.lib.damageObject(mo, inflictor, dmg)
		return false
	else
		-- don't damage
		return false
	end
end)

-- players: ignore missiles from friendly sources
local function player_missile_collide(mo1, mo2)
	if mo1 and mo1.valid and mo1.player
	and mo2 and mo2.valid and mo2.flags & MF_MISSILE
		
		if mo2.target and mo2.target.valid and mo2.target.player
			return false
		end
	end
end

addHook("MobjCollide", player_missile_collide, MT_PLAYER)
addHook("MobjMoveCollide", player_missile_collide, MT_PLAYER)