-- Status condition behaviours for both players and enemies

PSO.info.statusConditions = {
	
	[COND_BURN] = {
		
		duration = TICRATE*10,
		
		runfunc = function(mo, timer)
			
			local info
			if mo.player
				info = mo.player.pso
			else
				info = mo.pso
			end
			
			-- visuals
			local steam = P_SpawnMobj(mo.x+P_RandomRange(-45, 45)*mo.scale, mo.y+P_RandomRange(-45, 45)*mo.scale, mo.z+P_RandomRange(5, 30)*mo.scale, MT_SMOKE)
			P_SetObjectMomZ(steam, P_RandomRange(1, 3)*mo.scale)

			if not (leveltime%2)
				local fire = P_SpawnMobj(mo.x+P_RandomRange(-45, 45)*mo.scale, mo.y+P_RandomRange(-45, 45)*FRACUNIT, mo.z+P_RandomRange(5, 30)*mo.scale, MT_THOK)
				fire.sprite = SPR_FPRT
				fire.frame = $ & ~FF_TRANSMASK
				fire.frame = $|FF_FULLBRIGHT|TR_TRANS20
				fire.scale = mo.scale*5/2
				fire.momz = P_RandomRange(1, 3)*mo.scale
				fire.scalespeed = mo.scale/12
				fire.destscale = 1
				fire.tics = TICRATE
			end		
			
			
			if timer % (TICRATE) == 0
				
				-- deal damage
				
				local dmg = min(info.hp-1, info.maxhp/25)
				
				if dmg and dmg > 0
					PSO.lib.damageObject(mo, nil, dmg)
				end
			end
		end,
		
	},

	[COND_FREEZE] = {
		
		duration = TICRATE*5,
		
		startfunc = function(mo)
			mo.freezeframe = mo.frame
			mo.freezesprite2 = mo.sprite2
			mo.freezeangle = mo.angle
			mo.colorized = true
			mo.freezecolor = mo.color
			mo.color = SKINCOLOR_CYAN
			
		end,
		
		runfunc = function(mo, timer)
			
			local info
			if mo.player
				info = mo.player.pso
				mo.player.cmd.angleturn = mo.freezeangle/FRACUNIT
				mo.player.drawangle = mo.freezeangle
			else
				info = mo.pso
				mo.angle = mo.freezeangle
				mo.target = nil
			end
			
			mo.player.normalspeed = 0
			P_InstaThrust(mo, 0, 0)
			mo.momz = 0
			
			mo.sprite2 = mo.freezesprite2
			mo.frame = mo.freezeframe
			if mo.tics > 0
				mo.tics = $+1
			end
			
		end,
		
		endfunc = function(mo)
			mo.colorized = false
			mo.color = mo.freezecolor
			mo.freezeframe = nil
			mo.freezecolor = nil
			mo.freezeangle = nil
		end,
		
	},

	[COND_SHOCK] = {
		
		duration = TICRATE*10,
		
		startfunc = function(mo)
			
			mo.colorized = true
			mo.shockcolor = mo.color
			mo.color = SKINCOLOR_YELLOW
			
		end,
		
		runfunc = function(mo, timer)
			mo.momx = $/2
			mo.momy = $/2
			
		end,
		
		endfunc = function(mo)
			mo.colorized = false
			mo.color = mo.shockcolor
			mo.shockcolor = nil
		end,
	},

	[COND_POISON] = {
		
		duration = TICRATE*10,
		
		runfunc = function(mo, timer)
			
			local info
			if mo.player
				info = mo.player.pso
			else
				info = mo.pso
			end
			
			-- visuals
			for i = 1,2
				local psn = P_SpawnMobj(mo.x+P_RandomRange(-15,15)*FRACUNIT, mo.y+P_RandomRange(-15,15)*FRACUNIT, mo.z+P_RandomRange(0,15)*FRACUNIT, MT_THOK)
				psn.sprite = SPR_THOK
				psn.frame = A
				psn.color = SKINCOLOR_PURPLE
				psn.scale = FRACUNIT/2
				psn.destscale = FRACUNIT/99
				psn.momz = P_RandomRange(5, 12)*FRACUNIT
				psn.tics = TICRATE
			end
			
			
			if timer % (TICRATE) == 0
				
				-- deal damage
				
				local dmg = min(info.hp-1, info.maxhp/25)
				
				if dmg and dmg > 0
					PSO.lib.damageObject(mo, nil, dmg)
				end
			end
		end,
		
	},

	[COND_CONFUSE] = {
		
		duration = TICRATE*10,
		
		startfunc = function(mo)
			
			mo.colorized = true
			mo.confcolor = mo.color
			mo.color = SKINCOLOR_WHITE
			
		end,
		
		runfunc = function(mo, timer)
			-- stops enemy movement, players have a different behaviour
			if not mo.player
				mo.momx = 0
				mo.momy = 0
			end
			
		end,
		
		endfunc = function(mo)
			mo.colorized = false
			mo.color = mo.confcolor
			mo.confcolor = nil
		end,
	},
	
}

PSO.lib.removeStatus = function(mo)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	if not info.condition
		return
	end
	
	local dta = PSO.info.statusConditions[info.condition]
	
	if dta
		
		if dta.endfunc
			dta.endfunc(mo)
		end
		
		info.condition = nil
		info.conditiontimer = nil
	end
end

PSO.lib.inflictStatus = function(mo, cond)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	-- status is already applied
	if info.condition == cond
		return
	end

	PSO.lib.removeStatus(mo)
	
	local dta = PSO.info.statusConditions[cond]
	
	if dta
		
		if dta.startfunc
			dta.startfunc(mo)
		end
		
		info.condition = cond
		info.conditiontimer = dta.duration
	end
end

COM_AddCommand("psod_status", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	local cond = tonumber(arg)
	PSO.lib.inflictStatus(p.mo, cond)
	
end)

PSO.lib.handleStatus = function(mo)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	if info.condition
		
		local dta = PSO.info.statusConditions[info.condition]
		
		if dta.runfunc
			dta.runfunc(mo, info.conditiontimer)
		end
		
		info.conditiontimer = $-1
		
		if not info.conditiontimer
			PSO.lib.removeStatus(mo)			
		end
		
	end
	
	-- stackable buffs/debuffs
	-- we'll handle these separatly since it doesn't matter too much...
	if info.shifta
		info.shifta = $-1
	end
	
	if info.deband
		info.deband = $-1
	end
	
	if info.jellen
		info.jellen = $-1
	end
	
	if info.zallure
		info.zallure = $-1
	end
end