-- Bird Aircraft Strike Hazard
/*
	Stationary on ground, leaps, charges up and charges towards the target at high speeds.
*/

PSO.info.enemies[MT_VULTURE] = {
		
		name = "B.A.S.H.",
		sprite = SPR_VLTR,
		frame = A,
		painframe = A,
		
		bhp = 150,
		satk = 35,
		ratk = 20,
		tatk = 20,
		def = 25,
		exp = 7,
		weight = 1,
		
		
		drops = {
			-- Lv 1-10
			{100, 1, 9, 1, 15},				-- meseta (100% mult) 2/5 (Lv1-10)
			{"sword", 1, 9, 1, 100},		-- sword: 	1/100 (Lv1-10)
			{"rifle", 1, 9, 1, 100},		-- rifle: 	1/100 (Lv1-10)
			{"rod", 1, 9, 1, 100},			-- rod: 	1/100 (Lv1-10)
			{"gigush", 1, 9, 1, 50},
			
			{"monomate", 1, 9, 1, 25},	-- monomate: 1/5 (Lv1-10)
			
			-- Lv10-19
			
			{200, 10, 19, 1, 15},
			{"gigush", 10, 19, 1, 100},
			{"beam", 10, 19, 1, 100},
			{"cane", 10, 19, 1, 100},
			{"monomate", 10, 19, 1, 25},
			{"breaker", 10, 19, 1, 50},

			-- Lv20-29
			{300, 20, 29, 1, 15},
			{"breaker", 20, 29, 1, 150},
			{"doublegladius", 20, 29, 1, 150},
			{"laser", 20, 29, 1, 150},
			{"canon", 20, 29, 1, 150},
			{"ulrod", 20, 29, 1, 150},
			{"claymore", 20, 29, 1, 50},
			{"dimate", 20, 29, 1, 25},

			{650, 30, 50, 1, 15},
			{"claymore", 30, 50, 1, 50},
			{"alva claymore", 30, 50, 1, 75},
			{"vita claymore", 30, 50, 1, 100},
			{"vulture basher", 30, 50, 1, 250},
			{"rostig frame", 30, 50, 1, 30},
			{"aileron frame", 30, 50, 1, 30},			
			{"dimate", 30, 50, 1, 35},	
			{"trimate", 30, 50, 1, 35},				
			
		},
		
		stagger_duration = TICRATE/2,
		
		weakness = ELEM_ELEC,
		
		reset =	function(mo)
			if not mo.pso.blowdown
				mo.flags = $|MF_NOGRAVITY
				mo.momz = 0
			end
		end,
		
		thinker = function(mo)

			local diff = PSO.lib.questDifficulty()
			local attack_potency = 350
			local targetz = mo.floorz + mo.scale*32
			local rate = FRACUNIT*8/10
			local attack_wait = diff >= DIFF_HARD and TICRATE*3 or TICRATE*4
			local charge_visual_speed = 10
			local charge_speed = mo.scale*38
			local charge_extra = TICRATE/2
			
			PSO.lib.enemyManageAggro(mo)
			
			mo.flags = $ & ~(MF_FLOAT|MF_NOGRAVITY)
				
			if mo.z < targetz
				mo.momz = $ + rate
			elseif mo.momz < 0
				mo.momz = min(-mo.scale, $ + rate*2)
			end
			
			if mo.target and mo.target.valid
				
				if mo.actiontimer
					mo.actiontimer = $-1
					
					P_InstaThrust(mo, mo.angle, charge_speed)
					if leveltime & 1
						P_SpawnGhostMobj(mo)
					end
					
					if not S_SoundPlaying(mo, sfx_jet)
						S_StartSound(mo, sfx_jet)
					end
					
					-- update hitbox
					if mo.hitbox and mo.hitbox.valid
						local hx = mo.x + 64*cos(mo.angle)
						local hy = mo.y + 64*sin(mo.angle)
								
						P_MoveOrigin(mo.hitbox, hx, hy, mo.z)	
					end
					
				elseif not mo.chargestrike
					P_InstaThrust(mo, 0, 0)
					mo.chargestrike = attack_wait
				else
					P_InstaThrust(mo, 0, 0)
					mo.chargestrike = $-1
					
					mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
					
					if mo.chargestrike < attack_wait/2
						
						local t = P_SpawnMobj(mo.x + P_RandomRange(-128, 128)*mo.scale, mo.y + P_RandomRange(-128, 128)*mo.scale, mo.z + P_RandomRange(-128, 128)*mo.scale, MT_THOK)
						t.tics = charge_visual_speed
						t.color = SKINCOLOR_RED
						t.frame = A|FF_ADD
						t.scale = mo.scale
						t.destscale = 1
						t.scalespeed = mo.scale/12
						
						t.momx = (mo.x - t.x)/charge_visual_speed
						t.momy = (mo.y - t.y)/charge_visual_speed
						t.momz = (mo.z - t.z)/charge_visual_speed
						
						if not mo.chargestrike
							-- figure out how long we should charge
							local dist = R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y)
							local time = FixedDiv(dist, charge_speed) / FRACUNIT
							
							mo.actiontimer = 9999	--time + charge_extra
							mo.pso.superarmor = true
							
							-- spawn hurtbox
							local hx = mo.x + 64*cos(mo.angle)
							local hy = mo.y + 64*sin(mo.angle)
							
							local hb = PSO.lib.spawnHurtBox(hx, hy, mo.z, 80<<FRACBITS, 40<<FRACBITS, mo, attack_potency, DMT_STRIKE, mo.actiontimer, RT_KNOCKBACK)
							mo.hitbox = hb
							hb.angle = mo.angle
							hb.pso.reactionstrength = 80
							
						end
					end
				end
			end
			
			mo.sprite = SPR_VLTR
			mo.frame = A
		end,
}

addHook("MobjMoveBlocked", function(mo)
	if mo and mo.pso and mo.actiontimer
		-- bump the wall
		P_InstaThrust(mo, mo.angle, -FRACUNIT*8)
		mo.angle = $ + ANG1*180
		mo.momz = FRACUNIT*8
		mo.pso.blowdown = 1
		mo.pso.superarmor = false
		mo.actiontimer = nil
	end	
end, MT_VULTURE)