-- Black Flow
/*
	Depths Area boss, Phase 1
	
	Similar setting to Olga Flow from PSO epII
	Uses mostly ranged attacks with exploding projectiles.
*/

local BFS_NEUTRAL = 0
local BFS_TELEPORT = 1
local BFS_FRONTAOE = 2
local BFS_FRONTLINE = 3
local BFS_JUKE = 4
local BFS_LASER = 5
local BFS_CROSS = 6


freeslot("MT_PSO_BLACKFLOW", "SPR_BFLW", "S_PSO_BLACKFLOW_DEATH")
mobjinfo[MT_PSO_BLACKFLOW] = {
	doomednum = 2553,
	spawnhealth = 1000,
	deathstate = S_PSO_BLACKFLOW_DEATH,
	spawnstate = S_THOK,
	height = 120<<FRACBITS,
	radius = 50<<FRACBITS,
	flags = MF_ENEMY|MF_SOLID|MF_SHOOTABLE|MF_NOGRAVITY,
}

function A_BlackFlowDeathLoop(mo)
	mo.deathtimer = $ and $+1 or 1
	mo.flags = $|MF_NOBLOCKMAP|MF_NOCLIPTHING
	
	for p in players.iterate
		
		if p.mo and p.mo.valid
			p.mo.bflowdeath = mo.deathtimer
		end
	end
	
	if leveltime%2 == 0
	and mo.deathtimer < TICRATE*5
		
		local boom = P_SpawnMobj(mo.x + P_RandomRange(-128, 128)*FRACUNIT, mo.y + P_RandomRange(-128, 128)*FRACUNIT, mo.z + P_RandomRange(0, 256)*FRACUNIT, MT_SONIC3KBOSSEXPLODE)
		boom.colorized = true
		boom.color = SKINCOLOR_EMERALD
		boom.renderflags = $|RF_FULLBRIGHT
		boom.tics = $ + P_RandomRange(0, 4)
		boom.scale = FRACUNIT*3 + P_RandomRange(-FRACUNIT/3, FRACUNIT/3)
		
		S_StartSound(mo, sfx_s3kb4)
	end
	
	if mo.deathtimer == TICRATE*3
		S_StartSound(nil, sfx_bedie2)
	end
	
	if mo.deathtimer == TICRATE*5
		if mo.pteleport
			for p in players.iterate do
				if p.mo and p.mo.valid
					P_SetOrigin(p.mo, mo.pteleport.x, mo.pteleport.y, mo.pteleport.z)
				end
			end
		end
	end
	
	if mo.deathtimer == TICRATE*7
		for p in players.iterate
			if p.mo and p.mo.valid
				p.mo.bflowdeath = nil
			end
		end
		P_RemoveMobj(mo)
	end
	
end
states[S_PSO_BLACKFLOW_DEATH] = {SPR_BRAK, N, 1, A_BlackFlowDeathLoop, 1, 1, S_PSO_BLACKFLOW_DEATH}

-- VV annoying VV
hud.add(function(v, p, c)
	if p and p.mo and p.mo.valid
	and p.mo.bflowdeath	
		local timer = p.mo.bflowdeath
		
		local delay = TICRATE*3
		local t = (timer - delay)/5
		
		if t >= 0
		and timer <= TICRATE*7 -50
			v.fadeScreen(0, min(10, t))
		end
		
		if timer >= TICRATE*7 -50
		and timer <= TICRATE*7
			local t2 = (timer - (TICRATE*7-50)) /5
			v.fadeScreen(0, max(0, 10-t2))
		end
		
	end
end)

freeslot("MT_PSO_BFLOW_EXPLOSION")
mobjinfo[MT_PSO_BFLOW_EXPLOSION] = {
	doomednum = -1,
	spawnhealth = 1000,
	deathstate = S_NULL,	-- todo: change
	spawnstate = S_INVISIBLE,
	height = 64<<FRACBITS,
	radius = 32<<FRACBITS,
	flags = MF_NOGRAVITY,
}

addHook("MobjThinker", function(mo)
	mo.tics = -1
	mo.fuse = -1
	
	if mo.delay
		mo.delay = $-1
		return
	end
	
	mo.existtimer = $ and $+1 or 1
	
	/*local tflag = 4*cos(leveltime*ANG1*14) /FRACUNIT
	local splat = P_SpawnMobj(mo.x, mo.y, mo.z + FRACUNIT*8, MT_THOK)
	splat.flags2 = $ | MF2_SPLAT
	splat.frame = A|FF_FULLBRIGHT| (3 + tflag)<<FF_TRANSSHIFT
	splat.scale = FRACUNIT*2
	splat.color = SKINCOLOR_EMERALD
	splat.tics = 2*/
	
	if mo.existtimer == 1
		local s = P_SpawnMobj(mo.x, mo.y, mo.z + FRACUNIT*32, MT_THOK)
		s.state = S_ICESPARK1
		s.color = SKINCOLOR_WHITE
		s.scale = FRACUNIT*3
			
		S_StartSound(mo, sfx_sparks)
	elseif mo.existtimer == TICRATE
		for i = 1, 24
			local t = P_SpawnMobj(mo.x, mo.y, mo.z, i & 1 and MT_THOK or MT_SONIC3KBOSSEXPLODE)
			if i & 1
				t.frame = A|TR_TRANS10|FF_FULLBRIGHT
				t.fuse = 20
				t.tics = 20	
				t.scale = $*2/3
			else
				t.colorized = true
				t.renderflags = $|RF_FULLBRIGHT
				t.scale = $*3
			end
			
			t.color = SKINCOLOR_EMERALD
			t.destscale = 1
			t.scalespeed = mo.scale/12
			t.momx = P_RandomRange(-16, 16)*mo.scale
			t.momy = P_RandomRange(-16, 16)*mo.scale
			t.momz = P_RandomRange(-16, 16)*mo.scale			
		end	
		
		S_StartSound(mo, sfx_foie)
		
		-- spawn an explosion hitbox:
		local h = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, 256<<FRACBITS, 200<<FRACBITS, mo.target, mo.pso.potency, DMT_RANGED, 4, RT_BLOWDOWN)
		h.angle = mo.angle
	elseif mo.existtimer >= TICRATE*2
		P_RemoveMobj(mo)
		return
	end
end, MT_PSO_BFLOW_EXPLOSION)


freeslot("MT_PSO_BFLOW_LASER")
mobjinfo[MT_PSO_BFLOW_LASER] = {
	doomednum = -1,
	spawnhealth = 1000,
	deathstate = S_NULL,	-- todo: change
	spawnstate = S_INVISIBLE,
	height = 64<<FRACBITS,
	radius = 32<<FRACBITS,
	flags = MF_NOGRAVITY,
}
addHook("MobjThinker", function(mo)
	
	mo.existtimer = $ and $+1 or 1
	
	mo.sprite = SPR_BFLW
	mo.frame = (B + leveltime%4 / 2)|FF_FULLBRIGHT
	mo.colorized = true
	mo.color = SKINCOLOR_EMERALD
	mo.scale = FRACUNIT
	
	mo.spriteyscale = FRACUNIT*24
	
	if mo.existtimer == 1
		local h = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z-FRACUNIT*256, 256<<FRACBITS, 896<<FRACBITS, mo.target, mo.pso.potency, DMT_RANGED, TICRATE*4, RT_BLOWDOWN)
		h.angle = mo.angle
		h.pso.reactionradialangle = true
	end
	
	if mo.existtimer < 10
		mo.spritexscale = max(1, mo.existtimer*FRACUNIT/10)
	elseif mo.existtimer > TICRATE*4 - 10
		local tmr = mo.existtimer - (TICRATE*4 - 10)
		mo.spritexscale = max(1, FRACUNIT - tmr*FRACUNIT/10)
	else
		mo.spritexscale = FRACUNIT
	end
	
	if mo.existtimer >= TICRATE*4
		P_RemoveMobj(mo)
		return
	end
	
end, MT_PSO_BFLOW_LASER)

local function bflow_spawnWings(mo)
	
	mo.wings = {}
	for i = 1, 4
		
		local w = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		w.tics = -1
		w.fuse = -1
		w.sprite = SPR_BFLW
		w.frame = A|FF_FULLBRIGHT|FF_PAPERSPRITE|FF_ADD
		w.scale = FRACUNIT*2
		w.flags = $|MF_NOCLIPHEIGHT|MF_NOCLIP
		
		mo.wings[#mo.wings+1] = w
	end
	
	-- search for a boss waypoint to reference...
	for m in mobjs.iterate()
		
		if m and m.valid and m.type == MT_BOSS3WAYPOINT
			mo.tracer = m
			break
		end
	end

	for m in mobjs.iterate()
		
		if m and m.valid and m.type == MT_BOSSFLYPOINT
			mo.pteleport = m
			break
		end
	end
	
	if mo.tracer and mo.tracer.valid
		mo.refdist = R_PointToDist2(mo.x, mo.y, mo.tracer.x, mo.tracer.y) /FRACUNIT
		mo.refangle = R_PointToAngle2(mo.tracer.x, mo.tracer.y, mo.x, mo.y)
	end	
	
	assert(mo.tracer, "Black Flow needs an MT_BOSS3WAYPOINT for dist reference")
end

local function bflow_neutral(mo)
	return not mo.bstate
end

local function bflow_moveWings(mo)
	
	local behind_dist = 80
	
	for i = 1, 2
		local rot = ANG1*20
		local dist = 240
		local ang = mo.angle + ANG1*90 + ANG1*180*(i-1)
		local zpos = mo.z + mo.height*2/3
		
		for j = 1, 2
			
			local wing = mo.wings[(i-1)*2 +j]
			local behindx = mo.x - behind_dist*cos(mo.angle)
			local behindy = mo.y - behind_dist*sin(mo.angle)
			
			local wx = behindx + dist*cos(ang)
			local wy = behindy + dist*sin(ang)
			P_MoveOrigin(wing, wx, wy, zpos)
			wing.angle = ang
			wing.rollangle = rot
			wing.spritexscale = mo.spritexscale
			wing.spriteyscale = mo.spriteyscale
			
			dist = $ + 160
			rot = $+ANG1*20
			zpos = $ + FRACUNIT*80
		end
	end
end

local function bflow_float(mo)
	
	local amp = 64
	
	if mo.basez == nil
		mo.basez = mo.z
	end
	mo.floattimer = $ and $+1 or 1
	
	mo.z = mo.basez + amp*sin(mo.floattimer*ANG1*3)
end

-- initializes a teleport to the asked location or otherwise a random one
local function bflow_initTeleport(mo, target)
	
	mo.bstate = BFS_TELEPORT
	
	if target ~= nil
		mo.newang = target
	else
		mo.newang = P_RandomRange(0, 359)*ANG1
	end
	mo.teleport = 1
	mo.teleportstate = 1
	
	S_StartSound(nil, sfx_s3kb6)
end

-- frontal AOE cone attack
local function bflow_initForwardsAOE(mo)
	mo.bstate = BFS_FRONTAOE
	mo.frontaoe = 1
	
	S_StartSound(nil, sfx_beragh)
end

-- frontal line attack, faster than the cone attack
local function bflow_initForwardsLine(mo)
	mo.bstate = BFS_FRONTLINE
	mo.frontline = 1
	
	if P_RandomRange(0, 1)
		S_StartSound(nil, sfx_bewar4)
	else
		S_StartSound(nil, sfx_bewar1)
	end
end

-- starts juking, teleporting around to fake a position and then attacks the player
local function bflow_initJuke(mo)
	mo.bstate = BFS_JUKE
	mo.juke = 1
end

-- initializes a laser at the target location
local function bflow_initLaser(mo)
	mo.bstate = BFS_LASER
	mo.laser = 1

	mo.targetx = nil
	mo.targety = nil
	mo.targetz = nil
	
	S_StartSound(nil, sfx_s3k9d)
end

local function bflow_initCross(mo)
	mo.bstate = BFS_CROSS
	mo.cross = 1
end

local function bflow_doTeleport(mo)
	
	if mo.teleport
		
		mo.spriteyscale = mo.teleport*FRACUNIT
		mo.spritexscale = FRACUNIT/mo.teleport

		if mo.teleportstate == 1
			mo.teleport = $+1
			mo.colorized = true
			mo.color = SKINCOLOR_WHITE
			mo.renderflags = $|RF_FULLBRIGHT
		else
			mo.teleport = $-1
			
			if not mo.teleport
				mo.bstate = BFS_NEUTRAL
				mo.colorized = false
				mo.renderflags = $ & ~RF_FULLBRIGHT
			end
		end
		
		if mo.teleport == 6
			
			local nx = mo.tracer.x + mo.refdist*cos(mo.newang)
			local ny = mo.tracer.y + mo.refdist*sin(mo.newang)
			
			P_SetOrigin(mo, nx, ny, mo.basez)
			mo.refangle = mo.newang
			mo.teleportstate = 2
			mo.angle = mo.newang + ANG1*180
		end	
	end
end

local function bflow_doFrontAOE(mo)
	
	local diff = PSO.lib.questDifficulty()
	local wait = TICRATE*3/2
	local endlag = TICRATE*2
	local range = 3
	local angrange = ANG1*70
	local potency = 400
	
	if mo.frontaoe
		
		mo.frontaoe = $+1
		
		if mo.frontaoe < wait
			mo.frame = S
		else
			
			local tmr = mo.frontaoe-wait
			local frame = min(U, S +tmr/3)
			
			mo.frame = frame
			
			if mo.frontaoe == wait
				
				for i = 1, range
					local dist = 224 + 128*i
					local numexplosions = i*3
					local ang = mo.angle + angrange
					
					-- we sweep from ang-70 to ang+70 (140 deg)
					local sweep = (2*angrange) / (numexplosions-1)
					
					for j = 1, numexplosions
						
						local x = mo.x + dist*cos(ang)
						local y = mo.y + dist*sin(ang)
						
						local msl = PSO.lib.spawnMissile(mo, MT_PSO_BFLOW_EXPLOSION, 0, 0, potency, DMT_RANGED, 0, 0, 0)
						P_MoveOrigin(msl, x, y, mo.basez + FRACUNIT*64)
						
						msl.delay = i*TICRATE/(diff >= DIFF_HARD and 2 or 3)
						msl.angle = ang
						
						ang = $-sweep
					end
				end
			elseif mo.frontaoe >= wait + endlag
				mo.frontaoe = nil
				mo.bstate = BFS_NEUTRAL
			end
		end
	end
end

local function bflow_doFrontLine(mo)
	
	local wait = TICRATE/2
	local endlag = TICRATE*2
	local range = 16
	local potency = 400
	local diff = PSO.lib.questDifficulty()
	
	if mo.frontline
		
		mo.frontline = $+1
		
		if mo.frontline < wait
			mo.frame = S
		else
			
			local tmr = mo.frontline-wait
			local frame = min(U, S +tmr/3)
			
			mo.frame = frame
			
			if mo.frontline == wait
				
				for i = 1, range
					local dist = 224 + 96*i
					local ang = mo.angle
						
					local x = mo.x + dist*cos(ang)
					local y = mo.y + dist*sin(ang)
						
					local msl = PSO.lib.spawnMissile(mo, MT_PSO_BFLOW_EXPLOSION, 0, 0, potency, DMT_RANGED, 0, 0, 0)
					P_MoveOrigin(msl, x, y, mo.basez + FRACUNIT*64)
						
					if diff >= DIFF_HARD
						msl.delay = TICRATE/8 + i*TICRATE/12
					else
						msl.delay = TICRATE/6 + i*TICRATE/8
					end
					msl.angle = ang
				end
			elseif mo.frontline >= wait + endlag
				mo.frontline = nil
				mo.bstate = BFS_NEUTRAL
			end
		end
	end
end

local function bflow_doJuke(mo)
	
	local finaljuke = TICRATE*4
	local endlag = TICRATE*4
	
	if mo.juke
		
		mo.juke = $+1
		
		if mo.juke < finaljuke
			if mo.juke%24 == 0
				
				bflow_initTeleport(mo)
				if PSO.lib.questDifficulty() >= DIFF_HARD
				and mo.pso.hp < mo.pso.maxhp/2
					bflow_initForwardsLine(mo)
				end
				
			end
		elseif mo.juke == finaljuke
			
			-- check where "behind" is
			local behind = R_PointToAngle2(mo.target.x, mo.target.y, mo.tracer.x, mo.tracer.y)
			
			if abs(mo.target.angle - behind)/ANG1 <= 90
				behind = $ + ANG1*180
			end
			
			bflow_initTeleport(mo, behind)
		elseif mo.juke == finaljuke + 6
			bflow_initForwardsLine(mo)
		elseif mo.juke == finaljuke+endlag
			mo.juke = nil
			mo.bfstate = BFS_NEUTRAL
		end
	end
end

local function bflow_doLaser(mo)
	
	local laser_wait = TICRATE/4
	local laser_tracking = TICRATE*3 + TICRATE/2
	local wait = TICRATE*4 + TICRATE/2
	
	local endlag = TICRATE*2
	local potency = 400
	
	local range = 8
	
	if mo.laser
		
		mo.laser = $+1
		
		if mo.laser < wait
			mo.frame = S

			if mo.laser > laser_wait
			and mo.laser < laser_tracking
				mo.targetx = mo.target.x
				mo.targety = mo.target.y
				mo.targetz = mo.target.z
			end
			
			if mo.targetx ~= nil
			and leveltime & 1	
				local s = P_SpawnMobj(mo.targetx, mo.targety, mo.targetz, MT_THOK)
				s.sprite = SPR_TARG
				s.frame = (leveltime%14)/2 |FF_FULLBRIGHT
				s.tics = 1		
			end
			
		else
			
			local tmr = mo.laser-wait
			local frame = min(U, S +tmr/3)
			
			mo.frame = frame			
			if mo.laser == wait
				local msl = PSO.lib.spawnMissile(mo, MT_PSO_BFLOW_LASER, 0, 0, potency*3/2, DMT_RANGED, 0, 0, 0)
				P_MoveOrigin(msl, mo.targetx, mo.targety, mo.targetz)
				
				S_StartSound(nil, sfx_s3k83)
				
				-- spawn the explosions!
				for i = 1, range
					local dist = 128*i
					local numexplosions = 2 + 2*i
					
					local ang = 0
					local angrange = (360 / numexplosions) *ANG1
					
					for j = 1, numexplosions
						local x = msl.x + dist*cos(ang)
						local y = msl.y + dist*sin(ang)
						
						local boom = PSO.lib.spawnMissile(mo, MT_PSO_BFLOW_EXPLOSION, 0, 0, potency, DMT_RANGED, 0, 0, 0)
						boom.angle = ang
						boom.delay = TICRATE*3 + (i-1)*TICRATE
						P_MoveOrigin(boom, x, y, mo.basez + FRACUNIT*64)
						
						ang = $ + angrange
					end	
				end
				
				
			elseif mo.laser >= wait + endlag
				mo.laser = nil
				mo.bstate = BFS_NEUTRAL
				
				mo.targetx = nil
				mo.targety = nil
				mo.targetz = nil
			end
		end
	end	
end	

-- do 2 lasers in a cross pattern
local function bflow_doCross(mo)
	
	if mo.cross
		
		mo.cross = $+1
		
		if mo.cross == 2
			mo.cross_ang = P_RandomRange(0, 359)*ANG1
			bflow_initTeleport(mo, mo.cross_ang)
		elseif mo.cross == 14
			bflow_initForwardsLine(mo)
		elseif mo.cross == TICRATE+2
			mo.frontline = nil
			bflow_initTeleport(mo, mo.cross_ang+ANG1*90)
		elseif mo.cross == TICRATE+14
			bflow_initForwardsLine(mo)
			mo.cross = nil
		end
	end
end

local function bflow_intro(mo)
	
	local wait = TICRATE
	
	if not mo.introd
		mo.intro = $ and $+1 or 1

		for p in players.iterate do
		
			local pmo = p.mo
			if pmo and pmo.valid and (p.powers[pw_nocontrol] or R_PointToDist2(mo.x, mo.y, pmo.x, pmo.y) < FRACUNIT*2048)
				if mo.intro >= wait+TICRATE*3
					p.powers[pw_nocontrol] = 0
				else
					p.powers[pw_nocontrol] = 32767
				end
			end
		end
		
		if mo.intro < wait
			mo.frame = S
		else
			
			if mo.intro == wait
				S_StartSound(nil, sfx_bewar3)
			end
			
			local tmr = mo.intro-wait
			local frame = min(U, S +tmr/3)	
			mo.frame = frame
			
			if tmr < TICRATE*3
			and leveltime%5 == 0
				
				local howl = P_SpawnMobj(mo.x, mo.y, mo.z + FRACUNIT*192, MT_THOK)
				howl.state = S_ROAR1
				howl.angle = mo.angle
				howl.scale = FRACUNIT*8
				howl.destscale = FRACUNIT*32
				howl.scalespeed = FRACUNIT/3
				howl.color = SKINCOLOR_EMERALD
				howl.renderflags = $|RF_FULLBRIGHT
				P_InstaThrust(howl, howl.angle, FRACUNIT*16)
			end			
			
			
			if tmr > TICRATE*4
				mo.introd = true
				mo.intro = nil
				return
			end
		end
	end
end

PSO.info.enemies[MT_PSO_BLACKFLOW] = {
		
		name = "Black Flow",
		sprite = SPR_BRAK,
		frame = A,
		painframe = A,
		
		bhp = 5000,
		satk = 35,
		ratk = 35,
		tatk = 35,
		def = 20,
		exp = 250,
		weight = 4,

		boss = true,		-- enemy is a major boss!
		bossmusic = "FBOSS1",
		bossrushmusic = "BOSSR4",
		
		
		drops = {
			{"pine", 1, 29, 1, 100},
			{"claymore", 1, 29, 1, 100},
			{"alva pine", 1, 29, 1, 100},
			{"alva claymore", 1, 29, 1, 100},
			
			{"doubliss", 1, 29, 1, 100},
			{"doublecutter", 1, 29, 1, 100},
			{"alva doubliss", 1, 29, 1, 100},
			{"alva doublecutter", 1, 29, 1, 100},	
			
			{"shooter", 1, 29, 1, 100},
			{"burster", 1, 29, 1, 100},
			{"alva shooter", 1, 29, 1, 100},
			{"alva burster", 1, 29, 1, 100},				
			
			{"shot", 1, 29, 1, 100},
			{"concentrator", 1, 29, 1, 100},	
			{"alva shot", 1, 29, 1, 100},
			{"alva concentrator", 1, 29, 1, 100},	
			
			{"crest", 1, 29, 1, 100},
			{"crescent", 1, 29, 1, 100},
			{"alva crest", 1, 29, 1, 100},
			{"alva crescent", 1, 29, 1, 100},
				
			{"claw", 1, 29, 1, 100},
			{"cutter", 1, 29, 1, 100},
			{"alva claw", 1, 29, 1, 100},
			{"alva cutter", 1, 29, 1, 100},			
			
			{"beamgun", 1, 29, 1, 100},
			{"peacekeeper", 1, 29, 1, 100},
			{"alva beamgun", 1, 29, 1, 100},
			{"alva peacekeeper", 1, 29, 1, 100},			
					
			{"hammer", 1, 29, 1, 100},
			{"scepter", 1, 29, 1, 100},
			{"alva hammer", 1, 29, 1, 100},
			{"alva scepter", 1, 29, 1, 100},
			
			{"icicle frame", 1, 29, 1, 20},
			{"knight frame", 1, 29, 1, 20},
			
			{"unit_hp_2", 1, 29, 2, 50},
			{"unit_pp_2", 1, 29, 2, 50},
			{"unit_satk_2", 1, 29, 2, 50},
			{"unit_tatk_2", 1, 29, 2, 50},
			{"unit_ratk_2", 1, 29, 2, 50},

			{"digrinder", 1, 29, 5, 50},
				
			{10000, nil, nil, 1, 7},
		},
		
		stagger_duration = TICRATE/2,
		
		weakness = ELEM_LIGHT,
		
		reset =	function(mo)
			mo.actiontimer = 0	-- reset action if we get staggered etc.
		end,
		
		thinker = function(mo)
			
			PSO.lib.enemyManageAggro(mo)
				
			mo.sprite = SPR_BRAK
			mo.scale = FRACUNIT*2
			mo.shadowscale = FRACUNIT*2

			mo.frame = N

			if not mo.wings
				bflow_spawnWings(mo)
			end
			
			bflow_moveWings(mo)
			
			if not mo.introd
				bflow_intro(mo)
				return
			end
			
			mo.existtimer = $ and $+1 or 1
			if mo.existtimer == TICRATE*5
				P_StartQuake(FRACUNIT*30, 10)
				S_StartSound(nil, sfx_doord2)
				server.targetelevatorspeed = 100
			end	
			
			bflow_float(mo)
			
			if mo.target and mo.target.valid
				bflow_doTeleport(mo)
				bflow_doFrontLine(mo)
				bflow_doFrontAOE(mo)
				bflow_doJuke(mo)
				bflow_doLaser(mo)
				bflow_doCross(mo)
				
				if bflow_neutral(mo)
				and not mo.juke
				and not mo.cross
					mo.waittimer = $ and $+1 or 1
					
					if mo.waittimer > TICRATE*5
						
						local dist = R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y)
						
						-- numbers straight outta nowhere
						if dist < 684*FRACUNIT
							-- very close range:
							local rng = P_RandomRange(0, 5)
							
							if rng == 0
								bflow_initForwardsAOE(mo)
							elseif rng == 1
								bflow_initForwardsLine(mo)
							elseif rng == 2
								bflow_initLaser(mo)
								mo.waittimer = TICRATE*4
								return
							elseif rng == 3
								bflow_initCross(mo)
							elseif rng == 4
								bflow_initJuke(mo)
							else
								bflow_doTeleport(mo)
								mo.waittimer = TICRATE*4
								return
							end
							
						else
							
							local rng = P_RandomRange(0, 9)
							
							if rng < 4
								bflow_initCross(mo)
							elseif rng < 6
								bflow_initJuke(mo)
							elseif rng < 9
								bflow_initLaser(mo)
								mo.waittimer = TICRATE*4
								return							
							else
								bflow_doTeleport(mo)
								mo.waittimer = TICRATE*4
								return
							end
						end
						
						--bflow_initTeleport(mo)
						--bflow_initForwardsAOE(mo)
						--bflow_initForwardsLine(mo)
						--bflow_initJuke(mo)
						--bflow_initLaser(mo)
						--bflow_initCross(mo)
						mo.waittimer = 0
					end
				end
			end
		end,
}