-- Red Crawla
/*
	Drives towards players.
	Once close enough, attempt to charge with a quick charge that closes the gap. Has a cooldown.
	
	HARD MODE: Charges attack faster
*/

PSO.info.enemies[MT_REDCRAWLA] = {
		
		name = "Red Crawla",
		sprite = SPR_SPOS,
		frame = A,
		painframe = A,
		
		bhp = 200,
		satk = 35,
		ratk = 20,
		tatk = 20,
		def = 25,
		exp = 8,
		weight = 1,
		
		
		drops = {
			-- Lv 1-10
			{100, 1, 9, 1, 15},				-- meseta (100% mult) 2/5 (Lv1-10)
			{"sword", 1, 9, 1, 100},		-- sword: 	1/100 (Lv1-10)
			{"rifle", 1, 9, 1, 100},		-- rifle: 	1/100 (Lv1-10)
			{"rod", 1, 9, 1, 100},			-- rod: 	1/100 (Lv1-10)
			{"lockgun", 1, 9, 1, 50},
			
			{"monomate", 1, 9, 1, 25},	-- monomate: 1/5 (Lv1-10)
			
			-- Lv10-19
			
			{200, 10, 19, 1, 15},
			{"gigush", 10, 19, 1, 100},
			{"beam", 10, 19, 1, 100},
			{"cane", 10, 19, 1, 100},
			{"monomate", 10, 19, 1, 25},
			{"autogun", 10, 19, 1, 50},

			-- Lv20-29
			{300, 20, 29, 1, 15},
			{"breaker", 20, 29, 1, 150},
			{"doublegladius", 20, 29, 1, 150},
			{"laser", 20, 29, 1, 150},
			{"canon", 20, 29, 1, 150},
			{"ulrod", 20, 29, 1, 150},
			{"peacekeeper", 20, 29, 1, 50},
			{"dimate", 20, 29, 1, 25},

			-- Lv30-50
			{650, 30, 50, 1, 15},
			{"claw", 30, 50, 1, 50},
			{"alva claw", 30, 50, 1, 75},
			{"vita claw", 30, 50, 1, 100},
			{"poss shooter", 30, 50, 1, 200},
			{"rostig frame", 30, 50, 1, 30},
			{"fehler frame", 30, 50, 1, 100},			
			{"dimate", 30, 50, 1, 35},
			
		},
		
		stagger_duration = TICRATE/2,
		
		weakness = ELEM_ICE,
		
		reset = function(mo)
			mo.actiontimer = nil
			mo.attackcooldown = TICRATE
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			
			local attack_potency = 250
			local move_speed = mo.scale*(diff >= DIFF_HARD and 4 or 3)
			local attack_range = mo.scale*356
			local attack_charge = diff >= DIFF_HARD and TICRATE*2/3 or TICRATE
			local attack_duration = TICRATE
			local attack_delay = TICRATE
			local attack_speed = mo.scale*50
			local anim_speed = 1
			
			PSO.lib.enemyManageAggro(mo)
			
			mo.sprite = SPR_SPOS
			mo.frame = A
			
			if mo.target and mo.target.valid
				
				if not mo.actiontimer
					A_FaceTarget(mo)
					P_InstaThrust(mo, mo.angle, move_speed)
					
					mo.frame = (leveltime%(anim_speed*6)) /anim_speed
					
					-- wait for a cooldown
					if mo.attackcooldown
						mo.attackcooldown = $-1
					elseif R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y) < attack_range
						mo.actiontimer = 1
						S_StartSound(mo, sfx_spndsh)
					end
				else
					mo.actiontimer = $+1
					
					if mo.actiontimer < attack_charge
						if mo.actiontimer % 10 == 0
							S_StartSound(mo, sfx_spndsh)
						end
						
						local dust = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
						dust.angle = mo.angle + ANG1*180 + P_RandomRange(-30, 30)*ANG1
						P_InstaThrust(dust, dust.angle, P_RandomRange(8, 16)*FRACUNIT)
						dust.momz = P_RandomRange(4, 8)*FRACUNIT
						
					
					elseif mo.actiontimer == attack_charge
						mo.superarmor = true
						P_InstaThrust(mo, mo.angle, attack_speed)
						S_StartSound(mo, sfx_zoom)

						local hx = mo.x + 64*cos(mo.angle)
						local hy = mo.y + 64*sin(mo.angle)
						
						local hb = PSO.lib.spawnHurtBox(hx, hy, mo.z, 80<<FRACBITS, 40<<FRACBITS, mo, attack_potency, DMT_STRIKE, attack_duration, RT_KNOCKBACK)
						mo.hitbox = hb
						hb.angle = mo.angle
						hb.pso.reactionstrength = 80
						
					end
					
					if mo.actiontimer >= attack_charge
					and mo.actiontimer < attack_charge + attack_duration
						if mo.hitbox and mo.hitbox.valid
							local hx = mo.x + 64*cos(mo.angle)
							local hy = mo.y + 64*sin(mo.angle)
							
							P_MoveOrigin(mo.hitbox, hx, hy, mo.z)
						end
						
						P_SpawnGhostMobj(mo)
						
					elseif mo.actiontimer >= attack_charge + attack_duration
						mo.momx = $*94/100
						mo.momy = $*94/100
					
						if FixedHypot(mo.momx, mo.momy) < move_speed
							mo.superarmor = nil
							mo.actiontimer = 0
							mo.attackcooldown = TICRATE*5
						end
					end
				end
			end
		end,
		
}
