-- Robo Hood
/*
	Fires arrows towards the player.
	If the player gets too close, jumps back as it fires.
*/

PSO.info.enemies[MT_ROBOHOOD] = {
		
		name = "Dust Hood",
		sprite = SPR_ARCH,
		frame = A,
		painframe = A,
		
		bhp = 150,
		satk = 20,
		ratk = 35,
		tatk = 20,
		def = 20,
		exp = 7,
		weight = 1,
		
		
		drops = {
			-- Lv 1-10
			{100, 1, 9, 1, 15},				-- meseta (100% mult) 2/5 (Lv1-10)
			{"sword", 1, 9, 1, 100},		-- sword: 	1/100 (Lv1-10)
			{"rifle", 1, 9, 1, 100},		-- rifle: 	1/100 (Lv1-10)
			{"rod", 1, 9, 1, 100},			-- rod: 	1/100 (Lv1-10)
			{"lockgun", 1, 9, 1, 50},
			
			{"monomate", 1, 9, 1, 25},	-- monomate: 1/5 (Lv1-10)
			
			-- Lv10-19
			
			{200, 10, 19, 1, 15},
			{"gigush", 10, 19, 1, 100},
			{"beam", 10, 19, 1, 100},
			{"cane", 10, 19, 1, 100},
			{"monomate", 10, 19, 1, 25},
			{"autogun", 10, 19, 1, 50},

			-- Lv20-29
			{300, 20, 29, 1, 15},
			{"breaker", 20, 29, 1, 150},
			{"doublegladius", 20, 29, 1, 150},
			{"laser", 20, 29, 1, 150},
			{"canon", 20, 29, 1, 150},
			{"ulrod", 20, 29, 1, 150},
			{"peacekeeper", 20, 29, 1, 50},
			{"dimate", 20, 29, 1, 25},

			-- Lv30-50
			{650, 30, 50, 1, 15},
			{"hammer", 30, 50, 1, 50},
			{"alva hammer", 30, 50, 1, 75},
			{"vita hammer", 30, 50, 1, 100},
			{"rostig frame", 30, 50, 1, 20},		
			{"fehler frame", 30, 50, 1, 80},
			{"dimate", 30, 50, 1, 35},
			{"monogrinder", 30, 50, 1, 20},
			
		},
		
		stagger_duration = TICRATE/2,
		
		weakness = ELEM_EARTH,
		
		reset =	function(mo)
			mo.actiontimer = 0	-- reset action if we get staggered etc.
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			local arrow_wait = diff >= DIFF_HARD and TICRATE*5/2 or TICRATE*4
			local arrow_speed = 25
			local arrow_potency = 200
			
			PSO.lib.enemyManageAggro(mo)
			
			mo.sprite = SPR_ARCH
			mo.frame = A
			
			if mo.target and mo.target.valid
				
				local dist = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
				local zdist = mo.z - mo.target.z
				if mo.waittimer == nil
					mo.waittimer = arrow_wait
				end
				
				mo.waittimer = max(0, $-1)
				
				if mo.waittimer < arrow_wait/3
					mo.frame = C
				end
				
				mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
				
				if not P_IsObjectOnGround(mo)
					mo.frame = B
					return
				end
				
				if not mo.waittimer
					
					if dist > 256*mo.scale
					or abs(zdist) > 128*mo.scale
					or mo.jumpbuffer
						local msl = PSO.lib.spawnMissile(mo, MT_ARROW, mo.angle, arrow_speed*FRACUNIT, arrow_potency, DMT_RANGED, 0, 0, mo.height/2, RT_STAGGER)
						PSO.lib.aimMissile(msl, mo.target)				
						msl.flags = $|MF_NOGRAVITY
						mo.jumpbuffer = nil
						
						if diff >= DIFF_HARD
							local a = msl.angle +ANG1*90
							local dist = 64
							for i = 1, 2
								
								local msl2 = P_SpawnMobj(msl.x + dist*cos(a), msl.y + dist*sin(a), msl.z, msl.type)
								msl2.flags = msl.flags
								msl2.angle = msl.angle
								
								msl2.pso = {}
								for k, v in pairs(msl.pso)
									msl2.pso[k] = v
								end
								
								msl2.momx = msl.momx
								msl2.momy = msl.momy
								msl2.momz = msl.momz
								msl2.target = msl.target
								
								a = $+ANG1*180
							end
						end
						
						mo.waittimer = arrow_wait
						
						S_StartSound(mo, sfx_s3k67)
					else
						P_InstaThrust(mo, mo.angle, -mo.scale*10)
						mo.momz = mo.scale*8
						mo.waittimer = arrow_wait/3
						mo.jumpbuffer = true
					end
					
				end
			end
		end,	
}