-- Snapper Boss
/*
	Walks towards player.
	Legs are independant and must be destroyed to attack the head
	Otherwise, lunge at the player as they try to attack the head.
	When no legs are left, go in the shell 
	
	HARD MODE: 
	Leg slams cast Radiga on players
	Bodyslam spawns rocks that aim on players
	Adds laser attack when far away
*/

freeslot("MT_PSO_SNAPPERREMAIN", "MT_PSO_SNAPPERBOSS", "MT_PSO_SNAPPERBOSSHEAD", "MT_PSO_SNAPPERBOSSLEG", "S_PSO_SNAPPERDEATH")
freeslot("sfx_sroar1")

mobjinfo[MT_PSO_SNAPPERREMAIN] = {
	
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	radius = 1,
	height = 1,
	flags = MF_NOCLIPHEIGHT|MF_NOCLIP,
}

freeslot("SPR_HOWL", "S_ROAR1", "S_ROAR2", "S_ROAR3")
local tlookup = {TR_TRANS50, TR_TRANS70, TR_TRANS80}
for i = 0, 2 do
	states[S_ROAR1+i] = {SPR_HOWL, i|tlookup[i+1], 4, nil, 0, 0, i < 2 and S_ROAR2+i or S_NULL}
end

addHook("MobjThinker", function(mo)
	
	if leveltime & 1
		mo.flags2 = $|MF2_DONTDRAW
	else
		mo.flags2 = $ & ~MF2_DONTDRAW
	end
	
	mo.rollangle = $ + ANG1*32
	
	if leveltime%4 == 0
		S_StartSound(mo, sfx_pop)
		P_SpawnMobj(mo.x + P_RandomRange(-64, 64)*FRACUNIT, mo.y + P_RandomRange(-64, 64)*FRACUNIT, mo.z + P_RandomRange(0, 64)*FRACUNIT, MT_BOSSEXPLODE)
	end
	
end, MT_PSO_SNAPPERREMAIN)

function A_PSOSnapperDie(mo)
	
	local dta = PSO.info.enemies[mo.type]
	
	-- spawn a dummy:
	local dum = P_SpawnMobj(mo.x, mo.y, mo.z, MT_PSO_SNAPPERREMAIN)
	dum.scale = mo.scale
	dum.sprite = dta.sprite
	dum.frame = dta.frame
	
	dum.momx = P_RandomRange(-16, 16)*FRACUNIT
	dum.momy = P_RandomRange(-16, 16)*FRACUNIT
	dum.momz = P_RandomRange(12, 16)*FRACUNIT
	dum.fuse = TICRATE*4
end

states[S_PSO_SNAPPERDEATH] = {SPR_NULL, A, 1, A_PSOSnapperDie, 0, 0, S_NULL}

mobjinfo[MT_PSO_SNAPPERBOSS] = {
	doomednum = 2550,
	spawnhealth = 1000,
	deathstate = S_PSO_SNAPPERDEATH,
	spawnstate = S_INVISIBLE,
	height = 32<<FRACBITS,
	radius = 32<<FRACBITS,
	flags = MF_ENEMY|MF_SOLID,
}

mobjinfo[MT_PSO_SNAPPERBOSSHEAD] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	deathstate = S_PSO_SNAPPERDEATH,
	spawnhealth = 1,
	height = 32<<FRACBITS,
	radius = 16<<FRACBITS,
	flags = MF_ENEMY|MF_SHOOTABLE|MF_NOGRAVITY,
}

mobjinfo[MT_PSO_SNAPPERBOSSLEG] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	deathstate = S_PSO_SNAPPERDEATH,
	pawnhealth = 1,
	height = 32<<FRACBITS,
	radius = 16<<FRACBITS,
	flags = MF_ENEMY|MF_SHOOTABLE|MF_NOGRAVITY,
}

PSO.info.enemies[MT_PSO_SNAPPERBOSSHEAD] = {
		
		name = "Rock Snapper",
		sprite = SPR_GSNH,
		frame = A,
		painframe = A,
		
		part = true,
		partlinkdamage = true,
		
		bhp = 100,
		satk = 50,
		ratk = 40,
		tatk = 1,
		def = 30,
		exp = 0,
		weight = 4,
}

PSO.info.enemies[MT_PSO_SNAPPERBOSSLEG] = {
		
		name = "Rock Snapper",
		sprite = SPR_GSNL,
		frame = A,
		painframe = A,
		
		part = true,			-- enemy is a part of another enemy
		partlinkdamage = true,	-- damage taken by the part is also taken by the main body
		
		bhp = 19,	--% HP of the source object that spawned it.
		satk = 50,
		ratk = 40,
		tatk = 1,
		def = 30,
		exp = 0,
		weight = 4,
}

-- spawn my limbs kek
local function snapperBossSpawnLimbs(mo)
	
	local a = mo.angle
	local hdist = 32
	
	-- give me a solid body...
	/*mo.body = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	mo.body.state = S_INVISIBLE
	mo.body.flags = MF_SOLID*/

	local push = P_SpawnMobj(mo.x, mo.y, mo.z, MT_PSO_PUSHBOX)
	push.target = mo
	
	local hx = mo.x + FixedMul(mo.scale, hdist*cos(a))*3/2
	local hy = mo.y + FixedMul(mo.scale, hdist*sin(a))*3/2
	local hz = mo.z + 4*mo.scale
	
	local head = PSO.lib.enemyCreatePart(mo, hx, hy, hz, MT_PSO_SNAPPERBOSSHEAD)	--P_SpawnMobj(hx, hy, hz, MT_PSO_SNAPPERBOSSHEAD)
	head.scale = mo.scale
	
	push = P_SpawnMobj(hx, hy, hz, MT_PSO_PUSHBOX)
	push.target = head
	
	mo.head = head
	
	mo.legs = {}
	local angs = {mo.angle + ANG1*60, mo.angle + ANG1*60 + ANG1*50, mo.angle - ANG1*60, mo.angle - ANG1*60 - ANG1*50}
	
	for i = 1, 4
		
		local ang = angs[i]
		local hx = mo.x + FixedMul(mo.scale, hdist*cos(ang))
		local hy = mo.y + FixedMul(mo.scale, hdist*sin(ang))
		
		local leg = PSO.lib.enemyCreatePart(mo, hx, hy, hz, MT_PSO_SNAPPERBOSSLEG)	--P_SpawnMobj(hx, hy, hz, MT_PSO_SNAPPERBOSSLEG)
		mo.legs[i] = leg
		leg.scale = mo.scale

		push = P_SpawnMobj(hx, hy, hz, MT_PSO_PUSHBOX)
		push.target = leg	
		
	end
end

local function snapperBossTeleportLimbs(mo)
	
	
	local a = mo.angle
	local hdist = 32
	
	local hx = mo.x + FixedMul(mo.scale, hdist*cos(a))*3/2
	local hy = mo.y + FixedMul(mo.scale, hdist*sin(a))*3/2
	local hz = mo.z + 4*mo.scale
	local stomp_potency = 350
	
	P_MoveOrigin(mo.head, hx, hy, hz)
	mo.head.angle = mo.angle
	
	mo.head.sprite = SPR_GSNH
	mo.head.frame = A
	
	local angs = {mo.angle + ANG1*60, mo.angle + ANG1*60 + ANG1*50, mo.angle - ANG1*60, mo.angle - ANG1*60 - ANG1*50}
	for i = 1, #mo.legs

		local leg = mo.legs[i]
		
		if not leg or not leg.valid
			continue
		end
		
		leg.flags = $ & ~MF_SOLID
		
		local ang = angs[i]
		local hx = mo.x + FixedMul(mo.scale, hdist*cos(ang))
		local hy = mo.y + FixedMul(mo.scale, hdist*sin(ang))
		local hz = mo.z
		
		if leg.stomptimer
			if leg.stomptimer < TICRATE*3/2
				
				hz = $ + min(leg.stomptimer, 10)*mo.scale
				if leg.stomptimer > 10
					hz = $ + P_RandomRange(-1, 1)*mo.scale
				end
			
			elseif leg.stomptimer < TICRATE*3/2 + 5
				hz = $ + 14*mo.scale
			
			elseif leg.stomptimer < TICRATE*3/2 + 10
				hz = mo.z
				
			elseif leg.stomptimer == TICRATE*3/2 + 10
				
				for i = 1, 16
					local d = P_SpawnMobj(leg.x, leg.y, leg.z, MT_SPINDUST)
					d.angle = (360/16)*i*ANG1
					P_InstaThrust(d, d.angle, FRACUNIT*32)
					d.momz = FRACUNIT*8
				end
				
				local hb = PSO.lib.spawnHurtBox(leg.x, leg.y, leg.z - FRACUNIT*4, 320<<FRACBITS, 40<<FRACBITS, mo, stomp_potency, DMT_STRIKE, 8, RT_BLOWDOWN)
				hb.pso.reactionangle = leg.angle
				hb.pso.reactionstrength = 15
				
				S_StartSound(leg, sfx_pstop)
				
				leg.disablepushbox = nil
				mo.stomp = nil
			end
			
			
			leg.stomptimer = $+1
		end
		
		-- simple walking anim using basic grade school math trig
		if mo.moving
			local offset = (i == 1 or i == 4) and 180*ANG1 or 0
			
			hz = max(mo.z, $ + FixedMul(mo.scale, 6*cos(mo.moving*ANG1*4 + offset)))
			-- forward angle:
			local fa = mo.angle
			local dist = FixedMul(mo.scale, 5*cos(mo.moving*ANG1*4 + offset - ANG1*90))
			hx = $ + FixedMul(dist, cos(fa))
			hy = $ + FixedMul(dist, sin(fa))
		end
		
		leg.dest = {hx, hy, hz}
		
		--P_MoveOrigin(leg, hx, hy, hz)
		leg.momx = (leg.dest[1] - leg.x) /2
		leg.momy = (leg.dest[2] - leg.y) /2
		leg.momz = (leg.dest[3] - leg.z) /2
		
		leg.angle = R_PointToAngle2(mo.x, mo.y, leg.x, leg.y)
		
		leg.sprite = SPR_GSNL
		leg.frame = A
		
		if mo.slide
			leg.frame = B
		end
	end
end

local function snapperNeutral(mo)
	
	if mo.slide
	or mo.stomp
	or mo.chase
	or mo.jump
		return false
	end
	return true
end

-- tells us which leg the target is closest to
local function whichLeg(mo, target)
	local closest = target.scale*256
	local retleg
	
	for i = 1, 4
		
		local leg = mo.legs[i]
		if not leg or not leg.valid
			continue
		end
		
		local dist = R_PointToDist2(leg.x, leg.y, target.x, target.y)
		if dist < closest
			closest = dist
			retleg = leg
		end
	end
	
	return retleg
end

local function snapperInitStomp(mo, leg)
	
	if not snapperNeutral(mo) return end
	if mo.stompcd return end
	
	mo.stomp = leg	-- so we know that we're doing it
	leg.stomptimer = 1
	leg.disablepushbox = true
	
	mo.stompcd = TICRATE*10
	mo.playerpos = {}
	
	--print("begin stomping")
end

local function snapperHandleStompCD(mo)
	
	local radiga_potency = 300
	
	if mo.stompcd
		
		local diff = PSO.lib.questDifficulty()
		local digtime = TICRATE*2
		local stoptrack = TICRATE/2
		
		local tmr = mo.stompcd - TICRATE*5
		
		mo.stompcd = $-1
		
		-- radiga on players we can see
		if diff >= DIFF_HARD
			
			for p in players.iterate do
				if p and p.mo and p.mo.valid
					local pmo = p.mo
					
					if tmr > stoptrack
						local pmo = p.mo
						if P_CheckSight(mo, pmo)
							mo.playerpos[#p] = {pmo.x, pmo.y}
						end
					end
				
					if tmr < digtime
					and tmr
						if mo.playerpos[#p]
							local t = digtime - tmr
							local scale = max(1, (t*FRACUNIT*2/(digtime)))
							local surface = P_SpawnMobj(mo.playerpos[#p][1] + P_RandomRange(-4, 4)*scale, mo.playerpos[#p][2] + P_RandomRange(-4, 4)*scale, pmo.floorz, MT_THOK)
							surface.sprite = SPR_MNUD
							surface.frame = F
							surface.scale = scale
							surface.tics = 10
						end
					elseif tmr == 0
						
						-- cast radiga
						for i = 1, 8
							
							local x = mo.playerpos[#p][1] + P_RandomRange(-48, 48)*FRACUNIT*2
							local y = mo.playerpos[#p][2] + P_RandomRange(-48, 48)*FRACUNIT*2
							
							local r = P_SpawnMobj(x, y, pmo.z, MT_THOK)
							r.state = S_ROCKCRUMBLEA
							r.fuse = P_RandomRange(20, 35)
							r.scale = 1
							r.destscale = FRACUNIT
							r.scalespeed = FRACUNIT / P_RandomRange(3, 8)
								
							r.momz = P_RandomRange(4, 16)*FRACUNIT*2
							
							r.spriteyscale = FRACUNIT*3 + P_RandomRange(0, FRACUNIT-7)
							
							r.color = SKINCOLOR_ORANGE
							r.colorized = true
						end
						
						S_StartSound(pmo, sfx_rocks3)
						
						local hb = PSO.lib.spawnHurtBox(mo.playerpos[#p][1], mo.playerpos[#p][2], pmo.floorz, 96<<FRACBITS, 180<<FRACBITS, mo, radiga_potency, DMT_TECHNIQUE, 4, RT_BLOWDOWN)
						hb.pso.element = ELEM_EARTH
						hb.angle = R_PointToAngle2(mo.x, mo.y, pmo.x, pmo.y)
						
					end
				end
			end
		end
		
		if not mo.stompcd
			mo.stomp = nil
			mo.playerpos = nil
		end
	end
end


local function snapperInitSlide(mo)
	
	if not snapperNeutral(mo) return end
	
	local slide_potency = 400
	
	mo.slide = true
	P_InstaThrust(mo, mo.angle, FRACUNIT*32)
	
	-- put a bunch of hitboxes all around us!
	local ang = mo.angle - ANG1*36*2
	local hdist = 24
	
	mo.slidehitboxes = {}
	
	for i = 1, 5
		local hx = mo.x + FixedMul(mo.scale, hdist*cos(ang))
		local hy = mo.y + FixedMul(mo.scale, hdist*sin(ang))

		local hb = PSO.lib.spawnHurtBox(hx, hy, mo.z, 128<<FRACBITS, 40<<FRACBITS, mo, slide_potency, DMT_STRIKE, TICRATE*2, RT_BLOWDOWN)
		hb.pso.reactionangle = mo.angle
		hb.pso.reactionstrength = 16			
		
		mo.slidehitboxes[i] = hb
		
		ang = $ + ANG1*36
	end
	
	for i = 1, 5
		mo.slidehitboxes[i].pso.links = mo.slidehitboxes
	end

	mo.disablepushbox = true
	mo.head.disablepushbox = true
		
	for i = 1, 4
		if mo.legs[i] and mo.legs[i].valid
			mo.legs[i].disablepushbox = true
		end
	end
	
	--print("begin sliding")
end

local function snapperHandleSlide(mo)
	
	if mo.slide
		mo.friction = FRACUNIT - FRACUNIT/32
		for i = 1, #mo.slidehitboxes
			local hb = mo.slidehitboxes[i]
			
			if not hb or not hb.valid
				continue
			end
			
			hb.momx = mo.momx
			hb.momy = mo.momy
			hb.momz = mo.momz
			
			local dust = P_SpawnMobj(hb.x, hb.y, hb.z, MT_SPINDUST)
			dust.momx = mo.momx/2 + P_RandomRange(-8, 8)*FRACUNIT
			dust.momy = mo.momy/2 + P_RandomRange(-8, 8)*FRACUNIT
			dust.momz = P_RandomRange(0, 8)*FRACUNIT
		end
	end
	
	if FixedHypot(mo.momx, mo.momy) < FRACUNIT*2
		mo.slide = nil

		mo.disablepushbox = nil
		mo.head.disablepushbox = nil
		
		for i = 1, 4
			if mo.legs[i] and mo.legs[i].valid
				mo.legs[i].disablepushbox = nil
			end
		end
		
	end
end

local function snapperInitJump(mo)
	
	if not snapperNeutral(mo) return end
	
	mo.jump = 1
	mo.jumpdir = P_RandomRange(0, 1) and -1 or 1
	mo.momz = FRACUNIT*30
	mo.z = $+FRACUNIT
	mo.disablepushbox = true
	mo.head.disablepushbox = true
	
	for i = 1, 4
		if mo.legs[i] and mo.legs[i].valid
			mo.legs[i].disablepushbox = true
		end
	end
	
	S_StartSound(mo, sfx_beflap)
	
end

local function snapperHandleJump(mo)
	
	if not mo.jump return end
	local crush_potency = 400
	local rock_potency = 200
	local jump_time = TICRATE*3
	
	local diff = PSO.lib.questDifficulty()
	
	mo.jump = $+1
	
	if mo.momz > 0
		mo.momz = $ - P_GetMobjGravity(mo)/2
	elseif mo.jump < jump_time
		mo.momz = 0
	end
	
	mo.angle = $ + ANG1/3*mo.jumpdir
	
	if leveltime & 1
		local ghostme = {mo, mo.head}
		for i = 1, #mo.legs
			if mo.legs[i] and mo.legs[i].valid
				ghostme[#ghostme+1] = mo.legs[i]
			end
		end
		
		for i = 1, #ghostme
			
			local g = P_SpawnGhostMobj(ghostme[i])
			g.color = SKINCOLOR_GREEN
			g.colorized = true
		end
	end
	
	if leveltime%20 == 0
		for i = 1, 8
			local a = (360/8)*i*ANG1
			local d = P_SpawnMobj(mo.x, mo.y, mo.floorz, MT_SPINDUST)
			d.scale = $*2
			d.angle = a
			P_InstaThrust(d, d.angle, FRACUNIT*64)
		end
	end
	
	if P_IsObjectOnGround(mo)
		
		S_StartSound(leg, sfx_pstop)
		P_StartQuake(FRACUNIT*30, 10)
		
		local dist = 128
		local hitbox_group = {}
		
		for i = 1, 8
			local a = (360/8)*i*ANG1
			local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
			d.scale = $*2
			d.angle = a
			P_InstaThrust(d, d.angle, FRACUNIT*64)
			
			local hx = mo.x + dist*cos(a)
			local hy = mo.y + dist*sin(a)

			local hb = PSO.lib.spawnHurtBox(hx, hy, mo.z, 320<<FRACBITS, 40<<FRACBITS, mo, crush_potency, DMT_STRIKE, 3, RT_BLOWDOWN)
			hb.pso.reactionangle = a
			hb.pso.reactionstrength = 16
			hitbox_group[#hitbox_group+1] = hb
		end

		for i = 1, #hitbox_group
			hitbox_group[i].pso.links = hitbox_group
		end

		mo.disablepushbox = nil
		mo.head.disablepushbox = nil
		
		for i = 1, 4
			if mo.legs[i] and mo.legs[i].valid
				mo.legs[i].disablepushbox = nil
			end
		end

		
		if diff >= DIFF_HARD
			for i = 1, 96
				
				local msl = PSO.lib.spawnMissile(mo, MT_DIGA, mo.angle, 0, rock_potency, DMT_TECHNIQUE, 0, 0, mo.height/2, RT_BLOWDOWN)
				local x = mo.x + P_RandomRange(-256, 256)*FRACUNIT
				local y = mo.y + P_RandomRange(-256, 256)*FRACUNIT
				local z = mo.z + FRACUNIT*2
				
				msl.pso.element = ELEM_EARTH
				msl.scale = FRACUNIT
				msl.radius = $/2
				msl.height = $/2
				
				P_MoveOrigin(msl, x, y, z)
				
				msl.color = SKINCOLOR_ORANGE
				msl.colorized = true
				msl.scale = $*2
				
				local a = R_PointToAngle2(mo.x, mo.y, msl.x, msl.y)
				P_InstaThrust(msl, a, P_RandomRange(2, 16)*FRACUNIT)
				msl.momz = P_RandomRange(15, 26)*FRACUNIT
			end
		end
		
		mo.jump = nil
	end
end

local function snapperInitChase(mo)	
	if not snapperNeutral(mo) return end
	
	mo.chase = P_RandomRange(TICRATE*4, TICRATE*10)
	mo.haschased = mo.chase + TICRATE*7
	mo.moving = 1
end


local function snapperChaseHandler(mo)
	
	if mo.chase
		local dist = R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y)
		local max_move_speed = FRACUNIT*6
		local max_turn = ANG1*1
		
		-- count my legs...
		local numlegs = 0
		for i = 1, 4
			if mo.legs[i] and mo.legs[i].valid
				numlegs = $+1
			end
		end
		
		
		if not numlegs
		--or (mo.moving > TICRATE*4 and dist < 512*FRACUNIT)
			mo.chase = nil
			mo.moving = 0
			return
		end
		
		max_move_speed = $*numlegs/4
		
		mo.moving = $+1
		
		P_InstaThrust(mo, mo.angle, min(max_move_speed, mo.moving*(max_move_speed/(TICRATE))))
		
		-- change angle...
		local targetangle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
		local dif = PSO.lib.angToInt(targetangle - mo.angle)
		
		
		if dif < 180 and dif > max_turn/ANG1
			mo.angle = $ + max_turn
		elseif dif >= 180 and dif < PSO.lib.angToInt(InvAngle(max_turn))
			mo.angle = $ - max_turn
		else
			mo.angle = targetangle
			
			if (dist < 420*FRACUNIT or dist > 1024*FRACUNIT) and mo.moving > TICRATE*3/2
				local rng = P_RandomRange(0, 255)
				if rng < 1 or dist < 420*FRACUNIT
					mo.chase = nil
					mo.moving = 0
					snapperInitSlide(mo)
					mo.haschased = nil
					return
				end
			end
		end
		
		mo.chase = $-1
		
		if not mo.chase
			mo.moving = 0
		end
	end
end

local function initSnapperPinball(mo)
	
	if not snapperNeutral(mo) return end
	
	mo.pinball = 1
	mo.pinballangle = mo.angle
	mo.momz = FRACUNIT*32
	S_StartSound(mo, sfx_beflap)
end

local function snapperHandlePinball(mo)
	
	local pinball_potency = 400
	local pinball_duration = TICRATE*8
	local diff = PSO.lib.questDifficulty()
	
	if mo.pinball
		
		mo.pinball = $+1
		
		local angspeed = min(ANG1*45, ANG1/2*mo.pinball)
		
		mo.angle = $ + angspeed
		
		if mo.pinball == TICRATE
			mo.head.flags2 = $|MF2_DONTDRAW
			mo.head.flags = $|MF_NOBLOCKMAP|MF_NOCLIPTHING
			mo.flags = $|MF_NOCLIPTHING
			
			mo.head.disablepushbox = true
			mo.disablepushbox = true
		end
		
		if leveltime & 1
			local g = P_SpawnGhostMobj(mo)
			g.colorized = true
			g.color = SKINCOLOR_GREEN

			if diff >= DIFF_HARD
				for i = 1, 2
					
					local msl = PSO.lib.spawnMissile(mo, MT_DIGA, mo.angle, 0, rock_potency, DMT_TECHNIQUE, 0, 0, mo.height/2, RT_BLOWDOWN)
					local x = mo.x + P_RandomRange(-256, 256)*FRACUNIT
					local y = mo.y + P_RandomRange(-256, 256)*FRACUNIT
					local z = mo.z + FRACUNIT*2
					
					msl.pso.element = ELEM_EARTH
					msl.scale = FRACUNIT
					msl.radius = $/2
					msl.height = $/2
					
					P_MoveOrigin(msl, x, y, z)
					
					msl.color = SKINCOLOR_ORANGE
					msl.colorized = true
					msl.scale = $*2
					
					local a = R_PointToAngle2(mo.x, mo.y, msl.x, msl.y)
					P_InstaThrust(msl, a, P_RandomRange(2, 16)*FRACUNIT)
					msl.momz = P_RandomRange(15, 26)*FRACUNIT
				end
			end
			
		end
		
		if mo.pinball == TICRATE*5/2
			P_InstaThrust(mo, mo.pinballangle, FRACUNIT*48)
			mo.friction = FRACUNIT
			mo.flags = $|MF_BOUNCE
			mo.flags = $ & ~MF_SOLID
			
		elseif mo.pinball > TICRATE*5/2
		and mo.pinball < pinball_duration
			mo.friction = FRACUNIT	-- don't slow down
			
			local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, 420<<FRACBITS, 160<<FRACBITS, mo, pinball_potency, DMT_STRIKE, 3, RT_BLOWDOWN)
		
		elseif mo.pinball >= pinball_duration
			
			if mo.pinball == pinball_duration
				
				P_InstaThrust(mo, 0, 0)
				mo.momz = FRACUNIT*30				
			
			elseif mo.pinball == pinball_duration + TICRATE/2
				
				mo.head.flags2 = $ & ~MF2_DONTDRAW
				mo.disablepushbox = nil
				mo.head.disablepushbox = nil
				mo.pinball = nil
				mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
				mo.waitpinball = TICRATE*16
				
				mo.flags = $ & ~MF_NOCLIPTHING
				mo.head.flags = $ & ~(MF_NOBLOCKMAP|MF_NOCLIPTHING)
				
			end
		end
	end
end

local function snapperRoar(mo)
	
	if not mo.introd
		mo.introtimer = $ and $+1 or 1
		
		for p in players.iterate do
		
			local pmo = p.mo
			if pmo and pmo.valid and (p.powers[pw_nocontrol] or R_PointToDist2(mo.x, mo.y, pmo.x, pmo.y) < FRACUNIT*1536)
				if mo.introtimer == TICRATE*3
					p.powers[pw_nocontrol] = 0
				else
					p.powers[pw_nocontrol] = 32767
				end
			end
		end
		
		if mo.introtimer == TICRATE/2
			P_StartQuake(FRACUNIT*30, TICRATE*3)
			S_StartSound(nil, sfx_sroar1)
		end
		
		if mo.introtimer >= TICRATE/2
		and mo.introtimer <= TICRATE*3
			mo.head.frame = B
			
			if leveltime%5 == 0
				
				local howl = P_SpawnMobj(mo.head.x, mo.head.y, mo.head.z, MT_THOK)
				howl.state = S_ROAR1
				howl.angle = mo.head.angle
				howl.scale = FRACUNIT*8
				howl.destscale = FRACUNIT*32
				howl.scalespeed = FRACUNIT/3
				P_InstaThrust(howl, howl.angle, FRACUNIT*16)
			end
		end
		
		if mo.introtimer == TICRATE*3
			mo.introd = true
		end
	end	
end

PSO.info.enemies[MT_PSO_SNAPPERBOSS] = {
		
		name = "Rock Snapper",
		sprite = SPR_GSNP,
		frame = A,
		painframe = A,
		
		notarget = true,	-- can't be targetted.
		boss = true,		-- enemy is a major boss!
		bossmusic = "BOSS",
		bossrushmusic = "BOSSR1",
		
		bhp = 6000,
		satk = 35,
		ratk = 35,
		tatk = 35,
		def = 30,
		exp = 300,
		weight = 4,
		
		
		drops = {
			
			{"gigush", 1, 9, 2, 50},
			{"beam", 1, 9, 2, 50},
			{"baton", 1, 9, 2, 50},		
			{"giga calibur", 1, 9, 2, 50},
			{"giga edge", 1, 9, 2, 50},	
			{"burstbeam", 1, 9, 2, 50},
			{"longbeam", 1, 9, 2, 50},
			{"helixen", 1, 9, 2, 50},	
			{"half-helixen", 1, 9, 2, 50},		
			-- 16/50
			
			{"tutel frame", 1, 29, 1, 10},	-- 21/50
			{"unit_hp_1", 1, 29, 1, 20},		-- 23/50
			{"unit_pp_1", 1, 29, 1, 20},		-- 25/50
			{"unit_hppp_1", 1, 29, 1, 20},	-- 27/50


			-- lv30+
			{"orakl sword", 30, 50, 1, 400},
			{"orakl doublesaber", 30, 50, 1, 400},
			{"orakl soaringblade", 30, 50, 1, 400},
			{"orakl rifle", 30, 50, 1, 400},
			{"orakl shot", 30, 50, 1, 400},
			{"orakl orphan", 30, 50, 1, 400},
			{"orakl rod", 30, 50, 1, 400},
			{"orakl wand", 30, 50, 1, 400},
			{"smasher frame", 30, 50, 1, 32},

			{"unit_hp_2", 30, 50, 1, 20},
			{"unit_pp_2", 30, 50, 1, 20},
			{"unit_hppp_2", 30, 50, 1, 20},
			{"unit_power_2", 30, 50, 1, 20},
			{"unit_rainbow_2", 30, 50, 1, 20},
			
			-- permanent drops
			{"monogrinder", nil, nil, 5, 50},
			{"digrinder", nil, nil, 5, 50},
			
			{10000, nil, nil, 2, 50},
			{"moon atomizer", nil, nil, 5, 50},
			{"star atomizer", nil, nil, 5, 50},
		},
		
		weakness = ELEM_ELEC,
		
		reset =	function(mo)
			mo.actiontimer = 0	-- reset action if we get staggered etc.
		end,
		
		thinker = function(mo)
						
			PSO.lib.enemyManageAggro(mo)
			
			-- this enemy is too big to have the solid flag
			-- this would make us step ontop of the player lmao
			mo.flags = $ & ~MF_SOLID
			
			mo.sprite = SPR_GSNP
			mo.frame = A
			mo.scale = FRACUNIT*6
			
			if not mo.head
				snapperBossSpawnLimbs(mo)
				--mo.moving = 1
				mo.wait = 1
				mo.attack_cooldown = TICRATE*10	-- don't try to attack for the next 10 seconds
			end
			
			
			if not mo.introd
				snapperBossTeleportLimbs(mo)
				snapperRoar(mo)
				return
			end
			
			if mo.moving
				mo.moving = $+1
			end
			
			if mo.target and mo.target.valid
				
				-- count our legs (lol)
				local numlegs = 0
				for i = 1, 4
					if mo.legs[i] and mo.legs[i].valid
						numlegs = $+1
					end
				end
				
				if numlegs
					
					mo.hadlegs = true
					
					local dist = R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y)
									
					if not mo.chase
					and not mo.haschased
					--and dist > 768*FRACUNIT
						snapperInitChase(mo)
					elseif mo.haschased
						mo.haschased = $-1
						
						if not snapperNeutral(mo)
							mo.haschased = max(TICRATE*2, $)
						end
						
						--print(mo.haschased)
						
						if not mo.haschased
						and (dist > 768*FRACUNIT or not P_RandomRange(0, 2))
							
							snapperInitJump(mo)
							P_InstaThrust(mo, R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y), dist/128)
							
						end
					end
					
					-- stomp priority is last
					local leg = whichLeg(mo, mo.target)
					if leg
						snapperInitStomp(mo, leg)
					end				
				else
					
					if mo.hadlegs
						mo.chase = nil
						mo.jump = nil
						mo.stomp = nil
						mo.stompcd = nil
						mo.moving = nil
						
						mo.hadlegs = nil
						
						mo.disablepushbox = nil
						mo.head.disablepushbox = nil
						
						mo.waitpinball = TICRATE*2
					end
					
					if mo.waitpinball
						mo.waitpinball = $-1
						--print(mo.waitpinball)
						
						if not mo.waitpinball
							initSnapperPinball(mo)
						end
					end
					
				end
			else
				mo.chase = nil
				mo.jump = nil
				mo.stomp = nil
				mo.stompcd = nil
				mo.moving = nil
				
				if mo.pinball

					mo.head.flags2 = $ & ~MF2_DONTDRAW
					mo.pinball = nil
					mo.waitpinball = TICRATE*16
					mo.flags = $ & ~MF_NOCLIPTHING
					mo.head.flags = $ & ~(MF_NOBLOCKMAP|MF_NOCLIPTHING)					
				end

				mo.disablepushbox = nil
				mo.head.disablepushbox = nil
			end
			
			snapperHandlePinball(mo)
			snapperChaseHandler(mo)
			snapperHandleStompCD(mo)
			snapperHandleSlide(mo)
			snapperHandleJump(mo)
			snapperBossTeleportLimbs(mo)
			
		end,	
}