freeslot("MT_PSO_VIPERBOSS", "S_PSO_VIPERBOSS", "S_PSO_VIPERBOSSDEAD", "SPR_BVIP", "MT_PSO_VIPERAFTERLASER")
freeslot("MT_PSO_VIPERCHARGER")

mobjinfo[MT_PSO_VIPERBOSS] = {
	doomednum = 2551,
	spawnhealth = 1000,
	deathstate = S_PSO_VIPERBOSSDEAD,	-- todo: change
	spawnstate = S_PSO_VIPERBOSS,
	height = 60<<FRACBITS,
	radius = 30<<FRACBITS,
	flags = MF_ENEMY|MF_SOLID|MF_NOGRAVITY|MF_SHOOTABLE,
}

mobjinfo[MT_PSO_VIPERAFTERLASER] = {
	doomednum = -1,
	spawnhealth = 1000,
	spawnstate = S_INVISIBLE,
	height = 1,
	radius = 1,
	flags = 0,
}

mobjinfo[MT_PSO_VIPERCHARGER] = {
	doomednum = -1,
	spawnhealth = 1000,
	spawnstate = S_INVISIBLE,
	height = 60<<FRACBITS,
	radius = 30<<FRACBITS,
	flags = MF_ENEMY|MF_NOGRAVITY|MF_SHOOTABLE
}

addHook("MobjThinker", function(mo)
	
	mo.alivefor = $ and $+1 or 1
	
	mo.sprite = SPR_LASF
	mo.frame = (mo.alivefor-1) /3
	
	if mo.frame > G
		mo.frame = E + ((mo.alivefor-1)/3) %3
	end
	
	mo.frame = $|FF_FULLBRIGHT
	
	if not mo.colorized
		mo.colorized = true
		mo.color = SKINCOLOR_GOLD
	end
	mo.scale = FRACUNIT*2
	
	if leveltime%5 == 0
		local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, 256<<FRACBITS, 128<<FRACBITS, mo, 50, DMT_RANGED, 2, RT_NONE)	
		hb.continuous = 5
	end
	
end, MT_PSO_VIPERAFTERLASER)

local head_height = 1800
local head_dist = 384
local spline_step = FRACUNIT/4	-- # of segments between each point (4)

-- sets our body's waypoints up.
-- we want a "base", a head, and a middle.
local function viper_setupPoints(mo)
	mo.splinepath = {}
	
	-- point 1: the base
	local point1 = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	point1.state = S_INVISIBLE
	point1.tics = -1
	mo.splinepath[1] = point1	-- base.
	
	-- point 2: body middle
	-- in neutral stance, it's above the base.
	local point2 = P_SpawnMobj(mo.x, mo.y, mo.z + head_height*FRACUNIT, MT_THOK)
	point1.state = S_INVISIBLE
	point2.tics = -1
	mo.splinepath[2] = point2	-- middle
	
	-- point 3: head
	-- in neutral stance, it's forward compared to point2
	local p3x = point2.x + head_dist*cos(mo.angle)
	local p3y = point2.y + head_dist*sin(mo.angle)
	P_MoveOrigin(mo, p3x, p3y, mo.z + head_height*FRACUNIT)
	mo.splinepath[3] = mo
end

local function viper_makespline(mo, explosions)
	
	if not mo.splinepath return end
	local path = mo.splinepath
	
	local frac = 0
	local prev
	local num = 0
	while frac < ((#path-1) << FRACBITS)
		--print(frac..", "..(frac/FRACUNIT))
		local pos = PSO.lib.pointOnSpline(path, frac)
		
		if pos
			local seg = P_SpawnMobj(pos.x, pos.y, pos.z, MT_THOK)
			seg.tics = 1
			seg.sprite = SPR_BVIP
			seg.frame = B
			seg.scale = FRACUNIT*2
			seg.colorized = mo.colorized
			seg.color = mo.color
			
			if explosions
				
				if leveltime & 1
					seg.flags2 = $|MF2_DONTDRAW
				end
				
				local expl = P_SpawnMobj(pos.x + P_RandomRange(-128, 128)*FRACUNIT, pos.y + P_RandomRange(-128, 128)*FRACUNIT, pos.z + P_RandomRange(0, 256)*FRACUNIT, MT_THOK)
				expl.state = S_SONIC3KBOSSEXPLOSION1
				expl.scale = FRACUNIT*3
			end
		end
		
		if explosions and leveltime%5 == 0
			S_StartSound(mo, sfx_pop)
		end	

		if explosions
				
			if leveltime & 1
				mo.flags2 = $|MF2_DONTDRAW
			end
				
			local expl = P_SpawnMobj(mo.x + P_RandomRange(-128, 128)*FRACUNIT, mo.y + P_RandomRange(-128, 128)*FRACUNIT, mo.z + P_RandomRange(0, 256)*FRACUNIT, MT_THOK)
			expl.state = S_SONIC3KBOSSEXPLOSION1
			expl.scale = FRACUNIT*3
		end
		
		if frac < FRACUNIT
			frac = $ + spline_step/4
		else
			frac = $ + spline_step
		end
	end
end

-- update positions to look at the target player
-- sets mo.head_target_pos and mo.body_target_pos
local function viper_idleposition(mo, overwrite_target)
	
	local t = overwrite_target or mo.target
	if not t return end
	
	local origin = mo.splinepath[1]
	local a = R_PointToAngle2(origin.x, origin.y, t.x, t.y)
	
	local x = origin.x + head_dist*cos(a)
	local y = origin.y + head_dist*sin(a)
	local z = origin.z + head_height*FRACUNIT + 32*cos(leveltime*ANG1*4)
	
	mo.angle = a
	mo.head_target_pos = {x, y, z}
	mo.body_target_pos = {origin.x, origin.y, origin.z + head_height*FRACUNIT}
end


-- updates the position of the head and body
-- lag divides the speed, 1 makes it instant, for instance.
local function viper_update(mo, lag)
	
	lag = $ or 1
	mo.head_target_pos = $ or {mo.x, mo.y, mo.z}
	mo.body_target_pos = $ or {mo.splinepath[2].x, mo.splinepath[2].y, mo.splinepath[3].z}
	
	local nx = (mo.head_target_pos[1] - mo.x)/lag
	local ny = (mo.head_target_pos[2] - mo.y)/lag
	local nz = (mo.head_target_pos[3] - mo.z)/lag
	
	mo.momx = nx
	mo.momy = ny
	mo.momz = nz

	local nx2 = (mo.body_target_pos[1] - mo.splinepath[2].x)/lag
	local ny2 = (mo.body_target_pos[2] - mo.splinepath[2].y)/lag
	local nz2 = (mo.body_target_pos[3] - mo.splinepath[2].z)/lag
	
	mo.splinepath[2].momx = nx2
	mo.splinepath[2].momy = ny2
	mo.splinepath[2].momz = nz2
end

-- returns true if the viper is idle
local function viper_isneutral(mo)
	
	if mo.lunge
	or mo.shoot
	or mo.swing
	or mo.laser
		return false
	end
	
	return true
	
end

-- prepare a lunge, this spawns an object we'll use as a target
-- only lunge if a player isn't currently jumping for whatever reason
local function viper_preplunge(mo)
	
	local t = mo.target
	if not t or not t.valid return end
	if not t.player return end
	
	if t.player.pso.jumppad return end
	
	mo.lungetarget = P_SpawnMobj(t.x, t.y, t.z, MT_THOK)
	mo.lungetarget.state = S_INVISIBLE
	mo.lungetarget.tics = -1
	
	mo.lunge = 1
	
	S_StartSound(nil, sfx_s3k93)
end



-- do the lunge: start by leaning backwards and then rocket forward towards lungetarget
local function viper_dolunge(mo)

	local diff = PSO.lib.questDifficulty()

	local lunge_pullback = TICRATE
	local lunge_pullback_modifier = 24

	local lunge_wait = TICRATE/2

	local lunge_speedup = 2

	local lunge_endlag = TICRATE*7

	local lunge_contact_potency = 500
	local lunge_shake_potency = 400
	
	if not mo.lunge return end
	if not mo.lungetarget or not mo.lungetarget.valid
		mo.lunge = 0
		mo.lungetarget = nil
		return
	end
	
	-- face the lunge target
	if mo.lunge == 1
		mo.numlunge = 0
		viper_idleposition(mo, mo.lungetarget)
		viper_update(mo, 16)
		
		if FixedHypot(mo.momx, mo.momy) > FRACUNIT*4
			return
		end
	end
	
	-- if we made it here then we're facing the lunge target
	mo.lunge = $+1
	
	
	if mo.lunge < lunge_pullback
		
		-- for the speed to pull back with...
		-- we want to do "180 degrees" in lunge_pullback tics.
		local fake_ang = mo.lunge*180 / lunge_pullback
		local sine = sin(fake_ang*ANG1)
		local spd = lunge_pullback_modifier*sine
		
		local momx = FixedMul(spd, cos(mo.angle))
		local momy = FixedMul(spd, sin(mo.angle))
		
		-- update target position for the head
		mo.head_target_pos[1] = $ - momx
		mo.head_target_pos[2] = $ - momy
		
		-- ditto for the body
		mo.body_target_pos[1] = $ - momx
		mo.body_target_pos[2] = $ - momy
		
		mo.flags = $|MF_NOCLIPTHING
		
		mo.lungetime = nil
		
	end
	
	if mo.lunge >= lunge_pullback + lunge_wait
	and not (mo.lungetime and mo.lunge - lunge_pullback - lunge_wait > mo.lungetime)
			
		local nt = mo.lunge - lunge_pullback - lunge_wait
		local spd = nt*lunge_speedup*FRACUNIT
		
		local momx = FixedMul(spd, cos(mo.angle))
		local momy = FixedMul(spd, sin(mo.angle))
		
		--print("A")
		
		if mo.lunge == lunge_pullback + lunge_wait
			
			-- lazy hack to determine how long it takes us to clear the distance because i'm bad at maths
			local zdiff = mo.z - mo.lungetarget.z
			local time = 0
			local dist = R_PointToDist2(mo.x, mo.y, mo.lungetarget.x, mo.lungetarget.y)
			
			local cumulative = 0
			local fakespd = lunge_speedup*FRACUNIT
			
			while cumulative < dist
				time = $+1
				cumulative = $ + fakespd
				fakespd = $ + lunge_speedup*FRACUNIT
			end
			
			mo.lmomz = -zdiff / time
			mo.lungetime = time
			
			local hx = mo.x + FixedMul(mo.scale, 72*cos(mo.angle))
			local hy = mo.y + FixedMul(mo.scale, 72*sin(mo.angle))

			local hb = PSO.lib.spawnHurtBox(hx, hy, mo.z - mo.scale*16, 128<<FRACBITS, 30<<FRACBITS, mo, lunge_contact_potency, DMT_STRIKE, time, RT_BLOWDOWN)
			hb.pso.reactionangle = mo.angle
			--hb.pso.reactionradialangle = true
			hb.pso.reactionstrength = 16	
			mo.lungehitbox = hb
			
			S_StartSound(nil, sfx_s3k60)
		end
		
		-- update target position for the head
		mo.head_target_pos[1] = $ + momx
		mo.head_target_pos[2] = $ + momy
		
		if mo.lmomz ~= nil
			mo.head_target_pos[3] = $ + mo.lmomz
		end
		
		-- ditto for the body
		mo.body_target_pos[1] = $ + momx/2
		mo.body_target_pos[2] = $ + momy/2
		
		if mo.lmomz ~= nil
			mo.body_target_pos[3] = $ + mo.lmomz
		end
		
		viper_update(mo, 2)
		
		if mo.lungehitbox and mo.lungehitbox.valid
			mo.lungehitbox.momx = mo.momx
			mo.lungehitbox.momy = mo.momy
			mo.lungehitbox.momz = mo.momz
		end
		
	else
		viper_update(mo, 16)
		
		if mo.lungetime
			
			if mo.lunge == lunge_pullback + lunge_wait + mo.lungetime +1
				P_StartQuake(FRACUNIT*40, 20)
				for i = 1, 16
					local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
					d.angle = (360/16)*i*ANG1
					P_InstaThrust(d, d.angle, FRACUNIT*32)
					d.momz = FRACUNIT*8
				end
				
				S_StartSound(nil, sfx_kc40)
				
				local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z - FRACUNIT*48, 400<<FRACBITS, 192<<FRACBITS, mo, lunge_shake_potency, DMT_STRIKE, 5, RT_BLOWDOWN)
				hb.pso.reactionradialangle = true
				hb.pso.reactionstrength = 16
			
			elseif mo.lunge == lunge_pullback + lunge_wait + mo.lungetime + 5
				mo.flags = $ & ~MF_NOCLIPTHING
				
				if diff >= DIFF_HARD
				and not mo.numlunge
					mo.lunge = lunge_pullback/2
					viper_idleposition(mo, mo.lungetarget)
					mo.numlunge = 1
				end
			
			elseif mo.lunge > lunge_pullback + lunge_wait + mo.lungetime + lunge_endlag
				mo.lunge = nil
				mo.lungetime = nil
							
			end
		end
	end
end

-- prepares a projectile
local function viper_prepshoot(mo)
	
	local t = mo.target
	if not t or not t.valid return end
	if not t.player return end
	
	mo.shoot = 1
	
	S_StartSound(nil, sfx_kc67)
end

addHook("MobjThinker", function(mo)
	if mo and mo.valid and mo.pso and mo.trail
		
		local floordiff = abs(mo.z - mo.floorz)
		if floordiff < FRACUNIT*96

			local fire = P_SpawnMobj(mo.x, mo.y, mo.floorz, MT_PSO_VIPERAFTERLASER)
			fire.pso = {}
			fire.target = mo.target
			for k,v in pairs(mo.pso)
				fire.pso[k] = v
			end
					
			fire.tics = TICRATE*10
		end
	end
end, MT_ENERGYBALL)

-- do the lunge: start by leaning backwards and then rocket forward towards lungetarget
local function viper_doshoot(mo)
	
	local diff = PSO.lib.questDifficulty()

	local shoot_pullback = TICRATE
	local shoot_pullback_modifier = 24

	local shoot_wait = TICRATE/2

	local lunge_endlag = TICRATE*7

	local shoot_potency = 500
	
	if not mo.shoot return end
	if not mo.target or not mo.target.valid
		mo.shoot = 0
		return
	end
	
	-- face the lunge target
	if mo.shoot == 1
		viper_idleposition(mo)
		viper_update(mo, 16)
		
		if FixedHypot(mo.momx, mo.momy) > FRACUNIT*4
			return
		end
	end
	
	-- if we made it here then we're facing the lunge target
	mo.shoot = $+1
	
	if mo.shoot < shoot_pullback
		
		-- for the speed to pull back with...
		-- we want to do "180 degrees" in lunge_pullback tics.
		local fake_ang = mo.shoot*180 / shoot_pullback
		local sine = sin(fake_ang*ANG1)
		local spd = shoot_pullback_modifier*sine
		
		local momx = FixedMul(spd, cos(mo.angle))
		local momy = FixedMul(spd, sin(mo.angle))
		
		-- update target position for the head
		mo.head_target_pos[1] = $ - momx
		mo.head_target_pos[2] = $ - momy
		mo.head_target_pos[3] = $ - FRACUNIT*3
		
		-- ditto for the body
		mo.body_target_pos[1] = $ - momx
		mo.body_target_pos[2] = $ - momy
		mo.head_target_pos[3] = $ - FRACUNIT*2
		
		--mo.flags = $|MF_NOCLIPTHING
		
		for i = 1, 2
			local t = P_SpawnMobj(mo.x + P_RandomRange(-256, 256)*FRACUNIT, mo.y + P_RandomRange(-256, 256)*FRACUNIT, mo.z + P_RandomRange(-256, 256)*FRACUNIT, MT_THOK)
			t.color = SKINCOLOR_YELLOW
			t.frame = FF_FULLBRIGHT
			t.tics = 6
			t.scale = FRACUNIT*2
			t.destscale = 1
			
			t.momx = (mo.x - t.x)/6
			t.momy = (mo.y - t.y)/6
			t.momz = (mo.z - t.z)/6
		end
		
	end
	
	if mo.shoot == shoot_pullback + shoot_wait
		
		local msl = PSO.lib.spawnMissile(mo, MT_ENERGYBALL, mo.angle, 30*FRACUNIT, shoot_potency, DMT_RANGED, 0, 0, 0, RT_BLOWDOWN)
		PSO.lib.aimMissile(msl, mo.target)
		msl.color = SKINCOLOR_YELLOW
		msl.colorized = true
		msl.scale = FRACUNIT*2
		msl.fuse = TICRATE*2
		msl.trail = diff >= DIFF_HARD
		
		S_StartSound(nil, sfx_s3k54)
	end
	
	if mo.shoot == shoot_pullback + shoot_wait + 10
		mo.shoot = nil
	end
	
	viper_update(mo, 16)
end


-- prepares a swing
local swing_dist = 1184
local swing_startlag = TICRATE/4
local swing_duration = TICRATE*2
local swing_endlag = TICRATE*6
local swing_potency = 550

local function viper_prepswing(mo)
	
	local t = mo.target
	if not t or not t.valid return end
	if not t.player return end
	
	mo.swing = 1
	mo.swingdir = P_RandomRange(0, 1) and 1 or -1
	
	-- set the position
	local ang = mo.angle + ANG1*90 * mo.swingdir
	local center = mo.splinepath[1]
	local tx = swing_dist*cos(ang)
	local ty = swing_dist*sin(ang)
	
	mo.head_target_pos[1] = center.x + tx
	mo.head_target_pos[2] = center.y + ty
	mo.head_target_pos[3] = mo.z - FRACUNIT*215

	mo.body_target_pos[1] = center.x + tx/4
	mo.body_target_pos[2] = center.y + ty/4
	mo.body_target_pos[3] = mo.z - FRACUNIT*300
	
	mo.angle = R_PointToAngle2(mo.x, mo.y, mo.x+tx, mo.y+ty)
	
	mo.flags = $ | MF_NOCLIPTHING
	
	S_StartSound(nil, sfx_s3k8e)
end

local function viper_doswing(mo)
	
	if not mo.swing return end
	if not mo.target or not mo.target.valid
		mo.swing = 0

		if mo.swinghitboxes
			for i = 1, #mo.swinghitboxes
				if mo.swinghitboxes[i].valid
					P_RemoveMobj(mo.swinghitboxes[i])
					return
				end
			end
		end
		
		return
	end
	
	-- face the lunge target
	if mo.swing == 1
		--viper_idleposition(mo)
		viper_update(mo, 16)
		
		if FixedHypot(mo.momx, mo.momy) > FRACUNIT*4
			return
		end
	end
	
	mo.swing = $+1
	
	if mo.swing == 2
		-- assess the current positions to know how to swing...
		local center = mo.splinepath[1]
		mo.swing_headdist = R_PointToDist2(center.x, center.y, mo.x, mo.y)
		mo.swing_bodydist = R_PointToDist2(center.x, center.y, mo.splinepath[2].x, mo.splinepath[2].y)
		
		-- spawn a bunch of hitboxes and link them
		mo.swinghitboxes = {}
		for i = 1, 16
			local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z - FRACUNIT*48, 192<<FRACBITS, 192<<FRACBITS, mo, swing_potency, DMT_STRIKE, swing_startlag + swing_duration*3/4, RT_BLOWDOWN)
			hb.angle = mo.angle - ANG1*90 * mo.swingdir
			hb.pso.reactionstrength = 16	
			mo.swinghitboxes[#mo.swinghitboxes+1] = hb
		end
		
		-- this ensures the player can't be hit by more than 1 of those hitboxes
		for i = 1, 16
			mo.swinghitboxes[i].pso.links = mo.swinghitboxes
		end
	end
	
	
	if mo.swing > 2 + swing_startlag
	and mo.swing < 2 + swing_startlag + swing_duration
		
		if leveltime%10 == 0
			S_StartSound(nil, sfx_s3k6f)
		end
		
		local t = mo.swing - 2 - swing_startlag
		local spd = FixedAngle(4*sin( (t*180/swing_duration)*ANG1 ))
		
		mo.swing_headdist = $ - FRACUNIT
		mo.swing_bodydist = $ - FRACUNIT
		
		mo.angle = $ - spd * mo.swingdir
		
		-- update the targets
		local center = mo.splinepath[1]
		local head_x = center.x + FixedMul(mo.swing_headdist, cos(mo.angle))
		local head_y = center.y + FixedMul(mo.swing_headdist, sin(mo.angle))
		
		local body_x = center.x + FixedMul(mo.swing_bodydist, cos(mo.angle))
		local body_y = center.y + FixedMul(mo.swing_bodydist, sin(mo.angle))
		
		mo.head_target_pos[1] = head_x
		mo.head_target_pos[2] = head_y
		
		mo.body_target_pos[1] = body_x
		mo.body_target_pos[2] = body_y
		
		-- move the hitboxes....
		local hitbox_dist = 0
		for i = 1, 16
			
			local hb = mo.swinghitboxes[i]
			
			if not hb or not hb.valid
				continue
			end
			
			local tx = mo.x - hitbox_dist*cos(mo.angle)
			local ty = mo.y - hitbox_dist*sin(mo.angle)
			local tz = mo.z - FRACUNIT*96
			
			P_MoveOrigin(hb, tx, ty, tz)
			hb.angle = mo.angle - ANG1*90 * mo.swingdir
			
			hitbox_dist = 96
		end
	end
	
	if mo.swing == 2 + swing_startlag + swing_duration*3/4
		mo.flags = $ & ~MF_NOCLIPTHING
	end	
	
	if mo.swing > 2 + swing_startlag + swing_duration + swing_endlag
		mo.swing = nil
		return
	end
	
	viper_update(mo, 2)
end

-- prepares the laser swipe
local laser_dist = 630
local laser_startlag = TICRATE/4
local laser_duration = TICRATE*19
local laser_endlag = TICRATE*6
local laser_potency = 15000

local function viper_preplaser(mo)
	
	local t = mo.target
	if not t or not t.valid return end
	if not t.player return end
	
	mo.laser = 1
	mo.laserdir = P_RandomRange(0, 1) and 1 or -1
	
	-- set the position
	local ang = mo.angle + ANG1*90 * mo.laserdir
	local center = mo.splinepath[1]
	local tx = laser_dist*cos(ang)
	local ty = laser_dist*sin(ang)
	
	mo.head_target_pos[1] = center.x + tx
	mo.head_target_pos[2] = center.y + ty
	mo.head_target_pos[3] = mo.z - FRACUNIT*215

	mo.body_target_pos[1] = center.x - tx/2
	mo.body_target_pos[2] = center.y - ty/2
	mo.body_target_pos[3] = mo.z - FRACUNIT*300
	
	mo.angle = R_PointToAngle2(mo.x, mo.y, mo.x+tx, mo.y+ty)
	
	S_StartSound(nil, sfx_s3k53)
end

local function viper_dolaser(mo)
	
	if not mo.laser return end
	if not mo.target or not mo.target.valid
		mo.laser = 0
		
		if mo.laserhitboxes
			for i = 1, #mo.laserhitboxes
				if mo.laserhitboxes[i].valid
					P_RemoveMobj(mo.laserhitboxes[i])
				end
			end
		end
		return
	end
	
	-- face the lunge target
	if mo.laser == 1
		--viper_idleposition(mo)
		viper_update(mo, 16)

		for i = 1, 2
			local t = P_SpawnMobj(mo.x + P_RandomRange(-256, 256)*FRACUNIT, mo.y + P_RandomRange(-256, 256)*FRACUNIT, mo.z + P_RandomRange(-256, 256)*FRACUNIT, MT_THOK)
			t.color = SKINCOLOR_YELLOW
			t.frame = FF_FULLBRIGHT
			t.tics = 6
			t.scale = FRACUNIT*2
			t.destscale = 1
			
			t.momx = (mo.x - t.x)/6
			t.momy = (mo.y - t.y)/6
			t.momz = (mo.z - t.z)/6
		end
		
		if FixedHypot(mo.momx, mo.momy) > FRACUNIT*4
			return
		end
	end
	
	mo.laser = $+1
	
	if mo.laser == 2
		
		S_StartSound(nil, sfx_s3k54)
		
		-- assess the current positions to know how to swing...
		local center = mo.splinepath[1]
		mo.laser_headdist = R_PointToDist2(center.x, center.y, mo.x, mo.y)
		mo.laser_bodydist = R_PointToDist2(center.x, center.y, mo.splinepath[2].x, mo.splinepath[2].y)
		
		-- spawn a bunch of hitboxes and link them
		mo.laserhitboxes = {}
		for i = 1, 24
			local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z - FRACUNIT*48, 96<<FRACBITS, 96<<FRACBITS, mo, laser_potency, DMT_STRIKE, laser_startlag + laser_duration*95/100, RT_BLOWDOWN)
			hb.angle = mo.angle - ANG1*90 * mo.laserdir
			hb.pso.reactionstrength = 16	
			hb.scale = FRACUNIT*2
			mo.laserhitboxes[#mo.laserhitboxes+1] = hb
		end
		
		-- this ensures the player can't be hit by more than 1 of those hitboxes
		for i = 1, 24
			mo.laserhitboxes[i].pso.links = mo.laserhitboxes
		end
	end
	
	
	if mo.laser > 2 + laser_startlag
	and mo.laser < 2 + laser_startlag + laser_duration

		if leveltime%10 == 0
			S_StartSound(nil, sfx_s3k6f)
		end
		
		local t = mo.laser - 2 - laser_startlag
		local spd = min(ANG1, FixedAngle(4*sin( (t*180/laser_duration)*ANG1 )))
		
		mo.angle = $ - spd * mo.laserdir*3/4
		
		-- update the targets
		local center = mo.splinepath[1]
		local head_x = center.x + FixedMul(mo.laser_headdist, cos(mo.angle))
		local head_y = center.y + FixedMul(mo.laser_headdist, sin(mo.angle))
		
		local body_x = center.x - FixedMul(mo.laser_bodydist, cos(mo.angle))
		local body_y = center.y - FixedMul(mo.laser_bodydist, sin(mo.angle))
		
		mo.head_target_pos[1] = head_x
		mo.head_target_pos[2] = head_y
		
		mo.body_target_pos[1] = body_x
		mo.body_target_pos[2] = body_y
		
		-- move the hitboxes....
		local hitbox_dist = 128 + leveltime%3 * (64/3)
		for i = 1, 24
			
			local hb = mo.laserhitboxes[i]
			
			if not hb or not hb.valid
				continue
			end
			
			local tx = mo.x + hitbox_dist*cos(mo.angle)
			local ty = mo.y + hitbox_dist*sin(mo.angle)
			local tz = mo.z
			
			P_MoveOrigin(hb, tx, ty, tz)
			hb.angle = mo.angle - ANG1*90 * mo.laserdir
			
			hb.sprite = SPR_ENRG
			hb.frame = A|FF_FULLBRIGHT
			hb.colorized = true
			hb.color = SKINCOLOR_YELLOW
			
			if hb.tics < TICRATE*2
				hb.scale = max(1, hb.tics * (FRACUNIT*2) / (TICRATE*2))
				if leveltime & 1
					hb.sprite = SPR_NULL
				end
			end
			
			
			
			if i & 1
				local fz = hb.floorz
				local s = hb.subsector.sector
				
				if s.f_slope
					fz = P_GetZAt(s.f_slope, hb.x, hb.y)
				end
				
				
				if abs(hb.z - fz) < FRACUNIT*64
				and leveltime%10 == 0
				and R_PointToDist2(mo.x, mo.y, hb.x, hb.y) < 1184*FRACUNIT
					local fire = P_SpawnMobj(hb.x, hb.y, hb.floorz, MT_PSO_VIPERAFTERLASER)
					fire.pso = {}
					fire.target = hb.target
					for k,v in pairs(hb.pso)
						fire.pso[k] = v
					end
					
					fire.tics = TICRATE*10
				end
			end
			
			hitbox_dist = $+64
		end
	end
	
	if mo.laser > 2 + laser_startlag + laser_duration + laser_endlag
		mo.laser = nil
		return
	end
	
	viper_update(mo, 2)
end

local zondecharger_dist = 940
local function viper_spawnzondepods(mo, amount)
	
	local initangle = P_RandomRange(1, 3)
	mo.zondechargers = {}
	
	local a = mo.angle + initangle*ANG1*90
	for i = 1, amount
		
		local x = mo.splinepath[1].x + zondecharger_dist*cos(a)
		local y = mo.splinepath[1].y + zondecharger_dist*sin(a)
		
		local z = PSO.lib.spawnEnemyFromEnemy(mo, x, y, mo.splinepath[1].z + FRACUNIT*1600, MT_PSO_VIPERCHARGER)
		
		mo.zondechargers[#mo.zondechargers+1] = z
		z.target = mo.splinepath[1]
		
		a = $ + ANG1*90
	end
	
	mo.zondecharge = TICRATE*16 * (amount) + TICRATE*3*(amount-1)
end

local function viper_handlezondepods(mo)	
	if mo.zondechargers
		local i = #mo.zondechargers
		
		while i
			local z = mo.zondechargers[i]
			if not z or not z.valid
				table.remove(mo.zondechargers, i)
			end
			
			i = $-1
		end
		
		if not #mo.zondechargers
			mo.zondechargers = nil
			mo.zonded = TICRATE*3
			mo.shoot = nil
			mo.laser = nil

			if mo.laserhitboxes
				for i = 1, #mo.laserhitboxes
					if mo.laserhitboxes[i].valid
						P_RemoveMobj(mo.laserhitboxes[i])
					end
				end
			end
			mo.laserhitboxes = nil
			
			S_StartSound(nil, sfx_s3k79)
		end

		if mo.zondecharge
			mo.zondecharge = $-1
			
			if not mo.zondecharge
				
				local potency = 650 * #mo.zondechargers
				
				for p in players.iterate do
					
					local pmo = p.mo
					if p.mo and p.mo.valid and R_PointToDist2(mo.x, mo.y, pmo.x, pmo.y) < FRACUNIT*1280
						
						-- attack them!
						local zonde = P_SpawnMobj(pmo.x, pmo.y, pmo.z, MT_THOK)
						zonde.state = S_ZONDE
						zonde.color = SKINCOLOR_GOLD
						local hb = PSO.lib.spawnHurtBox(zonde.x, zonde.y, zonde.z, 32<<FRACBITS, 180<<FRACBITS, mo, potency, DMT_TECHNIQUE, 4)
						hb.pso.element = ELEM_ELEC					
					end
				end
				
				for i = 1, #mo.zondechargers
					P_RemoveMobj(mo.zondechargers[i])
				end
				mo.zondechargers = nil
			end
		end
	end	
end

-- we shocked ourselves lol
local function viper_selfzonde(mo)
	
	if mo.zonded
		mo.zonded = $-1
		
		if leveltime%4 == 0

			local zonde = P_SpawnMobj(mo.x +P_RandomRange(-80, 80)*FRACUNIT, mo.y +P_RandomRange(-80, 80)*FRACUNIT, mo.z - FRACUNIT*512, MT_THOK)
			zonde.state = S_ZONDE
			zonde.color = SKINCOLOR_GOLD
			zonde.scale = FRACUNIT*3
			zonde.spritetscale = FRACUNIT*12
		end
		
		if leveltime%4 < 2
			mo.color = SKINCOLOR_WHITE
			mo.colorized = true
		else
			mo.colorized = false
		end
		
		if mo.zonded == 0
			S_StartSound(nil, sfx_bkpoof)

			local damage = min(mo.pso.hp-1, mo.pso.maxhp/8)
			PSO.lib.damageObject(mo, mo, damage)
			
			mo.colorized = false
		end
	end
end

-- dying logic
function A_ViperDie(mo)
	viper_makespline(mo, true)
	viper_update(mo, 1)
	
	mo.deathtimer = $ and $+1 or 1
	
	mo.flags = $|MF_NOCLIP
	if mo.deathtimer > TICRATE
		P_RemoveMobj(mo)
	end
	--print("A")
end

states[S_PSO_VIPERBOSSDEAD] = {SPR_BVIP, A, 1, A_ViperDie, 0, 0, S_PSO_VIPERBOSSDEAD} 

PSO.info.enemies[MT_PSO_VIPERCHARGER] = {
		
		name = "Zonde Charger",
		sprite = SPR_GWLR,
		frame = A,
		painframe = A,
		
		bhp = 75,
		satk = 45,
		ratk = 45,
		tatk = 1,
		def = 20,
		exp = 0,
		weight = 4,
		
		drops = {},
		
		weakness = ELEM_EARTH,
		
		reset =	function(mo)
			mo.actiontimer = 0	-- reset action if we get staggered etc.
		end,
		
		thinker = function(mo)
			
			if not mo.target return end
			
			if not mo.cusval
				mo.dist = R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y)
				mo.angle = R_PointToAngle2(mo.target.x, mo.target.y, mo.x, mo.y)
				mo.basez = mo.z
				
				mo.cusval = 1 
			end
			
			mo.sprite = SPR_GWLR
			mo.frame = A
			mo.colorized = true
			mo.color = SKINCOLOR_GOLD
			
			mo.z = mo.basez + 8*cos(leveltime*ANG1*2)
			
			if not mo.stay
				
				mo.angle = $ + ANG1/4 --* (mo.rotdir or 1)
				
				local x = mo.target.x + FixedMul(mo.dist, cos(mo.angle))
				local y = mo.target.y + FixedMul(mo.dist, sin(mo.angle))
				P_MoveOrigin(mo, x, y, mo.z)
			end
			
		end,
}

local function viper_roar(mo)
	
	if not mo.introd
		mo.introtimer = $ and $+1 or 1
		
		for p in players.iterate do
		
			local pmo = p.mo
			if pmo and pmo.valid and R_PointToDist2(mo.x, mo.y, pmo.x, pmo.y) < FRACUNIT*1536
				if mo.introtimer == TICRATE*3
					p.powers[pw_nocontrol] = 0
				else
					p.powers[pw_nocontrol] = 32767
				end
			end
		end
		
		-- put em in the hole
		if mo.introtimer == 1

			local origin = mo.splinepath[1]
			mo.head_target_pos = {origin.x, origin.y, origin.z}
			mo.body_target_pos = {origin.x, origin.y, origin.z}
			
			P_MoveOrigin(mo.splinepath[2], origin.x, origin.y, origin.z)
			P_MoveOrigin(mo.splinepath[3], origin.x, origin.y, origin.z)
		end
		
		if mo.introtimer == TICRATE/2
			P_StartQuake(FRACUNIT*30, TICRATE*3)
			S_StartSound(nil, sfx_sroar1)
		end
		
		if mo.introtimer >= TICRATE
			viper_idleposition(mo)
			viper_update(mo, 16)			
		end
		
		if mo.introtimer == TICRATE + TICRATE/3
			for i = 1, 32
				
				local d = P_SpawnMobj(mo.splinepath[2].x, mo.splinepath[2].y, mo.splinepath[2].z, MT_SPINDUST)
				d.colorized = true
				d.color = SKINCOLOR_BLACK
				
				d.flags = $|MF_NOCLIP|MF_NOCLIPHEIGHT
				d.momx = P_RandomRange(-32, 32)*FRACUNIT
				d.momy = P_RandomRange(-32, 32)*FRACUNIT
				d.momz = P_RandomRange(32, 64)*FRACUNIT
				d.scale = FRACUNIT*10
				d.tics = $ + P_RandomRange(0, 20)
			end
		end
		
		if mo.introtimer == TICRATE*3
			mo.introd = true
		end

		mo.sprite = SPR_BVIP
		mo.frame = A
		mo.scale = FRACUNIT*2
		
	end	
end

PSO.info.enemies[MT_PSO_VIPERBOSS] = {
		
		name = "Viper",
		sprite = SPR_BVIP,
		frame = A,
		painframe = A,
		
		boss = true,		-- enemy is a major boss!
		bossmusic = "BOSS2",
		bossrushmusic = "BOSSR2",
		
		bhp = 3500,
		satk = 35,
		ratk = 35,
		tatk = 35,
		def = 30,
		exp = 300,
		weight = 4,
		

		drops = {
			{"breaker", 1, 29, 2, 100},
			{"pine", 1, 29, 1, 100},
			{"claymore", 1, 29, 1, 100},
			
			{"doublegladius", 1, 29, 2, 100},
			{"doubliss", 1, 29, 1, 100},
			{"doublecutter", 1, 29, 1, 100},	
			
			{"laser", 1, 29, 2, 100},
			{"shooter", 1, 29, 1, 100},
			{"burster", 1, 29, 1, 100},				
			
			{"canon", 1, 29, 2, 100},
			{"shot", 1, 29, 1, 100},
			{"concentrator", 1, 29, 1, 100},	
			
			{"ulrod", 1, 29, 2, 100},
			{"crest", 1, 29, 1, 100},
			{"crescent", 1, 29, 1, 100},
				
			{"slash", 1, 29, 3, 50},
			{"claw", 1, 29, 1, 100},
			{"cutter", 1, 29, 1, 100},
			
			{"autogun", 1, 29, 3, 50},
			{"beamgun", 1, 29, 1, 100},
			{"peacekeeper", 1, 29, 1, 100},
			
			{"batnara", 1, 29, 3, 50},			
			{"hammer", 1, 29, 1, 100},
			{"scepter", 1, 29, 1, 100},
			
			{"sneg frame", 1, 29, 1, 10},
			
			{"unit_hp_2", 1, 29, 2, 50},
			{"unit_pp_2", 1, 29, 2, 50},
			{"unit_satk_2", 1, 29, 2, 50},
			{"unit_tatk_2", 1, 29, 2, 50},
			{"unit_ratk_2", 1, 29, 2, 50},

			{"monogrinder", 1, 29, 5, 50},
			{"digrinder", 1, 29, 5, 50},

			-- lv30+
			{"orakl sword", 30, 50, 1, 400},
			{"orakl doublesaber", 30, 50, 1, 400},
			{"orakl soaringblade", 30, 50, 1, 400},
			{"orakl rifle", 30, 50, 1, 400},
			{"orakl shot", 30, 50, 1, 400},
			{"orakl orphan", 30, 50, 1, 400},
			{"orakl rod", 30, 50, 1, 400},
			{"orakl wand", 30, 50, 1, 400},
			{"sneg zapper", 30, 50, 1, 50},
			{"fehler frame", 30, 50, 5, 50},
			{"digrinder", 30, 50, 5, 50},
			{"trigrinder", 30, 50, 4, 50},

			{"unit_hpsatk_2", 30, 50, 1, 20},
			{"unit_hpratk_2", 30, 50, 1, 20},
			{"unit_hptatk_2", 30, 50, 1, 20},
			{"unit_rainbow_2", 30, 50, 1, 20},
				
			{10000, nil, nil, 1, 9},
		},
		
		weakness = ELEM_EARTH,
		
		reset =	function(mo)
			mo.actiontimer = 0	-- reset action if we get staggered etc.
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			
			PSO.lib.enemyManageAggro(mo)
			
			if not mo.splinepath
				viper_setupPoints(mo)
				mo.attackcount = 0
				mo.alivetimer = 0
			end
			
			viper_makespline(mo)
			
			if not mo.introd
				viper_roar(mo)
				return
			end
			
			if mo.zonded
				viper_selfzonde(mo)
				viper_update(mo, 1)
				return
			end
			
			if not mo.target or not mo.target.valid
				viper_idleposition(mo)
				viper_update(mo, 16)
				mo.lunge = 0
				mo.shoot = 0
				mo.shoot = 0
				mo.laser = 0

				if mo.laserhitboxes
					for i = 1, #mo.laserhitboxes
						if mo.laserhitboxes[i].valid
							P_RemoveMobj(mo.laserhitboxes[i])
						end
					end
				end
				mo.laserhitboxes = nil				
				return
			end
			
			if mo.target and mo.target.valid
				
				if viper_isneutral(mo)
				
					viper_idleposition(mo)
					viper_update(mo, 16)
					
					mo.waittimer = $ and $+1 or 1
					--print(mo.waittimer)
					
					if mo.waittimer > TICRATE*3
						
						-- debug shit lol
						--viper_preplunge(mo)
						--viper_prepshoot(mo)
						--viper_prepswing(mo)
						--viper_preplaser(mo)
						/*if not mo.zondechargers
							viper_spawnzondepods(mo, 3)
						else
							viper_prepshoot(mo)
						end*/

						mo.waittimer = 0
						
						if mo.zondechargers
							
							if mo.secondpod
							and (not mo.secondpodlaser)
								viper_preplaser(mo)
								mo.secondpodlaser = true	-- only do this once
							else
								viper_prepshoot(mo)
							end
							
							mo.waittimer = TICRATE*3/2
						else
							
							if mo.pso.hp < (mo.pso.maxhp*2/3)
							and not mo.firstlaser
								viper_preplaser(mo)
								mo.firstlaser = true
								--print("do first laser sweep")
							
							elseif mo.pso.hp < mo.pso.maxhp/2
							and not mo.firstpod
								viper_spawnzondepods(mo, diff >= DIFF_HARD and 3 or 2)
								mo.firstpod = true
								--print("spawn first pods")

							elseif mo.pso.hp < mo.pso.maxhp/4
							and not mo.secondpod
								viper_spawnzondepods(mo, diff >= DIFF_HARD and 4 or 3)
								mo.secondpod = true
								--print("spawn second pods")
								
							else
								
								-- do a counterable attack
								if not (mo.attackcount & 1)
									
									if mo.pso.hp < mo.pso.maxhp*3/4
									and not P_RandomRange(0, 2)
										viper_prepswing(mo)	-- do a swing
									else									
										viper_preplunge(mo)	-- do a lunge
									end
								else	-- uncounterable
									viper_prepshoot(mo)
								end
							end
						end
						
						mo.attackcount = $+1
						
					end
				else
					
					if mo.lunge
						--print("lunging")
						viper_dolunge(mo)
					end
					
					if mo.shoot
						viper_doshoot(mo)
					end
					
					if mo.swing
						viper_doswing(mo)
					end
					
					if mo.laser
						viper_dolaser(mo)
					end
				end
				
				viper_handlezondepods(mo)
			end
			
			mo.sprite = SPR_BVIP
			mo.frame = A
			mo.scale = FRACUNIT*2
		end,
		
}