-- Simple Authentication script


-- Server key generation
local KEYLENGTH = 12
local chars = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, "a", "b", "c", "d", "e", "f"}

local SERVERKEY
local needsavekey

local login_asks = {}		-- stores keys of clients wanting to log in
local deletion_asks = {}	-- stores keys of clients asking for their file to be deleted
local key_asks = {}			-- stores booleans when a client asks to receive a new login key
local logged_keys = {}		-- stores keys of clients currently logged in

local MYKEY	-- my login key

addHook("NetVars", function(net)
	SERVERKEY = net(SERVERKEY)
end)

--PSO.info.serverKeyFile

local function server_keyGen()
	local key = "SN"
	
	for i = 1, KEYLENGTH
		key = $..chars[P_RandomRange(1, 16)]
	end
	
	return key
end

-- saves the server key
local function server_saveKey(key)
	-- we don't need this to be synched...
	local f = io.openlocal(PSO.info.serverKeyFile, "w")
	if f
		f:write(key)
		f:close()
		return true
	else
		return false
	end
end

local function server_readKey(f)
	if f
		local key = f:read()
		--f:close()
		
		SERVERKEY = key
		CONS_Printf(server, "Server key read: "..SERVERKEY)
	else
		-- no file key
		CONS_Printf(server, "Server key not found, generating...")
		SERVERKEY = server_keyGen()
		needsavekey = true
		--server_saveKey(SERVERKEY)
	end
end

local function server_readData(f)
	if f
		local dta = f:read()
		
		local run = pcall(function()
			dta = json.decode(dta)
		end)
		
		if not run
			print("\x82".."WARNING: ".."Server data may be corrupted and cannot be read. Quest unlock progress will not be loaded.")
		else
			dta = PSO.lib.sparseIntTable(dta)
			server.psounlocks = dta
		end
	end
	-- otherwise, nothing to read or do
end

PSO.lib.server_saveData = function()
	local f = io.openlocal(netgame and PSO.info.serverDataFile or PSO.info.soloData, "w")
	
	assert(f, "Cannot open data file, check your permissions to luafiles and available disk space")
	
	local dta = PSO.lib.sparseStrTable(server.psounlocks)
	f:write(json.encode(dta))
	f:close()
end

PSO.lib.clientNukeData = function(p)	
	local f = io.openlocal(netgame and PSO.info.serverDataFile or PSO.info.soloData, "w")
	assert(f, "Couldn't create Savefile, please check your permissions to /luafiles and available disk space")
	f:write("[]")
	f:close()
	p.pso.nukesave = true
	--COM_BufInsertText(p, "exitgame") (this crashes the game)
end

-- check if we have a key, return it if possible
local function server_checkKey()
	
	if server
		if not server.init_key
			SERVERKEY = nil
			MYKEY = nil
			logged_keys = {}
			login_asks = {}
			key_asks = {}
			needsavekey = nil
		end
		
		if server
			if not server.init_key
			and netgame
			
				io.open(PSO.info.serverKeyFile, "r", server_readKey)
				server.init_key = true
			elseif not server.init_data
				
				io.open(netgame and PSO.info.serverDataFile or PSO.info.soloData, "r", server_readData)
				server.init_data = true
				
			end
		end
		
		if needsavekey
			server_saveKey(SERVERKEY)
			needsavekey = nil
		end
	end
end

PSO.lib.registerFrameFunc(server_checkKey)

-- client login stuff
-- warning: scary clientsided stuff :)

COM_AddCommand("__&client_asklogin", function(p, arg)
	-- ask the server to login. This doesn't work if the player is logged in already.
	if p.psologin return end
	
	if isserver
	and not login_asks[#p+1]
		login_asks[#p+1] = arg
	end
end)

-- player asks for a key for a first time login
COM_AddCommand("__&client_askkey", function(p)
	if p.psologin return end
	
	if isserver
	and not key_asks[#p+1]
		key_asks[#p+1] = true
	end
end)

COM_AddCommand("__&client_askdelete", function(p)
	
	if isserver
	and logged_keys[#p+1]	
		deletion_asks[#p+1] = true
	end
end)

local function client_updateLocalKeys(p, newserverkey, newclientkey)
	if p == consoleplayer
		-- do we have existing keys?
		local keys = {}
		local f = io.openlocal(PSO.info.clientKeyFile, "r")
		
		if f
			local run = pcall(function()
							keys = json.decode(f:read())
						end)
			
			if not run
				print("\x82".."WARNING: ".."\x80".."Failed to read local keys, the file may be corrupted.")
				keys = {}
			end
			
			f:close()
		end
		
		keys[newserverkey] = newclientkey
		
		local f = io.openlocal(PSO.info.clientKeyFile, "w")
		if f
			f:write(json.encode(keys))
			f:close()
		else
			error("Couldn't open keys file, are the permissions to luafiles OK? Are you out of disk space?")
		end
		
	end
end

local function client_checkLocalKey(p)
	
	print("Checking local key...")
	local f = io.openlocal(PSO.info.clientKeyFile, "r")
	local keys = {}
	
	if f
		--print("file OK")
		local run = pcall(function()
				keys = json.decode(f:read())
		end)
			
		if not run
			print("\x82".."WARNING: ".."\x80".."Failed to read local keys, the file may be corrupted.")
			keys = {}
		end
		f:close()
		
		/*print(SERVERKEY)
		for k,v in pairs(keys)
			print(k..":"..v)
		end*/
		
		return keys[SERVERKEY]
	end	
	return nil
end

-- singleplayer loading
local function client_localLoad(p)
	-- this is local so io.openlocal is fine
	local f = io.openlocal(PSO.info.soloSave, "r")
	
	-- if the file doesn't exist then we have nothing to load
	if f
		local dta = f:read()
		
		local run = pcall(function()
			dta = json.decode(dta)
		end)		
		
		if not run
			print("\x82".."WARNING: ".."\x80".."Savefile data cannot be read, file may be corrupted.")
			PSO.lib.openMenu(p, "type_set")
		else
			PSO.lib.deserializePlayer(p, dta)
		end
		
		if not p.pso.type_initialized
			PSO.lib.openMenu(p, "type_set")
		end
		
		f:close()
	else
		-- no save? Have the player set their thing
		PSO.lib.openMenu(p, "type_set")
	end
	
	p.psologin = true
end

PSO.lib.localSave = function(p)
	if netgame return end	-- no.
	if not p.pso return end	-- player is not ready.
	
	local dta = PSO.lib.serializePlayer(p)

	
	local f = io.openlocal(PSO.info.soloSave, "w")
	assert(f, "Couldn't create Savefile, please check your permissions to /luafiles and available disk space")
		
	f:write(dta)
	f:close()
end

-- deletes the local save
-- this simply opens and closes the file in w mode which wipes it.
PSO.lib.clientNukeSave = function(p)
	if netgame return end	-- no
	
	local f = io.openlocal(PSO.info.soloSave, "w")
	
	assert(f, "Couldn't create Savefile, please check your permissions to /luafiles and available disk space")
	f:write("[]")
	f:close()
	--COM_BufInsertText(p, "exitgame") (this crashes the game)
end

local can_ask_login = true
local nologin = 0
local LOGIN_DELAY = TICRATE*5

addHook("ThinkFrame", do
	-- hack that is only true on the title screen
	if not server
	and not can_ask_login
		--print("Login ask reset")
		can_ask_login = true
		nologin = 0
	end
end)

local function client_getLogin(p)
	-- ask things in cycles
	
	if not SERVERKEY and netgame return end	-- wait for this to be read or generated
	
	--if not p.asklogin
	if p == consoleplayer
	and can_ask_login
		
		if netgame
			local loc_key = client_checkLocalKey(p)
			if loc_key
				CONS_Printf(consoleplayer, "Requesting login from server...")
				COM_BufInsertText(p, "__&client_asklogin "..loc_key)
			else
				CONS_Printf(consoleplayer, "Asking server for new login key...")
				COM_BufInsertText(p, "__&client_askkey")
			end
		else
			CONS_Printf(consoleplayer, "Loading local save file...")
			client_localLoad(p)
		end
		
		--p.asklogin = true
		can_ask_login = false
		nologin = 0
	
	elseif not p.psologin
		if p == consoleplayer
			nologin = $+1
		end
		
		-- start over...
		if nologin > LOGIN_DELAY
			CONS_Printf(consoleplayer, "Login request appears not to have been received, trying again...")
			can_ask_login = true
			nologin = 0
		end
	end
end

PSO.lib.registerPlayerFunc(client_getLogin)

-- attempts to find the given player ID. Returns true if a save file exists for that ID, false if not.
local function server_checkPlayerID(id)
	local f = io.openlocal(PSO.info.serverSaves.."/"..id..".sav2", "r")
	
	if f
		f:close()
		return true
	else
		return false
	end
end

-- creates a player ID and an associated file
-- the ID must then be sent to the player
local function server_makePlayerKey()
	local key = "PN"
	
	for i = 1, KEYLENGTH
		key = $..chars[P_RandomRange(1, 16)]
	end
	
	return key
end

local function server_createPlayerID()
	local key = server_makePlayerKey()
	
	-- if somehow our key already exists keep making em until we make one that dosn't exist
	while server_checkPlayerID(key)
		key = server_makePlayerKey()
	end
	
	local f = io.openlocal(PSO.info.serverSaves..key..".sav2", "w")
	f:close()
	
	return key
end

local function server_savePlayer(pnum)
	if isserver
	and logged_keys[pnum]
		
		local p = players[pnum-1]
		assert(p and p.valid, "Attempting to save player "..pnum.." which does not exist")
		
		assert(p.pso, "Attempting to save player "..pnum.." which is not yet setup.")
		
		local dta = PSO.lib.serializePlayer(p)
		local f = io.openlocal(PSO.info.serverSaves..logged_keys[pnum]..".sav2", "w")
		if f
			f:write(dta)
			f:close()
		else
			error("File for player "..pnum.." does not exist, aborting save.")
		end
	end
end

COM_AddCommand("__&server_sendkey", function(host, key, pnum)
	if host ~= server return end
	local p = players[tonumber(pnum)-1]
	
	if p == consoleplayer
		MYKEY = key
		client_updateLocalKeys(p, SERVERKEY, MYKEY)
	end
	
	logged_keys[#p+1] = key
	
	server_savePlayer(#p +1)
	p.psologin = true
	
	-- we just sent the key so this is a new player
	-- open their class initialization menu
	PSO.lib.openMenu(p, "type_set")
end, COM_ADMIN)

local loading_player

local function server_loadPlayer(f)
	if f
		local dta = f:read()
		local run = pcall(function()
			dta = json.decode(dta)
		end)
		
		if run
			if dta
				PSO.lib.deserializePlayer(server.loadingplayer, dta)
			else
				print("No data to load for player "..#server.loadingplayer.."?")
				-- open this for the player to be set up
				PSO.lib.openMenu(p, "type_set")
				
			end
		else
			print("\x82".."WARNING: ".."\x80"..": Cannot load data for player "..#server.loadingplayer..": Data may be corrupted.")
			
			-- open this for the player to be set up
			PSO.lib.openMenu(server.loadingplayer, "type_set")
		end
		
		-- the loaded save does not have this player initialized
		if not server.loadingplayer.pso.type_initialized
			PSO.lib.openMenu(server.loadingplayer, "type_set")
		end
		
		server.loadingplayer.psologin = true
		server.loadingplayer = nil
		loading_player = nil
	else
		server.loadingplayer = nil
		loading_player = nil
		error("server_loadPlayer: File couldn't be opened? Did you ninja-delete it between verifications? Impressive timing! But the game's fucked now!")
	end
end

COM_AddCommand("__&server_loadplayer", function(host, key, pnum)
	if host ~= server return end
	local p = players[tonumber(pnum)-1]
	if p.psologin return end
	
	loading_player = true
	server.loadingplayer = p	-- keep track of who we're loading in
	logged_keys[tonumber(pnum)] = key
	
	io.open(PSO.info.serverSaves..key..".sav2", "r", server_loadPlayer)
end, COM_ADMIN)

local function server_deletePlayerFile(key)
	
	local f = io.openlocal(PSO.info.serverSaves..key..".sav2", "w")
	assert(f, "Couldn't delete player file")
	f:write("[]")
	f:close()
end	

COM_AddCommand("__&server_delplayer", function(host, pnum)
	if host ~= server return end
	pnum = tonumber($)
	print(pnum)
	local p = players[pnum-1]
	server_deletePlayerFile(logged_keys[pnum])
	logged_keys[pnum] = nil	-- remove that key
	
	if p == server
		p.pso.nukesave = true	-- this will call exitgame in a way that wont fucking crash
	else
		COM_BufInsertText(host, "kick "..(tostring(pnum)-1).." Save file deletion request")
	end
end, COM_ADMIN)

local function server_loginResponse()
	
	if gamestate ~= GS_LEVEL
		login_asks = {}
		key_asks = {}
		deletion_asks = {}
		loading_player = nil
		return
	end
	
	if isserver
	and netgame
	and not loading_player	--(this is fine actually we can load multiple files at once)
		--print(logged_keys[1])
		for i = 1, #players
			
			local p = players[i-1]
			
			if not p or not p.valid
				login_asks[i] = nil
				key_asks[i] = nil
				logged_keys[i] = nil
				
			elseif key_asks[i]
				-- player is asking for a new login key
				local newkey = server_createPlayerID()
				-- send the new key
				key_asks[i] = nil
				
				print("Sending key to player "..i)
				COM_BufInsertText(server, "__&server_sendkey "..newkey.." "..i)
				
			elseif login_asks[i]
				
				if server_checkPlayerID(login_asks[i])
					-- id exists, send command to begin making the player
					
					print("Logging in player "..i.."...")	--" with ID "..login_asks[i])
					COM_BufInsertText(server, "__&server_loadplayer "..login_asks[i].." "..i)
					
					login_asks[i] = nil
				else
					-- id doesn't exist, this is fucked up, so give this guy a new login key altogether?
					key_asks[i] = true
					login_asks[i] = nil
				end
			
			elseif deletion_asks[i]
				
				if not logged_keys[i]	-- bro doesn't exist
					deletion_asks[i] = nil
				else
					print("Deleting save from Player "..i.." as per requested by the client...")
					COM_BufInsertText(server, "__&server_delplayer "..i)
					
					deletion_asks[i] = nil
					--logged_keys[i] = nil
				end
			end
		end
	end
end

PSO.lib.registerFrameFunc(server_loginResponse)

-- when players quit, automatically save their data
addHook("PlayerQuit", function(p)
	if isserver
	and netgame
	and logged_keys[#p+1]	
	and p.pso
		server_savePlayer(#p+1)
	end
end)

addHook("GameQuit", function()
	
	if not server return end		-- title screen etc
	if not consoleplayer return end	-- fuck off dedicated servs
	
	if isserver
	and netgame
	and consoleplayer
	and consoleplayer == server
	and not (consoleplayer.pso and consoleplayer.pso.nukesave)
		server_savePlayer(#consoleplayer+1)
		--print("Netgame quit save")
		
	elseif not netgame
	and gamemap == PSO.info.LOBBY_MAP
	and not (server.pso and server.pso.nukesave)
		PSO.lib.localSave(server)
		--print("Solo quit save")
	end
end)

-- when maps change, also save
addHook("MapChange", function()

	if isserver
	and netgame
	
		for i = 1, #players
			local p = players[i-1]
			--print(logged_keys[i])
			if p and p.pso and logged_keys[i]
				--print("save "..i)
				server_savePlayer(i)
			end
		end
	elseif not netgame
		PSO.lib.localSave(server)
	end
end)