-- Used to determine the "party leader", as in the player that gets to do all the map switching and take decisions etc

local cv_pso_afktimer = CV_RegisterVar({
	name = "pso_afktimer",
	defaultvalue = 300,
	PossibleValue = CV_Unsigned,
	flags = CV_NETVAR,
})

COM_AddCommand("pso_setleader", function(p, arg)

	local compplayer
	
	for p2 in players.iterate do
		if p2 and p2.mo and p2.mo.valid
			compplayer = p2
			break
		end
	end
	
	local dedicated = compplayer ~= server
	
	if not dedicated
		CONS_Printf(p, "This only works on dedicated servers.")
		return
	end
	
	if not arg
		CONS_Printf(p, "pso_setleader <playernum>: Appoints the selected player as leader")
		return
	end
	
	arg = tonumber(arg)
	
	if players[arg] and players[arg].valid
		if players[arg].quittime
			CONS_Printf(p, "The selected player is currently disconnected.")
		else
			PSO.lib.setLeader(players[arg], true)
		end
	else
		CONS_Printf(p, "The selected player does not exist.")
	end
end, COM_ADMIN)

PSO.lib.getLeader = function(p)
	return server.playerLeader
end

PSO.lib.setLeader = function(p, log)
	
	local prevleader = server.playerLeader
	server.playerLeader = p
	
	if p
		p.afktimer = 0
		
		if prevleader ~= p
			if log
				chatprintf(p, "\x83".."*ATTENTION: You have acquired leadership and are now the player responsible of picking quests from the quest counter and taking other server-wide decisions.")
			end
			
			if p ~= consoleplayer
				chatprint("\x83".."*ATTENTION: "..p.name.." was appointed as leader")
			end
		end
	end
end

local function handlePlayerAFK()
	
	for p in players.iterate do
		local mo = p.mo
		if (not mo or not mo.valid) and not p.quittime continue end
		
		p.afktimer = $ or 0
		-- a player is considered AFK if they're simply not moving
		if not mo.momx and not mo.momy
		and not p.cmd.forwardmove and not p.cmd.sidemove
		or p.quittime
			p.afktimer = $+1
			
			if p == PSO.lib.getLeader()
				if p.afktimer == cv_pso_afktimer.value*TICRATE/2
					chatprintf(p, "\x85".."*ATTENTION: If you do not move within the next "..(cv_pso_afktimer.value/2).." seconds, your leadership will be transferred to another player.")
				elseif p.afktimer == cv_pso_afktimer.value*TICRATE*3/4
					chatprintf(p, "\x85".."*ATTENTION: If you do not move within the next "..(cv_pso_afktimer.value/4).." seconds, your leadership will be transferred to another player.")
				elseif p.afktimer == cv_pso_afktimer.value*TICRATE
				or p.quittime
					chatprintf(p, "\x85".."*ATTENTION: Your leadership has been transferred to another player.")
					PSO.lib.setLeader(nil)
				end
			end
		else
			p.afktimer = 0
		end
	end
end

local function findLeader()
	
	if not PSO.lib.getLeader()
		
		for p in players.iterate do
			
			local mo = p.mo
			if not mo or not mo.valid continue end
			
			if p.afktimer < cv_pso_afktimer.value
			and not p.quittime
				
				PSO.lib.setLeader(p, true)
				break
			end
		end
		
		-- even if there is no leader, the game won't break, nobody will be able to take decisions.
		-- but this is fine as nobody is playing anyways
	end	
end

local function setLeader()
	
	-- how to find out if we're in dedicated? 
	-- well normally we'd have a constant but this game is gay so we don't
	-- instead simply check if the first player players.iterate gets us is the server
	-- if it's not then we have a dedicated server.
	
	local compplayer
	
	for p in players.iterate do
		if p and p.mo and p.mo.valid
			compplayer = p
			break
		end
	end
	
	local dedicated = compplayer ~= server
	
	
	if not dedicated
		server.playerLeader = server	-- straightforward enough, in games hosted by a player, said player is always the leader
	else
		handlePlayerAFK()
		findLeader()
		
		-- if we have absolutely no players, return to lobby
		if not PSO.lib.countPlayers() and gamemap ~= PSO.info.LOBBY_MAP
		and gametype == GM_COOP
			G_SetCustomExitVars(1, 1)
			G_ExitLevel()
		end
	end
end

PSO.lib.registerFrameFunc(setLeader)

-- 06/07: Reset leader every map...
addHook("MapLoad", do	
	
	if not server return end 	-- fucking titlemap
	
	local leader = PSO.lib.getLeader()
	
	if isdedicatedserver and leader and leader.valid
		--print("pso_setleader "..(#leader))
		COM_BufInsertText(server, "pso_setleader "..(#leader))
	end
end)