-- inputs can be remapped etc...
-- make defaults

local inputs = {
	
	"IPT_CONFIRM",
	"IPT_BACK",
	"IPT_INTERACT",
	"IPT_ATTACK",
	"IPT_ATTACK2",
	"IPT_PALETTE",
	"IPT_SECONDPALETTE",
	"IPT_LOCKON",
	"IPT_MENU",
}

for i = 1, #inputs
	rawset(_G, inputs[i], i)
end

rawset(_G, "IPT_MENUSEP", IPT_BACK)	

PSO.info.inputDefaults = {
	[IPT_CONFIRM] = BT_JUMP,
	[IPT_BACK] = BT_SPIN,
	[IPT_INTERACT] = BT_JUMP,
	[IPT_ATTACK] = BT_ATTACK,
	[IPT_ATTACK2] = BT_FIRENORMAL,
	[IPT_PALETTE] = BT_SPIN,
	[IPT_SECONDPALETTE] = BT_CUSTOM1,
	[IPT_LOCKON] = BT_CUSTOM2,
	[IPT_MENU] = BT_TOSSFLAG,
}

PSO.info.btIcons = {
	[BT_JUMP] = "H_BT_JM",
	[BT_SPIN] = "H_BT_SN",
	[BT_ATTACK] = "H_BT_FR",
	[BT_FIRENORMAL] = "H_BT_NF",
	[BT_WEAPONNEXT] = "H_BT_WN",
	[BT_WEAPONPREV] = "H_BT_WP",
	[BT_TOSSFLAG] = "H_BT_TF",
	[BT_CUSTOM1] = "H_BT_C1",
	[BT_CUSTOM2] = "H_BT_C2",
	[BT_CUSTOM3] = "H_BT_C3",
}

PSO.info.btIcons2 = {
	[BT_JUMP] = "H_B2_JM",
	[BT_SPIN] = "H_B2_SN",
	[BT_ATTACK] = "H_B2_FR",
	[BT_FIRENORMAL] = "H_B2_NF",
	[BT_WEAPONNEXT] = "H_B2_WN",
	[BT_WEAPONPREV] = "H_B2_WP",
	[BT_TOSSFLAG] = "H_B2_TF",
	[BT_CUSTOM1] = "H_B2_C1",
	[BT_CUSTOM2] = "H_B2_C2",
	[BT_CUSTOM3] = "H_B2_C3",
}

-- get the button a player has set for a particular input
PSO.lib.playerButton = function(p, ipt)
	return p.pso.inputs[ipt]
end

-- gets the button icon by doing the same thing
PSO.lib.playerButtonIcon = function(p, ipt, flash)
	return flash and PSO.info.btIcons2[PSO.lib.playerButton(p, ipt)] or PSO.info.btIcons[PSO.lib.playerButton(p, ipt)]
end

PSO.lib.playerInputDown = function(p, ipt)
	if p.pso.buttons
		return ((p.pso.buttons[ipt] and p.pso.buttons[ipt] > 0) and p.pso.buttons[ipt] or 0)
	end
end

-- forces an input to be released.
-- if "only" is true, only does it for the input you want.
-- otherwise, also does it for every other input that shares the same key
PSO.lib.playerForceInputUp = function(p, ipt, only)
	
	if only
		p.pso.buttons[ipt] = -FRACUNIT
	else
		
		local refb = p.pso.inputs[ipt]
		--print(refb)
		for i = 1, #p.pso.inputs
			if p.pso.inputs[i] == refb
				--print("input "..i.." matches")
				p.pso.buttons[i] = -FRACUNIT
			end
		end
	end
end

local function playerInputTracker(p)
	if p.pso and p.pso.buttons
		
		local inpts = {
			IPT_CONFIRM,
			IPT_BACK,
			IPT_INTERACT,
			IPT_ATTACK,
			IPT_ATTACK2,
			IPT_PALETTE,
			IPT_SECONDPALETTE,
			IPT_LOCKON,
			IPT_MENU,
		}
		
		for i = 1, #inpts
			local bt = PSO.lib.playerButton(p, inpts[i])
			if p.cmd.buttons & bt
				p.pso.buttons[inpts[i]] = $ and $+1 or 1
			else
				if p.pso.buttons[inpts[i]]
					p.pso.buttons[inpts[i]] = 0
				end
			end
		end
		

		if p.cmd.forwardmove > 25
			p.pso.buttons["up"] = $ and $+1 or 1
		else
			p.pso.buttons["up"] = 0
		end
		
		if p.cmd.forwardmove < -25
			p.pso.buttons["down"] = $ and $+1 or 1
		else
			p.pso.buttons["down"] = 0
		end

		if p.cmd.sidemove > 25
			p.pso.buttons["right"] = $ and $+1 or 1
		else
			p.pso.buttons["right"] = 0
		end			

		if p.cmd.sidemove < -25
			p.pso.buttons["left"] = $ and $+1 or 1
		else
			p.pso.buttons["left"] = 0
		end
		
	end
end

PSO.lib.registerPlayerFunc(playerInputTracker)