-- Handgun PAs

PSO.info.photonArts["handgun_normal"] = {
		name = "Shot",
		desc = "A quick draw from a trusty handgun that allows shooting while retaining your freedom of movement",
		cost = 1,
		normal = true,
		potency = 100,
		speed = 70,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_HGUN1",
		combo = TICRATE/2,		-- can be reused after 17 frames on the timer!
		
		-- p: player
		-- timer: how long the animation has played for
		-- hold: if true, button is being held down, if not, it's been released
		-- input: timer on how long the button has been held for
		-- combo: non-zero if the attack was performed from a combo.
					-- useful to create attacks that are faster/different if used repeatedly.
		
		-- return true to end the animation
		
		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[HAND_LEFT]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle + ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = -ANG1*10	-- lat. angle
					
					if timer == 1
						
						if combo
							pos[5] = FixedDiv(mo.height, mo.scale)/2
							pos[6] = 0
							
						else	-- we've just started that attack
							local z = FixedDiv(mo.height, mo.scale)
							pos[4] = z
								
							-- set angle
							pos[6] = ANG1*90 -- vert. angle
						end
					end
					
					if timer < 5
					and not combo
						pos[6] = $ - ANG1*20
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					elseif timer == 5
						pos[6] = 0
					end
					
					
					if timer == 9
						pos[6] = ANG1*30
						
						-- fire the bullet!
						local msl = PSO.lib.spawnMissile(mo, MT_PSO_RIFLESHOT, p.drawangle, 65*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], mo.scale*30)
						msl.scale = $/3
						if winfo.effectcolor
							msl.color = winfo.effectcolor
						else
							msl.color = PSO.info.elementColours[wep.element]
						end	
						msl.pso.potency = potency	-- normals shouldn't benefit from PA%
												
						if p.pso.lockon and p.pso.lockon.valid
							PSO.lib.aimMissile(msl, p.pso.lockon)
						end

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end

						-- make the missile fade after it gets out of the range.
						local range = wep.range or 512
						msl.fuse = FixedDiv(range*FRACUNIT, FixedHypot(msl.momx, msl.momy))/FRACUNIT
						
					elseif timer > 9 and timer < 13
						pos[6] = $ - ANG1*10
					end
					
					if timer == TICRATE
						return true
					end
				end,
}

PSO.info.photonArts["handgun_pa_charge"] = {
		name = "Power Bullet",
		desc = "Fire a powerful bullet. Charge and release to power the attack up and make it explode on impact",
		cost = 15,
		potency = 300,
		speed = 40,
		
		icon = "H_HGUN2",
		
		use = 	function(p, timer, hold, input)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					local chargethreshold = PSO.lib.hasSkill(p, "ra_quickload") and TICRATE*2/3 or TICRATE
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle + ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = -ANG1*10	-- lat. angle
					
					if timer == 1
						local z = FixedDiv(mo.height, mo.scale)
						pos[4] = z
							
						-- set angle
						pos[6] = ANG1*90 -- vert. angle
					end
					
					if timer < 5
						pos[6] = $ - ANG1*20
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					elseif timer == 5
						pos[6] = 0
					end
					
					if timer >= 5
					and timer < chargethreshold*3/4

						if timer == 5
							S_StartSound(mo, sfx_s3ka3)
						end
						
						-- show the charge...
						local cang = p.drawangle + ANG1*5
						local cx = 50*cos(cang)
						local cy = 50*sin(cang)
						
						if leveltime & 1
							local t = P_SpawnMobj(mo.x + FixedMul(cx, mo.scale), mo.y + FixedMul(cy, mo.scale), mo.z + mo.height*2/3, MT_THOK)
							t.tics = 2
							t.fuse = 2
							t.scale = mo.scale/3
							t.frame = A
							t.color = PSO.info.elementColours[wep.element]
							
							local t2 = P_SpawnMobj(t.x + P_RandomRange(-64, 64)*mo.scale, t.y + P_RandomRange(-64, 64)*mo.scale, t.z + P_RandomRange(-64, 64)*mo.scale, MT_THOK)
							t2.scale = mo.scale/4
							t2.destscale = mo.scale/8
							t2.scalespeed = mo.scale/12
							t2.frame = A
							t2.color = PSO.info.elementColours[wep.element]
							
							t2.momx = (t.x - t2.x)/10
							t2.momy = (t.y - t2.y)/10
							t2.momz = (t.z - t2.z)/10
							
							t2.fuse = 10
							t2.tics = 10
							
						end
					end
					
					if timer >= chargethreshold/3
					and timer < TICRATE*20
					and not hold
						-- just let it go.
						p.pso.actiontimer = timer < chargethreshold and TICRATE*20 or TICRATE*30	-- first firing state.
						pos[6] = timer < chargethreshold and ANG1*40 or ANG1*90
						
						local msl = PSO.lib.spawnMissile(mo, MT_PSO_RIFLESHOT, p.drawangle, 70*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], mo.scale*30)
						msl.pso.combobreaker = true	-- charge shots are always combo breakers
						msl.extravalue1 = 1
						if winfo.effectcolor
							msl.color = winfo.effectcolor
						else
							msl.color = PSO.info.elementColours[wep.element]
						end	
						
						if p.pso.lockon and p.pso.lockon.valid
							PSO.lib.aimMissile(msl, p.pso.lockon)
						end

						local range = wep.range or 512
						msl.fuse = FixedDiv(range*FRACUNIT, FixedHypot(msl.momx, msl.momy))/FRACUNIT

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end
						S_StartSound(mo, sfx_s3k83)						

						if timer >= chargethreshold*3/4
							msl.pso.potency = $+200
							msl.extravalue2 = 1
							
							P_InstaThrust(mo, p.drawangle, -10*FRACUNIT)
							p.skidtime = 20
							mo.saveang = p.drawangle
							p.pso.actionspeed = 0	-- remove all cmd prompts.
							p.drawangle = mo.angle
							mo.state = S_PLAY_WALK
						else
							msl.scale = $ / 2
						end
					end
					
					-- if we keep holding we'll eventually reach full charge!
					if timer == chargethreshold*3/4
						S_StopSound(mo)
						local g = P_SpawnGhostMobj(mo)
						g.destscale = mo.scale*4
						g.scalespeed = mo.scale/12
						S_StartSound(mo, sfx_s3ka8)
					end
					
					if timer >= TICRATE*20
					and timer < TICRATE*20 + 5
						pos[6] = $ - ANG1*10
					elseif timer == TICRATE*21
						return true
					end
					
					
					if timer >= TICRATE*30
					and timer < TICRATE*31
						p.drawangle = mo.saveang
						p.cmd.angleturn = mo.saveang/FRACUNIT
					end	
					
					if timer >= TICRATE*30
					and timer < TICRATE*30 + 10
						pos[6] = $ - ANG1*10
					elseif timer == TICRATE*31
						P_InstaThrust(mo, 0, 0)
						p.drawangle = mo.saveang
						return true
					end
					
					-- hack to prevent us from reaching firing states.
					if p.pso.actiontimer > TICRATE*10
					and p.pso.actiontimer < TICRATE*20	-- if we're past here, don't do it.
						p.pso.actiontimer = TICRATE*10
					end
					
					
					/*if timer == 9
						pos[6] = -ANG1*30
						
						-- fire the bullet!
						PSO.lib.spawnMissile(mo, MT_ROCKET, p.drawangle, 40*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], pos[4])
						
					elseif timer > 9 and timer < 13
						pos[6] = $ + ANG1*10
					end
					
					if timer == TICRATE/2
						return true
					end*/
				end,
		
}