-- some player related constants and stats
freeslot("sfx_lvup", "sfx_rnkup", "sfx_rnkup2", "sfx_rank1", "sfx_rank2", "sfx_rank3", "sfx_rank4", "sfx_mopen")

PSO.info.dodgeroll_cost = 15

PSO.info.typestats = {
	
	[TYPE_HUNTER] = {
		
		hp = 250,
		pp = 85,
		
		satk = 20,
		ratk = 15,
		tatk = 10,
		def = 10,
		ppregen = 18,	-- 1 pp / 18f
		
		startingweapon = "sword",
		startingweaponr = "saber",	-- starting right h weapon
		startingweaponl = "handgun",		-- starting left h weapon
		
		-- weapon proficiency:
		/*
			0: Cannot equip
			1: 1-3*
			2: 4-6*
			3: 7-9*
			4: 10*-15*
		*/	
		weapons = {
			[WPT_SWORD] = 2,
			[WPT_DOUBLESABER] = 2,
			[WPT_RIFLE] = 0,
			[WPT_SHOTGUN] = 0,
			[WPT_ROD] = 0,
			[WPT_SABER] = 2,
			[WPT_HANDGUN] = 1,
			[WPT_WAND] = 0,
		}
	},

	[TYPE_RANGER] = {
		
		hp = 225,
		pp = 100,
		
		satk = 10,
		ratk = 20,
		tatk = 15,
		def = 10,
		ppregen = 6,	-- 1 pp / 7f
		
		startingweapon = "rifle",
		startingweaponr = "saber",	-- starting right h weapon
		startingweaponl = "handgun",		-- starting left h weapon
		
		weapons = {
			[WPT_SWORD] = 0,
			[WPT_DOUBLESABER] = 0,
			[WPT_RIFLE] = 2,
			[WPT_SHOTGUN] = 2,
			[WPT_ROD] = 0,
			[WPT_SABER] = 1,
			[WPT_HANDGUN] = 2,
			[WPT_WAND] = 0,
		}
	},

	[TYPE_FORCE] = {
		
		hp = 200,
		pp = 115,
		
		satk = 10,
		ratk = 15,
		tatk = 20,
		def = 10,
		ppregen = 4,	-- 1 pp / 5f
		
		startingweapon = "rod",
		startingweaponr = "wand",	-- starting right h weapon
		startingweaponl = "handgun",		-- starting left h weapon
		
		weapons = {
			[WPT_SWORD] = 0,
			[WPT_DOUBLESABER] = 0,
			[WPT_RIFLE] = 0,
			[WPT_SHOTGUN] = 0,
			[WPT_ROD] = 2,
			[WPT_SABER] = 0,
			[WPT_HANDGUN] = 1,
			[WPT_WAND] = 2,			
		}
	},
}

PSO.lib.initPlayerType = function(p, typ)

	local info = PSO.info.typestats[typ]
		
	local wep = PSO.lib.makeItem(info.startingweapon, {ELEM_ICE, 10})
	local wep_r = PSO.lib.makeItem(info.startingweaponr, {ELEM_ICE, 10})
	local wep_l = PSO.lib.makeItem(info.startingweaponl, {ELEM_ICE, 10})
	
	wep.equipped = true	-- cheat a little
	wep_r.equipped = true
	wep_l.equipped = true
	
	local armor = PSO.lib.makeItem("frame", {ELEM_ICE, 10})
	armor.equipped = true	-- cheat again
		
	-- defined like a shizo to test inventory sorting
	local inventory = {
		
		{wep},
		{wep_l},
		{wep_r},

		{armor},
		
		{"monomate", 5},
		{"dimate", 5},
		{"trimate", 2},
		{"moon atomizer", 1},
	}
		
	local palette = {
			
		-- [1]: consumables
		-- [2]: weapons: table because of dual wield
		-- [3]: armors
		{"monomate", "dimate", "trimate", "moon atomizer"},
		{{wep}, {wep_l, wep_r}},
		{armor},
	}
	
	p.pso.inventory = inventory
	p.pso.palette.layout = palette
end

-- EXP tables...
PSO.info.exptable = {50}
PSO.info.expdist = 2048 -- you have to be within 2048 units of something dying to get exp from it

-- the rest of the exp table
for i = 2, 201
	local mult = 100 + FixedDiv(90<<FRACBITS, FixedSqrt((i-1)<<FRACBITS)*2)/FRACUNIT
	PSO.info.exptable[i] = PSO.info.exptable[i-1] *mult/100
	--print("Lv "..i..": "..PSO.info.exptable[i].." ("..mult..")")
end

PSO.lib.givePlayerExp = function(p, exp)
	
	if p.pso.level >= PSO.info.LEVELCAP
		p.pso.exp = 0
		return
	end
	
	local needxp = PSO.info.exptable[p.pso.level]
	p.pso.exp = $ + exp
	
	--print("Got "..exp.." EXP")
	
	local nosound = false
	
	while p.pso.exp >= PSO.info.exptable[p.pso.level]
		p.pso.exp = $ - PSO.info.exptable[p.pso.level]
		PSO.lib.levelUpPlayer(p, nosound)
		nosound = true	-- protect your ears <3
	end
end

COM_AddCommand("psod_getexp", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local expn = tonumber(arg) or 0
	PSO.lib.givePlayerExp(p, arg)
end)

COM_AddCommand("psod_getmeseta", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local meseta = tonumber(arg) or 0
	p.pso.meseta = $+meseta
end)

COM_AddCommand("psod_gettp", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local tp = tonumber(arg) or 0
	p.pso.typepoints[p.pso.type] = $+tp
end)

-- calculates the player's stats for a given type & level
-- if units is true, also counts unit stats
PSO.lib.playerPreviewStats = function(p, type, level, units)
	local stats = {}
	local typinfo = PSO.info.typestats[type]
	stats.maxhp = typinfo.hp + (typinfo.hp/50)*(level-1)
	stats.maxpp = typinfo.pp + (level-1)
	
	stats.satk = typinfo.satk + (typinfo.satk/4)*(level-1)
	stats.ratk = typinfo.ratk + (typinfo.ratk/4)*(level-1)
	stats.tatk = typinfo.tatk + (typinfo.tatk/4)*(level-1)
	stats.def = typinfo.def + (typinfo.def/4)*(level-1)
	stats.ppregen = typinfo.ppregen
	
	if units
		local armour = PSO.lib.getPlayerActiveArmour(p)
		if armour			
			local def = PSO.info.items[armour.ref]
			if def.slots
				for i = 1, def.slots
					
					if armour.slots[i]
						local unit = armour.slots[i]
						stats.maxhp = $ + (unit.hp or 0)
						stats.maxpp = $ + (unit.pp or 0)
						stats.satk = $ + (unit.satk or 0)
						stats.ratk = $ + (unit.ratk or 0)
						stats.tatk = $ + (unit.tatk or 0)
						stats.ppregen = max(1, $ - (unit.ppregen or 0))
					end
				end
			end
			
		end
	end
	
	return stats
end

-- sets the player's stats for a given level using playerPreviewStats
PSO.lib.playerCalcStats = function(p, level)
	local stats = PSO.lib.playerPreviewStats(p, p.pso.type, level, true)
	
	for k, v in pairs(stats)
		p.pso[k] = max(1, v)
	end
	
	-- make sure we don't keep extra hp/pp if we've taken units off
	p.pso.hp = min($, p.pso.maxhp)
	p.pso.pp = min($, p.pso.maxpp)
end

-- changes the player's type.
-- if unequip is true, also clears the weapon palettes
PSO.lib.playerChangeType = function(p, newtype, skipunequip)
	p.pso.type = newtype
	local typinfo = PSO.info.typestats[p.pso.type]
	
	PSO.lib.playerCalcStats(p, p.pso.level)
	
	p.pso.hp = min($, p.pso.maxhp)
	p.pso.pp = min($, p.pso.maxpp)
	
	-- unequip all your shit
	if not skipunequip
		for i = 1, PSO.info.MAXPALSIZE
			
			if p.pso.palette.layout[2][i]
				for j = 1, 2
					if p.pso.palette.layout[2][i][j]
						p.pso.palette.layout[2][i][j].equipped = nil
					end
				end
			end
		end
		
		p.pso.palette.aweaponnum = 1
		p.pso.palette.weaponnum = 1
		
		p.pso.forceupdateweapon = true
		p.pso.palette.layout[2] = {}
		for i = 1, PSO.info.MAXPALSIZE
			p.pso.palette.layout[2][i] = nil
		end
		
		p.pso.palette.layout[2][1] = {}
	end
end

-- returns true/false, and a reason
-- if true, reason will always be 0 (OK)
-- if false, reason will either be 1 (level too low)
-- or 2, (proficiency too low)
PSO.lib.canEquip = function(p, item)
	
	local info = PSO.lib.itemInfo(item)
	local req = info.req or 1
	
	if p.pso.level < req
		return false, 1
	elseif info.type == IT_WEAPON
		-- weapons check for proficiency:
		local proficiency = p.pso.weaponproficiencies[p.pso.type][info.weapontype]
		
		if proficiency < (max(0, info.rarity -1)/3) +1
			return false, 2
		end
	end
	
	return true, 0
end

PSO.lib.levelUpPlayer = function(p, silent)
	
	--print("Level up!")
	
	if p.pso.level >= PSO.info.LEVELCAP
		return
	end
	
	local typinfo = PSO.info.typestats[p.pso.type]
	p.pso.level = $+1
	
	-- recalculate stats!
	PSO.lib.playerCalcStats(p, p.pso.level)
	
	p.pso.hp = p.pso.maxhp
	p.pso.pp = p.pso.maxpp
	
	if not silent
		S_StartSound(nil, sfx_lvup, p)
	end
end

PSO.lib.itemSort = function(a, b)
	
	local it1 = a[1]
	local it2 = b[1]
	
	local info1 = PSO.lib.itemInfo(it1)
	local info2 = PSO.lib.itemInfo(it2)
	
	if info1.type ~= info2.type
		return (info1.type or 1) < (info2.type or 1)
	end
	
	--print(tostring(itname(it1))..", "..tostring(itname(it2)))
	-- both items are weapons/armour
	if type(it1) == "table" and type(it2) == "table"
		
		if info1.type == info2.type
			
			-- comparing weapons!
			if info1.type == IT_WEAPON
			
				if info1.weapontype ~= info2.weapontype
					return (info1.weapontype or 0) < (info2.weapontype or 0)
				elseif info1.weapontype == info2.weapontype
					if info1.atk ~= info2.atk
						return (info1.atk or 1) > (info2.atk or 1)
					elseif info1.rarity ~= info2.rarity
						return (info1.rarity or 1) > (info2.rarity or 1)
					elseif it1.equipped ~= it2.equipped
						return it1.equipped
					else
						return info1.name < info2.name
					end
				end
			
			-- comparing armours!
			elseif info1.type == IT_ARMOR

				if info1.def ~= info2.def
					return (info1.def or 1) > (info2.def or 1)
				elseif info1.rarity ~= info2.rarity
					return (info1.rarity or 1) > (info2.rarity or 1)
				elseif it1.equipped ~= it2.equipped
					return it1.equipped					
				else
					return info1.name < info2.name		
				end
			
			elseif info1.type == IT_UNIT
				if info1.rarity ~= info2.rarity
					return (info1.rarity or 1) > (info2.rarity or 1)
				elseif it1.equipped ~= it2.equipped
					return it1.equipped
				else
					return info1.name < info2.name	-- nothing yet.
				end
			end
			
		else
			return info1.name < info2.name	-- yes this works as last resort.
		end
		
	-- userdata vs string reference, userdata always wins.	
	-- impossible case
	/*elseif type(it1) ~= type(it2)
		return type(it1) == "table"	-- take whichever is userdata*/
	
	-- string reference (not weapon/armor)
	elseif type(it1) == "string" and type(it2) == "string"
		
		if info1.itemtype ~= info2.itemtype
			return info1.itemtype < info2.itemtype
		elseif info1.sortpriority ~= info2.sortpriority
			return (info1.sortpriority or 9999) < (info2.sortpriorirty or 9999)
		elseif info1.rarity ~= info2.rarity
			return (info1.rarity or 1) < (info2.rarity or 1)
		else
			
			if info1.rarity ~= info2.rarity
				return info1.rarity > info2.rarity
			else
				return info1.name < info2.name	-- yes this works
			end
		end	
	end
end

-- sorts the player's inventory and sets the variables telling where each item category begins.
-- this uses PSO.lib.itemSort
PSO.lib.playerSortItems = function(p, sortwhat)
	--print("sort")
	sortwhat = $ or "inventory"
	table.sort(p.pso[sortwhat], PSO.lib.itemSort)
	p.pso[sortwhat.."_sections"] = {1, 0, 0, #p.pso[sortwhat]+1}
	
	local i = #p.pso[sortwhat]
	
	while i
		
		local it = p.pso[sortwhat][i][1]
		local info = PSO.lib.itemInfo(it)
		
		-- we start armours at the first thing that isn't a weapon!
		if info.type == IT_ARMOR or info.type == IT_UNIT
			p.pso[sortwhat.."_sections"][2] = i
		end
		
		-- we're looking for something that's nor an armor, nor a unit!
		if info.type > IT_UNIT
			p.pso[sortwhat.."_sections"][3] = i
		end
		
		i = $-1
	end
	
	i = 3
	-- go over sections and give them the max possible thing if they haven't been set...
	while i > 1
		p.pso[sortwhat.."_sections"][i] = $ or p.pso[sortwhat.."_sections"][i+1]
		i = $-1
	end
end

-- implement this sorting method to inventory, storage and shop
PSO.lib.playerSortInventory = function(p)
	PSO.lib.playerSortItems(p, "inventory")
end

PSO.lib.playerSortStorage = function(p)
	PSO.lib.playerSortItems(p, "storage")
end

PSO.lib.playerSortShop = function(p)
	PSO.lib.playerSortItems(p, "shop")
end


-- add items to a table, stacks items that can be stacked.
PSO.lib.playerAddItemToTable = function(p, item, qty, where, sortfunc, limit, ignorecarrylimit)
	local info = PSO.lib.itemInfo(item)
	qty = $ or 1
	
	if info.type == IT_WEAPON
	or info.type == IT_ARMOR
	or info.type == IT_UNIT	-- non stackable...
		
		if #p.pso[where] >= limit
			return false
		end
		
		p.pso[where][#p.pso[where]+1] = {item, 1}
		sortfunc(p)
		return true
		
	else	-- stackable
		-- search for an instance of this item...
		for i = 1, #p.pso[where]
			if p.pso[where][i][1] == item
				
				if not ignorecarrylimit	-- ie: storage doesn't check for carry limit
				and info.carrylimit
					local curr = p.pso[where][i][2]
					
					-- nope nope nope.
					if curr >= info.carrylimit
						return false
					end
					
					p.pso[where][i][2] = $+qty
					p.pso[where][i][2] = min($, info.carrylimit)
					
					-- return how many were taken (inventory doesn't need sorting)
					return min(qty, info.carrylimit - curr)
				else
					p.pso[where][i][2] = $+qty
				end
				--PSO.lib.playerSortInventory(p)
				return true
			end
		end
		
		-- if we made it here we haven't found an instance
		if #p.pso[where] >= limit
			return false
		end
		
		p.pso[where][#p.pso[where]+1] = {item, qty or 1}
		
		-- check if we CAN hold that many
		if not ignorecarrylimit
		and info.carrylimit
		and p.pso[where][#p.pso[where]][2] > info.carrylimit
			p.pso[where][#p.pso[where]][2] = min($, info.carrylimit)
			return info.carrylimit
		end
		
		--print("A")
		sortfunc(p)
		return true
	end
end

PSO.lib.playerGetItem = function(p, item, qty)
	return PSO.lib.playerAddItemToTable(p, item, qty, "inventory", PSO.lib.playerSortInventory, PSO.info.INVENTORYSIZE)
end


PSO.lib.playerRemoveItemFromTable = function(p, slot, qty, where, sortfunc)
	
	qty = $ or 1
	local it = p.pso[where][slot][1]
	if type(it) == "table"
		-- some sanity checks:
		if it.equipped return false end	-- cannot remove equipped item.
	end
	
	p.pso[where][slot][2] = $ or 1	-- if somehow the qty had never been set
	p.pso[where][slot][2] = max($ - qty, 0)
	
	if not p.pso[where][slot][2]
		table.remove(p.pso[where], slot)
		sortfunc(p)
	end
	return true
end

PSO.lib.playerLoseItem = function(p, slot, qty)
	return PSO.lib.playerRemoveItemFromTable(p, slot, qty, "inventory", PSO.lib.playerSortInventory)
end


-- sends an item from the player's inventory to the storage
PSO.lib.playerSendToStorage = function(p, itslot, qty)
	
	local item = p.pso.inventory[itslot][1]
	-- try to add it to the storage, if it fails, don't deplete it from the inventory
	local res = PSO.lib.playerAddItemToTable(p, item, qty, "storage", PSO.lib.playerSortStorage, PSO.info.STORAGESIZE, true)	
	-- failure
	if not res
		--print("\x82".."WARNING: ".."\x80".."Couldn't send item to storage!")
		return false
	end
	
	if type(res) == "number"
		return PSO.lib.playerLoseItem(p, itslot, res)
	else
		-- if we're here, also lose the items here.
		return PSO.lib.playerLoseItem(p, itslot, qty)
	end
end

-- retrieves an item from the player's storage and adds it to their inventory
PSO.lib.playerRetrieveFromStorage = function(p, itslot, qty)
	local item = p.pso.storage[itslot][1]
	
	-- try to add it to their inventory, if it fails, don't deplete it from the storage
	local res = PSO.lib.playerGetItem(p, item, qty)
	-- failure
	if not res
		return false
	end
	
	--print(res)
	if type(res) == "number"
		return PSO.lib.playerRemoveItemFromTable(p, itslot, res, "storage", PSO.lib.playerSortStorage)
	else
		return PSO.lib.playerRemoveItemFromTable(p, itslot, qty, "storage", PSO.lib.playerSortStorage)
	end
end

-- returns the number of the inventory slot the desired item is in, or 0 if it isn't there
PSO.lib.findPlayerItemInTable = function(p, item, where)
	for i = 1, #p.pso[where]
		local it = p.pso[where][i][1]
		
		if it == item
		or type(it) == "table" and it.ref == item
			return i
		end
	end
	return 0
end

-- unclean fixer for units that somehow show as equipped when they shouldn't be
PSO.lib.playerFixUnits = function(p)
	
	local armours = {}
	local units = {}
	
	-- make a list of armours and units
	for i = 1, #p.pso.inventory
		local it = p.pso.inventory[i][1]
		if it and it.type == IT_ARMOR
		and PSO.lib.itemInfo(it).slots	-- this shit better have slots.
			armours[#armours+1] = it
		elseif it and it.type == IT_UNIT and it.equipped	-- only check for units that say they're equipped
			units[#units+1] = it
		end
	end
	
	-- check if the units are equipped
	for i = 1, #armours
		local a = armours[i]
		for j = 1, PSO.lib.itemInfo(a).slots
			if a.slots[j]
				local k = #units
				while k
					if a.slots[j] == units[k]
						table.remove(units, k)	-- it's equipped so it doesn't need to be here anymore.
					end
					k = $-1
				end
			end
		end
	end
	
	-- every unit left at this point is unequipped and must be killed
	for i = 1, #units
		local u = units[i]
		u.equipped = false
		u.equippedarmours = 0	-- liar.
	end
end

PSO.lib.findPlayerItem = function(p, item)
	return PSO.lib.findPlayerItemInTable(p, item, "inventory")
end

PSO.lib.findPlayerStorageItem = function(p, item)
	return PSO.lib.findPlayerItemInTable(p, item, "storage")
end

PSO.lib.getPlayerActiveArmour = function(p)
	local pal = p.pso.palette.layout[3][p.pso.palette.armornum]
	return pal
end

-- sets trailing palette slots to nil
PSO.lib.playerFixWeaponPalette = function(p)
	local firstwep
	
	local pal = p.pso.palette.layout[2]
	
	for i = 1, PSO.info.MAXPALSIZE
		if not PSO.lib.slotHasWeapon(pal[i])
			pal[i] = nil
		end
	end
	
	-- if we set that to nil, don't do it.
	pal[p.pso.palette.aweaponnum] = $ or {}
end

-- get the weapon we're displaying the actions of.
PSO.lib.getPlayerHUDWeapon = function(p)
	
	local slot = 1
	local pal = p.pso.palette.layout[2][p.pso.palette.aweaponnum]
	local onehanded = false
	
	for i = 1, 2
		if PSO.lib.slotIsOneHanded(pal)
			onehanded = true
		end
	end
		
	if onehanded and (not p.pso.backpalette or not pal[1])
	and pal[2]	
		slot = 2	-- by default, use right handed weapon
	end
	
	return pal[slot]
end

-- get the player's active weapon.
-- similar to HUD weapon but is locked to the action the player is currently doing
PSO.lib.getPlayerActiveWeapon = function(p)
	
	local slot = 1
	local pal = p.pso.palette.layout[2][p.pso.palette.aweaponnum]
	local onehanded = false
	
	if p.pso.actionslot
		slot = p.pso.actionslot
	else
		for i = 1, 2
			if PSO.lib.slotIsOneHanded(pal)
				onehanded = true
			end
		end
		
		if onehanded and (not p.pso.backpalette or not pal[1])
		and pal[2]	
			slot = 2	-- by default, use right handed weapon
		end
	end
	
	return pal[slot]
end

PSO.lib.slotIsOneHanded = function(slot)
	return slot and (slot[2] or (slot[1] and PSO.info.weaponHand[slot[1].weapontype] ~= HAND_BOTH))
end

-- returns true if the slot has at least 1 weapon
PSO.lib.slotHasWeapon = function(slot)
	return slot and (slot[1] or slot[2])
end

PSO.lib.getPlayerArmor = function(p)
	return p.pso.palette.layout[3][p.pso.palette.armornum]
end

PSO.lib.resetPlayerAction = function(p)
	p.pso.action = nil
	p.pso.actionslot = nil
	p.pso.taction = nil
	p.pso.actionbutton = nil
	p.pso.actionweapon = nil
	p.pso.actiontimer = 0
	p.pso.actionheld = false
	p.pso.actionhold = 0
	p.pso.actionnum = 0
	p.pso.actionspeed = nil
	p.pso.actioncombo = 0	
end

-- resets player actions etc.
-- for use when teleporting / between maps.
PSO.lib.softResetPlayer = function(p)
	
	p.pso.teleport = nil
	p.pso.teleporttarget = nil
	
	p.pso.jumppad = nil
	if p.mo and p.mo.valid
		p.thrustfactor = skins[p.mo.skin].thrustfactor
	end
	
	PSO.lib.removeStatus(p)
	PSO.lib.resetPlayerAction(p)
end

local function calculateQuestScore(p)
	if server.questdata
	and p.pso.quest
	and gamemap == server.questdata.map	
		local q = PSO.info.quests[gamemap][server.questdata.index]
		local total = 0
		local totalgrade = 0
		local diff = PSO.lib.questDifficulty()
		
		p.pso.quest.grades = {}
		
		if q.grading
			
			for i = 1, #q.grading do
				local playerval = 0
				local grade = 0
				local g = q.grading[i]	
				local gtype = g[1]
				
				p.pso.quest.grades[i] = {}	-- {player_val, score_val}
				
				if gtype == QG_TIME
					
					local diffmult = 100
					
					if diff >= DIFF_HARD
						diffmult = 130
					end
					
					local full = g[2]
					local low = g[3]
					local mintime = g[4]*diffmult/100
					local maxtime = g[5]*diffmult/100
					
					if p.pso.quest.time < mintime
						grade = full
					elseif p.pso.quest.time > maxtime
						grade = low
					else
						local normal = FixedDiv((p.pso.quest.time - mintime)*FRACUNIT, (maxtime - mintime)*FRACUNIT)
						-- and then we use black magic also known as more maths:
						grade = (normal*low + (FRACUNIT - normal)*full) /FRACUNIT
					end
					
					total = $ + full
					playerval = p.pso.quest.time
					
					--print("TIME: "..grade.."/"..full)
					
				elseif gtype == QG_KILLS
					
					local full = g[2]
					local low = 0
					local minkills = 0
					local maxkills = g[3]
					local checktype = g[4]
					
					-- count kills
					local kills = 0
					for k,v in pairs(p.pso.kills)
						if not checktype or checktype == k
							kills = $ + v
						end
					end
					
					if kills > maxkills
						grade = full
					elseif not kills
						grade = low
					else
						local normal = FixedDiv((kills - minkills)*FRACUNIT, (maxkills - minkills)*FRACUNIT)
						-- and then we use black magic also known as more maths:
						grade = (normal*low + (FRACUNIT - normal)*full) /FRACUNIT
					end
					
					total = $ + full
					playerval = kills
					
					--print("KILLS: "..grade.."/"..full)
				
				elseif gtype == QG_NOINCAP
					grade = (not p.pso.quest.incapacitations) and g[2] or 0
					playerval = p.pso.quest.incapacitations and 1 or 0
					total = $ + g[2]
					
					--print("INCAPACITATIONS: "..grade.."/"..g[2])
					
				elseif gtype == QG_DAMAGETAKEN

					local diffmult = 100
					
					if diff >= DIFF_HARD
						diffmult = 400
					end
				
					local full = g[2]
					local low = g[3]
					local mindamage = g[4]*diffmult/100
					local maxdamage = g[5]*diffmult/100
					
					if p.pso.quest.damagetaken < mindamage
						grade = full
					elseif p.pso.quest.time > maxdamage
						grade = low
					else
						local normal = FixedDiv((p.pso.quest.damagetaken - mindamage)*FRACUNIT, (maxdamage - mindamage)*FRACUNIT)
						--print(normal)
						-- and then we use black magic also known as more maths:
						grade = (normal*low + (FRACUNIT - normal)*full) /FRACUNIT
					end
					
					total = $ + full
					playerval = p.pso.quest.damagetaken
					
					--print("DAMAGETAKEN: "..grade.."/"..full)
				end
				totalgrade = $ + grade
				p.pso.quest.grades[i] = {playerval, grade}
			end

		/*grading = {
			{QG_TIME, full_point_time, 0_point_time, full_time, low_time},
			{QG_KILLS, max_points, max_kills, [MT_ type]},
			{QG_NOINCAP, points_for_no_incapacitation},
			{QG_DAMAGETAKEN, full_point_damagetaken, 0_point_damagetaken, low_damage_taken, high_damage_taken},
		}*/		
		end
		
		-- set our total rank:
		local percentage = FixedDiv(totalgrade<<FRACBITS, total<<FRACBITS)
		local gradelookup = {
			
			[RANK_S] = FRACUNIT,
			[RANK_A] = FRACUNIT*8/10,
			[RANK_B] = FRACUNIT*4/10,
			[RANK_C] = 0
		}
		
		for i = 1, #gradelookup
			if percentage >= gradelookup[i]
				p.pso.quest.ranking = i
				break
			end
		end
		
		p.pso.quest.totals = {totalgrade, total}
		p.pso.quest.typepoints = totalgrade	-- add difficulty scalar
		
		if p.pso.quest.ranking == RANK_S
			p.pso.quest.typepoints = $*2
		elseif p.pso.quest.ranking == RANK_A
			p.pso.quest.typepoints = $*3/2
		end
		
		if diff >= DIFF_HARD
			p.pso.quest.typepoints = $*5/2	-- x2.5 in hard
		end
		
		p.pso.typepoints[TYPE_HUNTER] = $+p.pso.quest.typepoints
		p.pso.typepoints[TYPE_RANGER] = $+p.pso.quest.typepoints
		p.pso.typepoints[TYPE_FORCE] = $+p.pso.quest.typepoints
		
		--print("GRADE: "..totalgrade.."/"..total)
	end
end

PSO.lib.clearQuest = function()
	
	if not server.questdata return end
	server.questdata.complete = true
	
	S_ChangeMusic("QCLEAR", true)
	
	for p in players.iterate
		
		if p.deadtimer or p.pso.hp <= 0
			PSO.lib.revivePlayer(p, p, p.pso.maxhp)
		end
		
		calculateQuestScore(p)
		p.pso.quest.completetimer = 1
		PSO.lib.playerForceInputUp(p, IPT_CONFIRM)
		PSO.lib.rollPlayerShop(p)	-- reroll the player's shop
	end
	
	-- despawn all enemies
	for m in mobjs.iterate() do
		if m and m.valid and m.flags & MF_ENEMY
			P_RemoveMobj(m)
		end
	end
	
	local q = PSO.info.quests[gamemap][server.questdata.index]
	
	if q.unlocks
		for i = 1, #q.unlocks
			local unlock = q.unlocks[i]
			
			local mapn = unlock[1]
			local qstn = unlock[2]
			local diff = unlock[3] or server.questdata.difficulty
			
			if not PSO.lib.questUnlocked(mapn, qstn)
			or PSO.lib.questUnlocked(mapn, qstn) < diff
				PSO.lib.unlockQuest(mapn, qstn, diff)
			end
		end
	end
	
	S_StartSound(nil, sfx_rnkup)
end

PSO.lib.noPPNotif = function(p)
	
	S_StartSound(nil, sfx_mnot, p)
	
	if not p.pso.nopp
		p.pso.nopp = TICRATE
	end
end

-- sets a notification on the player's screen
PSO.lib.playerNotification = function(p, msg)
	S_StartSound(nil, sfx_mopen, p)
	p.pso.notification = {
		timer = TICRATE*4,
		text = msg,
	}
end

-- jsonify the player's important variables...
PSO.lib.serializePlayer = function(p)
	
	if not p or not p.valid or not p.pso return end
	
	local d = p.pso
	local tabl = {
		
		-- whether the player has passed the first type selection check.
		type_initialized = d.type_initialized,
		
		-- core stats
		type = d.type,
		hp = d.hp,
		maxhp = d.maxhp,
		pp = d.pp,
		maxpp = d.maxpp,
		exp = d.exp,
		level = d.level,
		satk = d.satk,
		ratk = d.ratk,
		tatk = d.tatk,
		def = d.def,
		ppregen = d.ppregen,
		
		-- money
		meseta = d.meseta,
		
		-- items
		inventory = PSO.lib.copyTable(d.inventory),
		storage = PSO.lib.copyTable(d.storage),	-- will not need any fix because you can't put shit with equipment in there
		
		-- type info
		typepoints = PSO.lib.copyTable(d.typepoints),
		weaponproficiencies = PSO.lib.copyTable(d.weaponproficiencies),
		typeskills = PSO.lib.copyTable(d.typeskills),
		maxskillpoints = d.maxskillpoints,
		
		-- PAs
		photonarts = PSO.lib.compactPow2Table(d.photonarts),	-- (this drops the .techniques attribute)
		techniques = d.photonarts.techniques,	-- techs are a single value, we must separate it from PAs here
		-- otherwise we'll end up with a mixed table
		
		-- Palette
		palette = PSO.lib.copyTable(d.palette)
	}
	
	tabl.palette.layout = {{}, {}, {}}
	tabl.palette.qlayout = nil	-- this should not be saved
	
	-- we can't serialize palette shit as-is so we're gonna do some hacking and save object # instead of references
	for i = 2, 3
		local pal = p.pso.palette.layout[i]
		for j = 1, PSO.info.MAXPALSIZE
			
			if pal[j]
				if i == 2	-- weapons
					tabl.palette.layout[i][j] = {0, 0}
					for k = 1, 2
						if pal[j][k]
							local num = PSO.lib.findPlayerItem(p, pal[j][k])
							tabl.palette.layout[i][j][k] = num or 0
						else
							--print("j: "..j.." k: "..k)
							tabl.palette.layout[i][j][k] = 0	-- save a 0 instead of nil.
							-- this avoids errors if you have, say, only a saber equipped in a slot.
						end
					end
					
				else
					
					local num = PSO.lib.findPlayerItem(p, pal[j])
					tabl.palette.layout[i][j] = num or 0
				end
			else
				if i == 2
					tabl.palette.layout[i][j] = {0, 0}
				else
					tabl.palette.layout[i][j] = 0
				end
			end
		end
	end
	
	-- also check for empty spaces in the item palette,......
	for i = 1, PSO.info.MAXPALSIZE
		local pal = p.pso.palette.layout[1]
		
		if not pal[i]
			tabl.palette.layout[1][i] = 0
		else
			tabl.palette.layout[1][i] = pal[i]
		end
	end	
	
	-- serialize units and weapon actions properly.
	-- there can be gap in the setup so we must account for them
	-- or the json will error.
	local tabls = {"inventory", "storage"}
	
	for tn = 1, #tabls
		local t = tabls[tn]
		for i = 1, #p.pso[t]
			local it = p.pso[t][i][1]
			
			if type(it) == "table"
			
				if it.type == IT_ARMOR
					local info = PSO.lib.itemInfo(it)
					if info.slots
						for j = 1, info.slots
							if it.slots[j]
								
								local slotnum = PSO.lib.findPlayerItem(p, it.slots[j])
								tabl[t][i][1].slots[j] = slotnum
							else
								it.slots[j] = 0
							end
						end
					end
				
				elseif it.type == IT_WEAPON
					
					local num_actions = (PSO.info.weaponHand[it.weapontype] == HAND_BOTH) and 4 or 2
					for j = 1, num_actions
						if not it.actions[j]
							tabl[t][i][1].actions[j] = 0
						end
					end
				end
			end
		end
	end
	
	-- proficiencies is a sparse table, fix that
	for i = 1, #tabl.weaponproficiencies
		tabl.weaponproficiencies[i] = PSO.lib.compactPow2Table(tabl.weaponproficiencies[i])
	end
	
	return json.encode(tabl)
end

PSO.lib.deserializePlayer = function(p, data)
	
	if not data or not data.type return end	-- this is fine too, happens with empty files
	
	if not p or not p.valid or not p.pso return end
	
	local t = p.pso
	t.type_initialized = data.type_initialized
	
	t.type = data.type
	t.hp = data.hp
	t.maxhp = data.maxhp
	t.pp = data.pp
	t.maxpp = data.maxpp
	t.exp = data.exp
	t.level = data.level
	t.satk = data.satk
	t.ratk = data.ratk
	t.tatk = data.tatk
	t.def = data.def
	t.ppregen = data.ppregen
	
	t.meseta = (data.meseta or 5000)
	
	t.typepoints = data.typepoints
	t.weaponproficiencies = data.weaponproficiencies
	for i = 1, #t.weaponproficiencies	-- make it sparse again
		t.weaponproficiencies[i] = PSO.lib.sparsePow2Table(t.weaponproficiencies[i])
	end
	
	t.photonarts = PSO.lib.sparsePow2Table(data.photonarts)
	t.photonarts["techniques"] = data.techniques
	
	t.typepoints = data.typepoints
	t.maxskillpoints = data.maxskillpoints
	t.typeskills = data.typeskills
	
	t.storage = data.storage	-- this is fine as is, armor have units unequipped when they're sent to storage
	
	-- now the annoying parts:
	t.inventory = data.inventory
	-- fix inventory refs: units & weapon actions
	local tabls = {"inventory", "storage"}
	
	for tn = 1, #tabls
		
		local t = tabls[tn]
		
		for i = 1, #p.pso[t]
			local it = p.pso[t][i][1]
			
			if type(it) == "table"
			
				if it.type == IT_ARMOR
					local info = PSO.lib.itemInfo(it)
					if info.slots
						for j = 1, info.slots
							if it.slots[j]
								it.slots[j] = p.pso[t][$][1]
								-- the item is already pre-marked as equipped
							else
								it.slots[j] = nil
							end
						end
					end
				
				elseif it.type == IT_WEAPON
					local num_actions = (PSO.info.weaponHand[it.weapontype] == HAND_BOTH) and 4 or 2
					for j = 1, num_actions
						if not it.actions[j]
							it.actions[j] = nil
						end
					end
				end
			end
		end
	end
	
	t.palette.layout = data.palette.layout
	-- same deal here
	for i = 2, 3
		local pal = p.pso.palette.layout[i]
		for j = 1, PSO.info.MAXPALSIZE
			
			if pal[j]
				if i == 2	-- weapons
					
					for k = 1, 2
						if pal[j][k]
							p.pso.palette.layout[i][j][k] = p.pso.inventory[$][1]
						end
					end
					
				else
					p.pso.palette.layout[i][j] = p.pso.inventory[$][1]
				end
			else
				p.pso.palette.layout[i][j] = nil
			end
		end
	end
	
	PSO.lib.playerSortInventory(p)
	PSO.lib.playerFixUnits(p)
	PSO.lib.playerFixWeaponPalette(p)
	PSO.lib.playerSortStorage(p)
	PSO.lib.rollPlayerShop(p)
	
	-- this recalculates stats
	-- useful if we change stats during a patch etc
	PSO.lib.playerChangeType(p, p.pso.type, true)
	
	p.pso.forceupdateweapon = true
end

COM_AddCommand("psod_testnotif", function(p, msg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	PSO.lib.playerNotification(p, msg)
end)

COM_AddCommand("psod_clearquest", function()

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	PSO.lib.clearQuest()
end)