-- basic HUD

local display_dist = 3184

local function HU_login(v, p, c)
	if not p.psologin
		
		local w = 160
		local h = 32
		local x = 160 - (w/2)
		local y = 100 - (h/2)
		
		v.drawFill(x, y, w, h, 253)
		v.drawString(160, 100, netgame and "Logging in..." or "Loading...", V_YELLOWMAP, "center")
		
		return
	end
end

--------------------------	
	-- low hp strawberry jam
--------------------------
local function HU_lowHPOverlay(v, p)
	
	if p.pso.lowhp
		
		local base_intensity = 3
		local max_screen = 1
		if p.pso.hp < p.pso.maxhp*15/100
			base_intensity = 4
			max_screen = 2
		end
		
		v.fadeScreen(35, min(max_screen, p.pso.lowhp/8))
		
		local fadein = p.pso.lowhp/5
		local tflag = min(fadein, (base_intensity + (3*cos(leveltime*ANG1*4))/(FRACUNIT+2)))<<V_ALPHASHIFT
		
		v.drawScaled(0, 0, FRACUNIT/2, v.cachePatch("H_LOWHP2"), V_SNAPTOTOP|V_SNAPTOLEFT|V_ADD|tflag)
		v.drawScaled(320<<FRACBITS, 0, FRACUNIT/2, v.cachePatch("H_LOWHP2"), V_SNAPTORIGHT|V_SNAPTOTOP|V_ADD|V_FLIP|tflag)	
		v.drawScaled(0, 200<<FRACBITS, FRACUNIT/2, v.cachePatch("H_LOWHP"), V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_ADD|tflag)
		v.drawScaled(320<<FRACBITS, 200<<FRACBITS, FRACUNIT/2, v.cachePatch("H_LOWHP"), V_SNAPTOBOTTOM|V_SNAPTORIGHT|V_ADD|V_FLIP|tflag)
	end
end


--------------------------
	-- open menu helper
--------------------------
local function HU_menuReminder(v, p)
	if gamemap == PSO.info.LOBBY_MAP
	and not (p.pso.menu and p.pso.menu.ref)
		local pp = PSO.lib.playerButtonIcon(p, IPT_MENU)
		v.draw(2, 2, v.cachePatch(pp), V_SNAPTOTOP|V_SNAPTOLEFT)
		v.drawString(20, 3, "OPEN MENU", V_YELLOWMAP|V_SNAPTOTOP|V_SNAPTOLEFT)
	end
end

--------------------------	
	-- low pp indicator
-------------------------
local function HU_lowPP(v, p)
	if p.pso.nopp
		local pnum = (leveltime%10 / 5) +1
		
		if p.pso.nopp > TICRATE/3
		or leveltime & 1
			v.draw(160, 80, v.cachePatch("H_NOPP"..pnum))
		end
	end
end

--------------------------	
	-- target reticle:
-------------------------	
local function HU_targetReticle(v, p, c)
	if p.pso.target and p.pso.target.valid
		
		local ang = leveltime*4*ANG1
		local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, p.pso.target.x, p.pso.target.y, p.pso.target.z + p.pso.target.height/2)
		local sprite, flip, offc = v.getSpritePatch(SPR_AIMR, p.pso.lockon and B or A, 0, ang)
		
		local offs = 0
		if offc
			offs = 4*(flip and -1 or 1)
		end
		
		v.drawScaled(cx, cy+(offs*FRACUNIT), FRACUNIT, sprite)
	end
end

local function HU_damageNumbers(v, p, c)
	
	if server.dmgnums
		for i = 1, #server.dmgnums
			
			local m = server.dmgnums[i]
			
			if R_PointToDist(m.x, m.y) < FRACUNIT*display_dist
				
				local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, m.x, m.y, m.z)
				
				if cx > 0 and cy > 0
					if m.tics > TICRATE/2 or m.tics & 1
						
						if m.extravalue1
							if not m.extravalue2
								if p.pso.level >= PSO.info.LEVELCAP
									v.drawString(cx/FRACUNIT, cy/FRACUNIT - 20, "\x82".."EXP FULL!", 0, "center")
								else
									v.drawString(cx/FRACUNIT, cy/FRACUNIT - 20, "\x82"..m.angle.." EXP", 0, "center")
								end
							end
						else
							local halfscale = true
							if m.source == p.mo
							or m.target == p.mo
								halfscale = nil
							end
							
							local col = SKINCOLOR_WHITE
							
							if m.extravalue2
								col = SKINCOLOR_EMERALD
							elseif m.target and m.target.player
								col = SKINCOLOR_RED
							end
							
							PSO.lib.HUD_drawDamageNumber(v, cx/FRACUNIT -5, cy/FRACUNIT, m.angle, 0, col, m.extravalue2, halfscale)
							--v.drawNum(cx/FRACUNIT + (#tostring(m.angle) -1)*6, cy/FRACUNIT, m.angle)
						end
					end
				end
			end
		end
	end
end


--------------------------	
	-- chain numbers:
-------------------------
local function HU_chainCombo(v, p, c)
	if server.chaintrackers
		for i = 1, #server.chaintrackers
			
			local m = server.chaintrackers[i]
				
			if R_PointToDist(m.x, m.y) < FRACUNIT*display_dist and m.pso.chain and m.pso.chain > 1
				
				local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, m.x, m.y, m.z)

				local timer = m.pso.chainbreak and m.pso.chainbreak or m.pso.chaintime
				local maxtimer = m.pso.chainbreak and m.pso.maxchainbreak or m.pso.maxchaintime
				
				local col = m.pso.chainbreak and "\x82" or "\x88"
				local ccol = abs(timer-maxtimer) < 5 and "\x80" or col
				local col2 = m.pso.chainbreak and 73 or 130
				
				local str = ccol..string.format("%02d", m.pso.chain-1)..col.." CHAIN"
				local w = v.stringWidth(str)
				
				if cx > 0 and cy > 0
					if timer > TICRATE or timer & 1
						v.drawString(cx/FRACUNIT, cy/FRACUNIT +4, str, 0, "center")
					end
					
					v.drawFill(cx/FRACUNIT - w/2, cy/FRACUNIT + 13, w, 2, 26)
					v.drawFill(cx/FRACUNIT - w/2, cy/FRACUNIT + 13, timer*(w-1)/maxtimer, 1, col2)
				end
			end
		end
	end
end

	
--------------------------	
	-- status bar:
-------------------------
local function HU_playerStatus(v, p)
	local x = 4 - p.pso.hidehudtimer*16
	local y = 165
	
	for i = 1, 10
		v.drawFill(x-2, y-2 + (i-1), 48 + (i-1), 1, 253|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	end
	
	v.drawFill(x-2, y+8, 132, 25, 253|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	
	local lvnum = string.format("%03d", p.pso.level)
	
	v.drawString(x, y-1, "\x82".."Lv", V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	v.drawString(x + 16, y-1, lvnum, V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	
	v.drawFill(x, y+8, 39, 2, 26|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	
	if p.pso.level >= PSO.info.LEVELCAP
		v.drawFill(x, y+8, 38, 1, 73|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	else
		v.drawFill(x, y+8, p.pso.exp*38/PSO.info.exptable[p.pso.level], 1, 73|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	end
	
	v.drawString(x + 60, y-1, p.name, V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	
	local hpnum = string.format("%03d", p.pso.hp)
	local hpmnum = string.format("%03d", p.pso.maxhp)
	local ppnum = string.format("%03d", p.pso.pp)
	local ppmnum = string.format("%03d", p.pso.maxpp)
	
	v.drawString(x, y+10, "\x82".."HP", V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	v.drawString(x + 40, y+10, hpnum.."/"..hpmnum, V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT|(p.pso.hp <= p.pso.maxhp/3 and V_REDMAP or 0))
	
	local hp_col = {35, 34, 54, 73, 189, 98, 98, 98, 98, 98}
	--print(p.pso.hp*10/p.pso.maxhp)
	local usecol = hp_col[(p.pso.hp*10/p.pso.maxhp) or 1] or 98
	
	v.drawFill(x, y+19, 129, 2, 26|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	v.drawFill(x, y+19, p.pso.hp*128/p.pso.maxhp, 1, usecol|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	
	v.drawString(x, y+21, "\x82".."PP", V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	v.drawString(x + 40, y+21, ppnum.."/"..ppmnum, V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT)

	local pp_col = {168, 166, 164, 162, 148, 132, 132, 132, 132, 132}
	--print(p.pso.hp*10/p.pso.maxhp)
	local ppusecol = pp_col[(p.pso.pp*10/p.pso.maxpp) or 1] or 132
	
	v.drawFill(x, y+30, 129, 2, 26|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	v.drawFill(x, y+30, p.pso.pp*128/p.pso.maxpp, 1, ppusecol|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	
	-- status conditions:
	local statusx = x + 134
	if p.pso.condition
	and (p.pso.conditiontimer > TICRATE*3 or p.pso.conditiontimer & 1)
		v.draw(statusx, y+8, v.cachePatch("H_STATUS"..p.pso.condition), V_SNAPTOBOTTOM|V_SNAPTOLEFT)
	end
	
	-- s/d j/z
	local st = {"shifta", "deband", "jellen", "zallure"}
	local pp = {"H_STATUS8", "H_STATUS16", "H_STATUS32", "H_STATUS64"}
	local sx = statusx
	
	for i = 1, #st
		if p.pso[st[i]]
			if (p.pso[st[i]] > TICRATE*10 or p.pso[st[i]] & 1)
				v.draw(sx, y+8+12, v.cachePatch(pp[i]), V_SNAPTOBOTTOM|V_SNAPTOLEFT)
			end
			v.drawString(sx+8, y+8+12+6, p.pso[st[i]]/TICRATE, V_SNAPTOBOTTOM|V_SNAPTOLEFT, "small-thin-right")
			sx = $+ 12
		end
	end
end


	
--------------------------	
	-- item palette:
-------------------------
local function HU_playerItemPalette(v, p)
	local px = 268 + p.pso.hidehudtimer*16 + p.pso.palette.open*32
	local py = 182
	
	-- indication for switching:
	local bt = v.cachePatch(PSO.lib.playerButtonIcon(p, IPT_PALETTE, flash))
	v.draw(px - 16, py+2, bt, V_SNAPTOBOTTOM|V_SNAPTORIGHT)
	
	-- weapon:
	local wdta = p.pso.palette.layout[2][p.pso.palette.weaponnum]
	local onehanded = PSO.lib.slotIsOneHanded(wdta)
	
	v.drawFill(px, py, 32, 16, 253|V_SNAPTOBOTTOM|V_SNAPTORIGHT)
	
	if wdta
		for i = 1, 2
			local wp = wdta[i]
			if wp
				local col = PSO.info.elementColours[wp.element]
				local ico = PSO.info.weaponToIcon[wp.weapontype]
				local tflag = 0
				
				if onehanded
					
					if i == 1 and not p.pso.backpalette and wdta[2]
					or i == 2 and p.pso.backpalette and wdta[1]
						tflag = V_50TRANS
					end
				end
				
				v.draw(px, py, v.cachePatch(ico), V_SNAPTOBOTTOM|V_SNAPTORIGHT|tflag, v.getColormap(TC_DEFAULT, col))
			end
		end
	end
	
	-- armour:
	local arm = p.pso.palette.layout[3][p.pso.palette.armornum]
	v.drawFill(px + 33, py, 16, 16, 253|V_SNAPTOBOTTOM|V_SNAPTORIGHT)
	
	if arm
		local col = PSO.info.elementColours[arm.element]
		v.draw(px + 33, py, v.cachePatch("H_ARMR"), V_SNAPTOBOTTOM|V_SNAPTORIGHT, v.getColormap(TC_DEFAULT, col))
	end
end



--------------------------	
	-- action palette:
-------------------------
local function HU_playerActionPalette(v, p)
	local wp = PSO.lib.getPlayerHUDWeapon(p)
	local hand = HAND_BOTH
	local offset = 0

	local px = 268 + p.pso.hidehudtimer*16 + p.pso.palette.open*32
	local py = 182
	
	if wp
		hand = PSO.info.weaponHand[wp.weapontype]
	end

	
	-- the thing for the back palette...
	local bpalb = PSO.lib.playerButtonIcon(p, IPT_SECONDPALETTE, p.pso.backpalette and leveltime%4 < 2)
	v.drawFill(px + 21, py - 15 - 26, 22, 12, 253|V_SNAPTORIGHT|V_SNAPTOBOTTOM)
	v.draw(px + 22, py - 15 - 25, v.cachePatch("H_BPAL"), V_SNAPTORIGHT|V_SNAPTOBOTTOM)
	v.draw(px + 32, py - 15 - 26, v.cachePatch(bpalb), V_SNAPTORIGHT|V_SNAPTOBOTTOM)

	
	if wp and hand == HAND_BOTH
		offset = p.pso.backpalette and 2 or 0	-- if we hold swap palette, offset by 2
	end
	
	local butts = {IPT_ATTACK, IPT_ATTACK2}		
	local cy = py - 15
		
	for i = 1 + offset, 2 + offset
			
		local flash = false
			
		if p.pso.actionnum == i
		and p.pso.actiontimer < TICRATE/2
		and p.pso.actiontimer%4 < 2
			flash = true
		end
			
		local bt =  PSO.lib.playerButtonIcon(p, butts[i - offset], flash)
			
		v.drawFill(px - 48, cy, 68, 12, 253|V_SNAPTORIGHT|V_SNAPTOBOTTOM)
		v.drawFill(px + 21, cy, 22, 12, 253|V_SNAPTORIGHT|V_SNAPTOBOTTOM)
			
		-- action name / ico:
		local dta
		
		if wp and wp.actions[i]
			dta = PSO.info.photonArts[wp.actions[i]]
		end
				
		if dta
			local ico = v.cachePatch(dta.icon)
			v.drawString(px - 46, cy+2, dta.name, V_ALLOWLOWERCASE|V_SNAPTORIGHT|V_SNAPTOBOTTOM, "thin")
			v.draw(px + 21, cy, ico, V_SNAPTORIGHT|V_SNAPTOBOTTOM)
		else
			v.drawString(px - 46, cy, "---", V_ALLOWLOWERCASE|V_SNAPTORIGHT|V_SNAPTOBOTTOM)
		end
			
		v.draw(px + 32, cy, v.cachePatch(bt), V_SNAPTORIGHT|V_SNAPTOBOTTOM)
			
		cy = $ - 13
	end
end

--------------------------	
	-- keycards:
-------------------------
local function HU_displayKeycards(v, p)
	if server.keycards
	and not (p.pso.menu and p.pso.menu.ref)
		local minimap_on = cv_useminimap.value and (gamemap ~= PSO.info.LOBBY_MAP or p.mo.y > PSO.info.LOBBY_YPOS*FRACUNIT)
		local tx = 320 - 64 - 2
		local ty = minimap_on and 68 or 2
		
		v.drawFill(tx, ty, 64, 21, 253|V_SNAPTOTOP|V_SNAPTORIGHT)
		v.drawString(tx+2, ty+2, "\x82".."KEYS", V_SNAPTOTOP|V_SNAPTORIGHT)
		
		for i = 1, 4
			if server.keycards & (2^(i-1))
				v.draw(tx+2, ty+11, v.cachePatch("H_KCARD"..i), V_SNAPTOTOP|V_SNAPTORIGHT)
				tx = $+9
			end
		end
	end	
end


--------------------------	
	-- target info:
-------------------------
local function HU_targetInfo(v, p)
	if p.pso.target and p.pso.target.valid and not p.pso.menu.ref
		local t = p.pso.target
		local tx = 2
		local ty = 2
				
		if t.type == MT_PSO_DROP
					
			local it = t.psoitem
			if t.psoitems
				it = t.psoitems[#p+1]
			end
			
			-- meseta
			if it and type(it) == "number"
				v.drawFill(tx, ty, 140, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
				v.drawString(tx+2+17, ty+2, PSO.lib.numCommas(it).." MESETA", V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)
				
				local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
				v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
				v.drawString(tx+2+20, ty+15, "PICK UP", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
				
			elseif it	-- normal item
				
				v.drawFill(tx, ty, 140, 33, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
				
				local qty = t.psoqty
				if not qty
					qty = 1
				end
				
				local info
				if type(it) == "string"
					info = PSO.info.items[it]
				else
					info = PSO.info.items[it.ref]
				end
				
				local name = PSO.lib.getItemName(it)
				local rarity = info.rarity
				local ico = PSO.lib.getItemIcon(it)
				
				v.draw(tx+2+7, ty+2, v.cachePatch(ico), V_SNAPTOTOP|V_SNAPTOLEFT)
				v.drawString(tx+2+17, ty+2, name..(qty > 1 and " ("..qty..")" or ""), V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)
				PSO.lib.HUD_drawItemRarity(v, tx+2, ty+12, V_SNAPTOLEFT|V_SNAPTOTOP, rarity or 0)
			
				-- check if it can be picked up:
				local nopickup
				
				if #p.pso.inventory >= PSO.info.INVENTORYSIZE
					nopickup = "INVENTORY IS FULL"
				
				elseif info.carrylimit 
					local slot = PSO.lib.findPlayerItem(p, it)
					if slot
						local dta = p.pso.inventory[slot]
						if dta and dta[2] and dta[2] >= info.carrylimit
							nopickup = info.name.." STOCK FULL"
						end
					end
				end
			
				if nopickup
					v.drawString(tx+2, ty+22, nopickup, V_REDMAP|V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
				else
					local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					
					v.draw(tx+2, ty+19, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
					v.drawString(tx+2+18, ty+24, "PICK UP", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
				end
			end
		
		elseif t.type == MT_PSO_UNLOCKCARD

			v.drawFill(tx, ty, 128, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			local name = "KEYCARD (ID: "..(t.id or "INVALID")..")"
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)

			local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					
			v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
			v.drawString(tx+2+20, ty+15, "PICK UP", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")			
		
		elseif t.type == MT_PSO_CARDREADER

			v.drawFill(tx, ty, 128, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			local name = "NEED ID: "..(t.id or "INVALID")
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)
			
			local keycards = server.keycards or 0
			
			if keycards & (t.id or 0)
				local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
						
				v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
				v.drawString(tx+2+20, ty+15, "USE KEYCARD", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
			else
				v.drawString(tx+2, ty+15, "KEYCARD NOT FOUND", V_REDMAP|V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
			end

		elseif t.type == MT_PSO_TELEPORTER
		or t.type == MT_PSO_BOSSTELEPORTER

			v.drawFill(tx, ty, 128, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			local name = "TELEPORTER"
			if t.spawnpoint and t.spawnpoint.angle == 1
				name = "TO LOW ORBIT STATION"
			end
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)

			local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					
			v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
			v.drawString(tx+2+20, ty+15, "TELEPORT", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")	

		elseif t.type == MT_PSO_JUMPPAD

			v.drawFill(tx, ty, 128, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			local name = "JUMP PAD"
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)

			local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					
			v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
			v.drawString(tx+2+20, ty+15, "USE", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")	
		
		elseif t.flags & MF_ENEMY	-- enemy
			
			local boxw = 128
			v.drawFill(tx, ty, boxw, 24, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			
			if t.pso and t.pso.part
				if t.tracer
					t = t.tracer
				end
			end
			
			local name = tostring(t)
			if PSO.info.enemies[t.type]
				name = "LV"..(t.pso.level or "?").." "..PSO.info.enemies[t.type].name or $
			end
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)
			
			-- weakness?
			if t.pso.weakness
				local str = "WEAK:"
				v.drawString(tx+2, ty+10, "\x82"..str, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)
				local pp = v.cachePatch("H_ELEM"..t.pso.weakness)
				local sw = v.stringWidth(str)
				
				v.draw(tx+4+sw, ty+10, pp, V_SNAPTOTOP|V_SNAPTOLEFT)
			end
			
			-- targetted?
			if t.target == p.mo
				v.draw(tx + boxw - 16, ty+10, v.cachePatch("H_HATE"), V_SNAPTOTOP|V_SNAPTOLEFT)
			end
			
			local info = t.pso
			if info
				v.drawFill(tx+2, ty+20, boxw-4, 2, 26|V_SNAPTOTOP|V_SNAPTOLEFT)
				v.drawFill(tx+2, ty+20, info.hp*(boxw-5)/info.maxhp, 1, 98|V_SNAPTOTOP|V_SNAPTOLEFT)
			end
		else	-- other shootable objects
			v.drawFill(tx, ty, 128, 24, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
		
			local name = tostring(t)
			if t.pso and t.pso.name
				name = t.pso.name
			end
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)		
		end
	end
end

local function HU_showIncapacitations(v, p, c)
	for p2 in players.iterate do
			
		if p2 and p2.mo and p2.mo.valid
		and p2.pso and p2.pso.hp <= 0	-- blud is dead	
		and R_PointToDist(p2.mo.x, p2.mo.y) < FRACUNIT*display_dist
				
			local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, p2.mo.x, p2.mo.y, p2.mo.z+p2.mo.height/3)
			cy = $ + 2*cos(leveltime*ANG1*4)
			
			v.drawScaled(cx, cy, FRACUNIT, v.cachePatch("H_INCAP"))
		end
	end
end

local function playerHUD_base(v, p, c)
	
	
	HU_login(v, p, c)
	if not p.pso return end
	
	if not p.pso.type_initialized return end	-- only display the type selection menu
	

	HU_lowHPOverlay(v, p)
	HU_menuReminder(v, p)
	HU_lowPP(v, p)
	HU_showIncapacitations(v, p, c)
	HU_targetReticle(v, p, c)
	HU_damageNumbers(v, p, c)
	HU_chainCombo(v, p, c)
	HU_playerStatus(v, p)
	HU_playerItemPalette(v, p)
	HU_playerActionPalette(v, p)
	HU_displayKeycards(v, p)
	HU_targetInfo(v, p)
end

PSO.lib.registerHudFunc(playerHUD_base)