freeslot("SPR_MMPL")

local MINIMAP_RANGE = FRACUNIT*1536

local minimap_lines = {}
local minimap_objects = {}

addHook("MapChange", function()
	if server and server.valid
		server.sectorsvisited = {}
	end
end)

local function playerPopulateSectorsList(p)
	if p.mo and p.mo.valid
		local sec = p.mo.subsector.sector
		server.sectorsvisited[#sec] = sec
	end
end

PSO.lib.registerPlayerFunc(playerPopulateSectorsList)


local function blockmap_linesearch(mo, line)
	
	if line.frontsector
	and type(line.frontsector) == "userdata"	-- wtf
	and server.sectorsvisited[#line.frontsector]
	or line.backsector
	and type(line.backsector) == "userdata"	-- wtf
	and server.sectorsvisited[#line.backsector]
	
		minimap_lines[#minimap_lines+1] = line
	end
end

local function blockmap_objsearch(mo, mo2)
	if mo2 and mo2.valid
		
		if mo2 == mo
			return
		end
		
		if mo2.flags & MF_ENEMY
		or mo2.type == MT_PLAYER
		
		or mo2.type == MT_PSO_LASERTOWER
		and not mo2.extravalue2	-- activated
		and mo2.hnext	-- make sure it has a new target
		
		or mo2.type == MT_PSO_UNLOCKCARD
		and mo2.extravalue1
			minimap_objects[#minimap_objects+1] = mo2
		end
	end
end

local function line_slopespecial(l)
	return (l.special >= 700 and l.special < 800)
end

local function playerMinimapHUD(v, p, c)
	
	if not p.mo return end
	local mo = p.mo
	
	if not cv_useminimap.value return end
	if p.pso.menu and p.pso.menu.ref return end				-- in menus, hide the minimap
	if gamemap == PSO.info.LOBBY_MAP or (mo.y < PSO.info.LOBBY_YPOS*FRACUNIT) return end	-- in lobby
	
	local rot = 0
	if cv_maprotation.value
		rot = p.mo.angle + ANG1*90
	end
	
	
	local w = 64
	local h = 64
	
	local x = 320 - 2 - w
	local y = 2
	
	local bounds = {x, y, x+w, y+h}
	
	-- easier
	local function calcXY(cxpos, cypos)
	
		if cv_maprotation.value
			
			local dist = R_PointToDist2(mo.x, mo.y, cxpos, cypos)
			local angle = R_PointToAngle2(mo.x, mo.y, cxpos, cypos)
			
			angle = $ - rot
			cxpos = mo.x - FixedMul(dist, cos(angle))
			cypos = mo.y - FixedMul(dist, sin(angle))
		end
		
		local rx = x +w/2 +((cxpos - mo.x)/FRACUNIT * (w)/2) / (MINIMAP_RANGE/FRACUNIT)
		local ry = y +h/2 -((cypos - mo.y)/FRACUNIT * (h)/2) / (MINIMAP_RANGE/FRACUNIT) 

		return rx, ry
	end
	
	v.drawFill(x, y, w, h, 253|V_SNAPTOTOP|V_SNAPTORIGHT)
	
	minimap_lines = {}
	minimap_objects = {}
	
	searchBlockmap("lines", blockmap_linesearch, p.mo, p.mo.x - MINIMAP_RANGE*2, p.mo.x + MINIMAP_RANGE*2, p.mo.y - MINIMAP_RANGE*2, p.mo.y + MINIMAP_RANGE*2)
	searchBlockmap("objects", blockmap_objsearch, p.mo, p.mo.x - MINIMAP_RANGE*2, p.mo.x + MINIMAP_RANGE*2, p.mo.y - MINIMAP_RANGE*2, p.mo.y + MINIMAP_RANGE*2)
	
	-- draw the lines
	for i = 1, #minimap_lines
		local l = minimap_lines[i]
		
		local dx1, dy1 = calcXY(l.v1.x, l.v1.y)
		local dx2, dy2 = calcXY(l.v2.x, l.v2.y)
		
		-- what colour should we draw the line?
		local frontsec = l.frontsector
		local backsec = l.backsector
		local linecol = 0
		
		
		/*if l.flags & ML_IMPASSIBLE
		or frontsec
		and (frontsec.floorheight >= frontsec.ceilingheight or frontsec.floorheight >= backsec.ceilingheight)
		and not line_slopespecial(l)
		or backsec
		and (backsec.floorheight >= backsec.ceilingheight or backsec.floorheight >= frontsec.ceilingheight)
		and not line_slopespecial(l)
			linecol = 0*/
		
		if frontsec and backsec and frontsec.floorheight == backsec.floorheight
		or line_slopespecial(l)
			linecol = 156
			
		else
			linecol = 0
		end
		
		PSO.lib.drawLine(v, dx1, dy1, dx2, dy2, linecol|V_SNAPTOTOP|V_SNAPTORIGHT, bounds)
	end
	
	-- draw the objects
	for i = 1, #minimap_objects
		local o = minimap_objects[i]

		local dx, dy = calcXY(o.x, o.y)
		
		local pp
		
		if o.flags & MF_ENEMY
			
			if o.pso and o.pso.boss
				pp = (o.target == p.mo) and "H_MMEN2_2" or "H_MMEN2_1"
			else
				pp = (o.target == p.mo) and "H_MMEN1_2" or "H_MMEN1_1"
			end
		
		elseif o.type == MT_PLAYER
			pp = "H_MMP2"
		
		elseif o.type == MT_PSO_UNLOCKCARD
		and leveltime%10 < 5
			v.drawFill(dx-2, dy-3, 3, 5, 35|V_SNAPTOTOP|V_SNAPTORIGHT)
			continue
		
		elseif o.type == MT_PSO_LASERTOWER
			
			-- get hnext and draw a red line between the two
			local o2 = o.hnext

			local dx2, dy2 = calcXY(o2.x, o2.y)
			
			PSO.lib.drawLine(v, dx, dy, dx2, dy2, 35|V_SNAPTOTOP|V_SNAPTORIGHT, bounds)
			continue
		end
		
		if pp
		and dx > x and dx < x+w
		and dy > y and dy < y+h
			v.drawScaled(dx<<FRACBITS, dy<<FRACBITS, FRACUNIT/2, v.cachePatch(pp), V_SNAPTOTOP|V_SNAPTORIGHT)
		end
	end
	
	-- finally, draw the player at the center of the map
	local a
	
	if cv_maprotation.value
		a = PSO.lib.angToInt((p.drawangle + ANG1*90) - rot) + 90
	else
		a = PSO.lib.angToInt(p.drawangle) --+ 180
	end
	
	local plpatch, flip, offc = v.getSpritePatch(SPR_MMPL, A, 0, a*ANG1)
	local offs = 0
		
	if offc
		offs = 2*(flip and -1 or 1)
	end
		
	--print(offc)
		
	v.drawScaled((x+w/2)*FRACUNIT, (y+h/2 +offs)*FRACUNIT, FRACUNIT/2, plpatch, V_SNAPTOTOP|V_SNAPTORIGHT)

	
	/*local p1 = {x+w/2, y+h/2}
	
	local p2 = {
		
		p1[1] + (24*cos(leveltime*ANG1*4)) /FRACUNIT,
		p1[2] + (24*sin(leveltime*ANG1*4)) /FRACUNIT,
	}
	
	PSO.lib.drawLine(v, p1[1], p1[2], p2[1], p2[2], V_SNAPTOTOP|V_SNAPTORIGHT)*/
end

PSO.lib.registerHudFunc(playerMinimapHUD)