
-- for late joiners, teleport em to the checkpoint or the leader, whichever is available
local function playerFirstTeleport(p)
	if server.questdata
	and gamemap ~= PSO.info.LOBBY_MAP
		
		local ref = server.keycheckpoint
		local ref2 = PSO.lib.getLeader()
		
		if ref and ref.valid
							
			local tempx = ref.x + 128*cos(ref.angle)
			local tempy = ref.y + 128*sin(ref.angle)
			local temp = P_SpawnMobj(tempx, tempy, ref.z, MT_THOK)
			temp.state = S_INVISIBLE
			temp.tics = TICRATE
			P_MoveOrigin(p.mo, temp.x, temp.y, temp.z)
			p.mo.angle = R_PointToAngle2(p.mo.x, p.mo.y, ref.x, ref.y)
			
		elseif ref2 and ref2.mo and ref2.mo.valid
			P_MoveOrigin(p.mo, ref2.mo.x, ref2.mo.y, ref2.mo.z)
		elseif server.lastspawnpoint	
			local pp = server.lastspawnpoint
			
			P_MoveOrigin(p.mo, pp[1], pp[2], pp[3])
			p.mo.angle = pp[4]
			
			if pp[5] & 2
				p.mo.flags2 = $|MF2_OBJECTFLIP
			end
		end	-- if we're out of options, well fuck you.
	end
end 

-- Initialize all PSO-related variables within the player...
local function playerInit(p)
	if not p.pso
		
		--COM_BufInsertText(p, "sessionanalog 0")	-- experimental
		
		local typ = TYPE_HUNTER
		local info = PSO.info.typestats[typ]
		
		local inputs = {}
		for k,v in pairs(PSO.info.inputDefaults)
			inputs[k] = v
		end
		
		
		p.pso = {
			
			type = typ,
			
			-- monies
			meseta = 5000,
			
			-- level stats
			level = 1,
			exp = 0,
			
			-- stats
			maxhp = info.hp,
			hp = info.hp,
			
			maxpp = info.pp,
			pp = info.pp,
			
			satk = info.satk,
			ratk = info.ratk,
			tatk = info.tatk,
			def = info.def,
			
			pb = 0,
			
			-- pp regen
			ppregen = info.ppregen,
			ppregentimer = 0,	-- recover 1 pp when regentimer > regen
			
			-- action variable
			backpalette = nil,	-- use actions 3, 4 instead of 1, 2
			action = nil,		-- action ref being used
			taction = nil,		-- tech action being used
			actionnum = 0,		-- action num being used (this is mostly for the HUD)
			actionbutton = nil,	-- which button did we use for that? keep track of that for holding.
			actionweapon = nil,	-- which weapon is performing the action?
			actiontimer = 0,	-- how long has it been performed for?
			actionheld = false,	-- stays ture until the button is released, cannot be set to true again 
			actionhold = 0,		-- are we still holding the action button? (doesn't matter for every anim)
			actionspeed = nil,	-- multiplier for movement speed: forces cmd forward/sidemove to lower values (0-100)
			actioncombo = 0,	-- for actions performed back to back
			
			-- targetting:
			target = nil,		-- targetted enemy / item
			
			-- inventory / inputs / palette
			inventory = {},
			storage = {},	-- player storage
			shop = {},		-- shop (generates per player)
			
			-- inputs & preferences
			inputs = inputs,	-- used to remap controls on a player to player baisis
			buttons = {},		-- buttons being pressed by the player, in correlation to inputs
			toggle_lockon = false,		-- toggle lockon mode
			toggle_apalette = false,	-- toggle action palette mode
			toggle_bpalette = false,	-- toggle back palette mode
			
			-- weapon stuff:
			weaponpos = {
				{nil, 0, 0, 0, 0, 0, 0},	-- ref, x, y, z, angle, rollangle, slopeangle
				{nil, 0, 0, 0, 0, 0, 0},	-- ref, x, y, z, angle, rollangle, slopeangle (secondary weapon)
				-- secondary weapon is for dual wielding, but is also used for stuff like twin sabers and TMGs
			},
			
			-- palette stuff:
			palette = {
				
				layout = {{}, {}, {}},
				itemnum = 1,
				weaponnum = 1,
				aweaponnum = 1,	-- active weapon num
				armornum = 1,
				open = 0,	-- animation for opening the palette. goes up to TICRATE/4
				column = 0,	-- current column
				scroll = 0,
			},
			
			-- menu info:
			menu = {
				
				-- current opened menu index
				ref = nil,
				choice = 1,
				
				timer = 0,
				
				-- 0: up/down, 1: left/right
				inputtype = 0,			
			},
			
			-- used when opening menus.
			-- hidehudtimer will go up to TICRATE/2 when hidehud is true, and back down to 0 when it's false.
			hidehud = false,
			hidehudtimer = 0,
			
			-- saves cursor position for menus
			-- eg: menumemory["main"] = 2
			menumemory = {},
		}
		
		-- ready weapon proficiencies:
		p.pso.weaponproficiencies = {}
		local types = {TYPE_HUNTER, TYPE_RANGER, TYPE_FORCE}
		for i = 1, #types
			p.pso.weaponproficiencies[types[i]] = {}
			
			for k, v in pairs(PSO.info.typestats[types[i]].weapons)
				p.pso.weaponproficiencies[types[i]][k] = v
			end
		end
		
		-- ready class skills (well, the fact that you don't have any)
		p.pso.typeskills = {}
		p.pso.typepoints = {}
		p.pso.maxskillpoints = 4
		for i = 1, #types
			p.pso.typeskills[types[i]] = {}
			p.pso.typepoints[i] = 0
		end	
		
		-- ready base weapon photon arts etc
		-- we will always default to having 2 available action per weapons
		-- for most weapons this is normal + 1 PA
		-- for techs this is barta and foie
		p.pso.photonarts = {}
		for k,v in pairs(PSO.info.weaponDefaultActions)
			p.pso.photonarts[k] = 1|2	-- unlock the first two actions for every weapon/tech.
		end
		
		local foie = 2^0
		local barta = 2^5
		local zonde = 2^10
		local diga = 2^13
		
		p.pso.photonarts.techniques = foie|barta|zonde|diga
		
		PSO.lib.initPlayerType(p, TYPE_HUNTER)	-- by default
		
		-- generate sections for player inventory & storage
		PSO.lib.playerSortInventory(p)
		PSO.lib.playerSortStorage(p)
		PSO.lib.rollPlayerShop(p)
		
		-- preferences:
		PSO.lib.playerLoadPrefs(p)
		
		-- TP into the quest
		playerFirstTeleport(p)
		
		--print(PSO.lib.serializePlayer(p))
	end
	
	-- fuck OFF with your stupid controls
	p.pflags = $ & ~PF_ANALOGMODE
	--local lmaogalog = CV_FindVar("sessionanalog")
	--CV_StealthSet(lmaogalog, 0)
end

PSO.lib.registerPlayerFunc(playerInit)

rawset(_G, "render_hack", 0)
/*hud.add(function(v)
	render_hack = 0	--v.renderer()
end)*/

/*local cv_fuck = CV_RegisterVar({
	name = "ang",
	defaultvalue = 0,
	PossibleValue = CV_Unsigned,
})*/


-- used specifically for lockon
addHook("PlayerCmd", function(p, cmd)
	if p.pso and p.pso.lockon and p.pso.lockon.valid and p.mo and p.mo.valid
		cmd.angleturn =  R_PointToAngle2(p.mo.x, p.mo.y, p.pso.lockon.x, p.pso.lockon.y)/FRACUNIT
	end
end)

-- Make the player slow as dicks and put them in a wheelchair:
local function playerMovement(p)
	if p.mo and p.mo.valid
		local mo = p.mo
		
		if p.pso.dodgeroll
			p.normalspeed = 0
			p.thrustfactor = 0
		elseif not p.pso.menu.ref
			p.normalspeed = FRACUNIT*14
			p.thrustfactor = skins[p.mo.skin].thrustfactor
		end
		p.runspeed = FRACUNIT*24
		p.jumpfactor = 0
		p.charability = 0
		p.charability2 = 0
		
		
		if p.pso.target
		and p.pso.target.valid
		and not p.pso.lockon
		and p.pso.target.flags & MF_ENEMY	-- don't lock on to other shit lol
		and PSO.lib.playerInputDown(p, IPT_LOCKON) == 1
			p.pso.lockon = p.pso.target
			-- force the button to be released if we use toggle lockon
			if p.pso.toggle_lockon
				PSO.lib.playerForceInputUp(p, IPT_LOCKON)
			end
		end
		
		if p.pso.lockon
		and not p.pso.menu.ref
		and not P_PlayerInPain(p)
			
			if not p.pso.lockon.valid
			or (not p.pso.toggle_lockon and not PSO.lib.playerInputDown(p, IPT_LOCKON))
			or (p.pso.toggle_lockon and PSO.lib.playerInputDown(p, IPT_LOCKON) == 1)
			or p.pso.lockon ~= p.pso.target	-- if that somehow happens
				p.pflags = $ & ~PF_FORCESTRAFE
				p.pso.lockon = nil
			else
				p.mo.angle = R_PointToAngle2(mo.x, mo.y, p.pso.lockon.x, p.pso.lockon.y)
				p.drawangle = mo.angle
				p.pflags = $|PF_FORCESTRAFE
			end
		end
		
		-- Update weapon position
		-- handles sheathe position outside of actions too.
		-- note: weapon position is relative from the player's
		
		if p.pso.palette.layout[2][p.pso.palette.aweaponnum]
			
			if not PSO.lib.slotHasWeapon(p.pso.palette.layout[2][p.pso.palette.aweaponnum])
			and p.pso.palette.aweaponnum ~= p.pso.palette.weaponnum
				p.pso.palette.aweaponnum = p.pso.palette.weaponnum
			end
			
			for i = 1, 2	--#p.pso.palette.layout[2][p.pso.palette.aweaponnum]
				
				-- if the player switched weapon and is not doing an action, switch the weapon!
				if (p.pso.forceupdateweapon or p.pso.palette.weaponnum ~= p.pso.palette.aweaponnum)
				and not p.pso.action
					for j = 1, 2
						if p.pso.weaponpos[j][1] and  p.pso.weaponpos[j][1].valid
							P_RemoveMobj( p.pso.weaponpos[j][1])
							p.pso.weaponpos[j][1] = nil
						end
					end
					p.pso.palette.aweaponnum = p.pso.palette.weaponnum
					p.pso.forceupdateweapon = nil
				end
				
				local weapon = p.pso.palette.layout[2][p.pso.palette.aweaponnum][i]
				
				if weapon
					local info = PSO.info.items[weapon.ref]
					
					if not p.pso.weaponpos[i][1] or not p.pso.weaponpos[i][1].valid	-- how.
						
						local wp = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
						wp.tics = -1
						wp.fuse = -1
						
						wp.sprite = info.sprite
						wp.frame = info.frame | FF_PAPERSPRITE
						wp.color = info.color or SKINCOLOR_NONE
												
						p.pso.weaponpos[i][1] = wp
					end
					
					if not (p.pso.action and p.pso.actionslot == i)
						
						local sheatheinfo = PSO.info.defaultSheathePosition[weapon.weapontype]
						
						-- update positions to sheathe the weapon:
						local x = sheatheinfo[1]*cos(p.drawangle + ANG1*180)
						local y = sheatheinfo[2]*sin(p.drawangle + ANG1*180)
						local z = FixedDiv(mo.height*sheatheinfo[3]/100, mo.scale)
						
						p.pso.weaponpos[i][2] = x
						p.pso.weaponpos[i][3] = y
						p.pso.weaponpos[i][4] = z
						p.pso.weaponpos[i][5] = sheatheinfo[4]
						p.pso.weaponpos[i][6] = sheatheinfo[5]
						
						-- pslope_t data
						p.pso.weaponpos[i][7] = {
							use = false,
							o = {},
							xy = 0,
							za = 0,
						}
					end
					
					-- move the weapon to its destination position
					local tx = mo.x + FixedMul(p.pso.weaponpos[i][2], mo.scale)
					local ty = mo.y + FixedMul(p.pso.weaponpos[i][3], mo.scale)
					local tz = mo.z + FixedMul(p.pso.weaponpos[i][4], mo.scale)
					
					P_MoveOrigin(p.pso.weaponpos[i][1], tx, ty, tz)
					p.pso.weaponpos[i][1].angle = p.drawangle + p.pso.weaponpos[i][5]
					p.pso.weaponpos[i][1].tics = 2
					
					-- update splats:
					if p.pso.weaponpos[i][7]
						local dta = p.pso.weaponpos[i][7]
						local wep = p.pso.weaponpos[i][1]
						if dta.use
							
							if not wep.floorspriteslope
								-- ready pslope_t.
								-- this is useful for most striking weapons.
								
								if info.splatframe ~= nil
									wep.frame = $ & ~FF_FRAMEMASK
									wep.frame = $ | info.splatframe
								end
								
								P_CreateFloorSpriteSlope(wep)
							end
							
							wep.frame = $ & ~FF_PAPERSPRITE
							wep.flags2 = $ | MF2_SPLAT
							wep.renderflags = $|RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD
							wep.floorspriteslope.o = dta.o
							wep.floorspriteslope.xydirection = dta.xy
							wep.floorspriteslope.zangle = dta.za
							wep.angle = 0
						else
							if wep.floorspriteslope
								-- unset data
								P_RemoveFloorSpriteSlope(wep)
								wep.flags2 = $ & ~MF2_SPLAT
								wep.renderflags = $ & ~(RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD|RF_NOSPLATROLLANGLE)

								if info.frame ~= nil
									wep.frame = $ & ~FF_FRAMEMASK
									wep.frame = $ | info.frame
								end

								wep.frame = $ | FF_PAPERSPRITE
							end
						end
					end
					
					-- lol renderer is broken
					if render_hack == "software"
					and R_PointToAngle(tx, ty, tz) - p.pso.weaponpos[i][1].angle > 0
					and p.pso.weaponpos[i][1].frame & FF_PAPERSPRITE
						p.pso.weaponpos[i][1].rollangle = -p.pso.weaponpos[i][6]	-- this is retarded
					else
						p.pso.weaponpos[i][1].rollangle = p.pso.weaponpos[i][6]
					end
				end
			end	
		else	-- aweaponnum invalid
			
			if p.pso.palette.aweaponnum ~= p.pso.palette.weaponnum
				p.pso.palette.aweaponnum = p.pso.palette.weaponnum
			end
		end
	end
end	

local function playerActionFreeze(p)
	return 
		p.powers[pw_nocontrol] 
		or p.pso.menu.ref 
		or p.pso.palette.open 
		or P_PlayerInPain(p) 
		or p.pso.teleport 
		or p.pso.jumppad 
		or p.pso.dodgeroll
		or p.pso.stagger
		or p.pso.blowdown
		or p.pso.knockback
		or p.pso.condition == COND_FREEZE
end

local function playerWeaponActions(p)
	local weapon = PSO.lib.getPlayerActiveWeapon(p)
	local loc_weapon = PSO.lib.getPlayerHUDWeapon(p)
	
	if p.mo and p.mo.valid
		local mo = p.mo
		if weapon
			if not p.pso.action and not playerActionFreeze(p)
			or p.pso.action and PSO.info.photonArts[p.pso.action]
			and PSO.info.photonArts[p.pso.action].combo 
			and p.pso.actiontimer >= PSO.info.photonArts[p.pso.action].combo	-- in that case, we can perform an action.
			and weapon == loc_weapon
			and not playerActionFreeze(p)
				
				local offset = 0
				local hand = PSO.info.weaponHand[weapon.weapontype]
				
				if hand == HAND_BOTH	-- 2-handed weapons can have 4 actions
					offset = p.pso.backpalette and 2 or 0	-- set to 2 if we're holding back palette
				end	
					
				local input = {IPT_ATTACK, IPT_ATTACK2}
				
				for i = 1 + offset, 2 + offset
					if weapon.actions[i]
					
						local butt = PSO.lib.playerInputDown(p, input[i - offset])
						
						if butt == 1
							
							local action = weapon.actions[i]
							local actiondta = PSO.info.photonArts[action]
							
							if not actiondta return end	-- shouldn't happen but in case it does
							
							if p.pso.action 
							and not (actiondta.combo and p.pso.actiontimer >= actiondta.combo and p.pso.actionnum == i)
								--print("Cannot perform again")
								continue
							elseif p.pso.action
								p.pso.actioncombo = $+1
								--print("Combo!")
							end
							
							local ppcost = actiondta.cost or 0
							if weapon.pa_cost
								ppcost = $*weapon.pa_cost/100
							end
							
							if ppcost
							and p.pso.pp < ppcost
								--print("Insufficient PP")
								PSO.lib.noPPNotif(p)
								continue
							else
								p.pso.pp = $ - ppcost
							end
							
							--print("Using action "..i)
							
							if actiondta.technique
								--print("Technique used, using weapon cast action...")
								p.pso.action = PSO.info.castArt[weapon.weapontype]
								p.pso.taction = action
								actiondta = PSO.info.photonArts[action]	-- make sure to re-read that data
							else
								p.pso.action = action
								p.pso.taction = nil
							end
							
							p.pso.actionnum = i
							p.pso.actionbutton = input[i - offset]
							p.pso.actionweapon = weapon
							p.pso.actionslot = (hand == HAND_RIGHT and 2 or 1)	-- know which weapon slot we're using.
							p.pso.actiontimer = 1
							p.pso.actionheld = true
							p.pso.actionhold = 1
							
							if actiondta.speed ~= nil
								p.pso.actionspeed = actiondta.speed
								-- remove any speed we have above that value...
								
								-- let's figure out what our new max speed should be...
								local ang = R_PointToAngle2(0, 0, p.mo.momx, p.mo.momy)
								local maxmomx = FixedMul(p.mo.scale, FixedMul(p.normalspeed, cos(ang)))
								local maxmomy = FixedMul(p.mo.scale, FixedMul(p.normalspeed, sin(ang)))
								
								if maxmomx > 0 and mo.momx > maxmomx
								or maxmomx < 0 and mo.momx < maxmomx
									mo.momx = maxmomx
								end
								
								if maxmomy > 0 and mo.momy > maxmomy
								or maxmomy < 0 and mo.momy < maxmomy
									mo.momy = maxmomy
								end
								
							else
								p.pso.actionspeed = 100
							end
							
							break
						end
					end
				end
			end
			
			if p.pso.action	-- menu OK: we need to let the player finish their action
				-- we're performing an action!
				local action = PSO.info.photonArts[p.pso.action]
				
				if action
					
					if action.use
						if action.use(p, p.pso.actiontimer, p.pso.actionheld, p.pso.actionhold, p.pso.actioncombo)
							PSO.lib.resetPlayerAction(p)
							return
						end	
					end
					
					p.pso.actiontimer = $ + 1
					p.pso.actionhold = PSO.lib.playerInputDown(p, p.pso.actionbutton)
					if p.pso.actionheld and not p.pso.actionhold
						p.pso.actionheld = false
					end
				end
			end
		end
	end
end

local function playerTargetting(p)
	
	local range = 512
	local ang = 60
	
	local weapon = PSO.lib.getPlayerActiveWeapon(p)
	local mo = p.mo
	
	if not mo or not mo.valid return end
	if p.pso.menu.ref return end	-- useless
	
	p.pso.target = nil	-- remove the target each frame
	-- some weapons have higher/lower range
	
	local info = PSO.lib.itemInfo(weapon)
	if weapon
		if weapon.range
			range = weapon.range
		elseif weapon.angle
			ang = weapon.angle
		end
	end
	
	-- range x2.5 in lockon
	if p.pso.lockon
		range = $*5/2
	end
	
	-- use searchblockmap to find drops or cards in range.
	-- priority is always:
	-- enemy > card > others > drops
	
	local dropfound
	local cardfound
	local tpfound
	local npcfound
	
	-- yeah that should be far enough maybe probably perhaps
	local closest = FRACUNIT*10000
	local objectfound
	
	if not p.pso.action	-- if attacking, attacks take priority on the current target
	and not p.pso.lockon
		searchBlockmap("objects", function(ref, found)
			
			if found and found.valid and abs(ref.z - found.z) < ref.scale*128
				
				local dist = R_PointToDist2(ref.x, ref.y, found.x, found.y)
				if dist < closest
					
					if (found.type == MT_PSO_UNLOCKCARD and found.extravalue1)		-- make sure it's visible.
					or (found.type == MT_PSO_CARDREADER and not found.extravalue2)	-- make sure it's not already on
						--cardfound = found
						objectfound = found
						closest = dist
					
					elseif found.type == MT_PSO_DROP
					and (not found.collected or not (found.collected & (2^(#p))))
					and (not found.psoitems or found.psoitems[#p+1])
					and not dropfound
					--and dist <= found.radius
						--dropfound = found
						objectfound = found
						closest = dist
					
					elseif found.type == MT_PSO_JUMPPAD
						--tpfound = found	-- hack
						objectfound = found
						closest = dist
					
					elseif (found.type == MT_PSO_TELEPORTER or found.type == MT_PSO_BOSSTELEPORTER)
					and not found.hold
						--tpfound = found
						objectfound = found
						closest = dist
					
					elseif found.type == MT_PSO_NPC_INTERACT
						--npcfound = found
						objectfound = found.target
						closest = dist
					
					end
				elseif dist == closest	-- somehow the objects are at the exact same coordinates...
					
					-- prioritise cards and then drops
					if (found.type == MT_PSO_UNLOCKCARD and found.extravalue1)		-- make sure it's visible.
					or (found.type == MT_PSO_CARDREADER and not found.extravalue2)	-- make sure it's not already on					
						cardfound = found
					
					elseif found.type == MT_PSO_DROP
					and (not found.collected or not (found.collected & (2^(#p))))
					and (not found.psoitems or found.psoitems[#p+1])
					and not cardfound
					--and dist <= found.radius
						dropfound = found				
					end
				end
			end
		end, mo)
		
		-- priority 1 is card
		if cardfound
			p.pso.target = cardfound
			return
		end
		
		-- then drops
		if dropfound
			p.pso.target = dropfound
			return
		end
		
		/*-- then teleporters
		if tpfound
			p.pso.target = tpfound
			return
		end
		
		-- and then npcs.
		if npcfound
			p.pso.target = npcfound.target
		end*/
		
		-- if we make it here, get the closest object to us!
		if objectfound
			p.pso.target = objectfound
			return
		end
	end
	
	local enemies = {}
	-- use searchblockmap to find enemies within our range...
	
	local cx = mo.x + range/2 * FixedMul(mo.scale, cos(mo.angle))
	local cy = mo.y + range/2 * FixedMul(mo.scale, sin(mo.angle))
	
	local x1 = cx - (range/2 *mo.scale)
	local x2 = cx + (range/2 *mo.scale)
	local y1 = cy - (range/2 *mo.scale)
	local y2 = cy + (range/2 *mo.scale)
	
	local function search(ref, found)
		
		if found and found.valid and (found.flags & (MF_ENEMY|MF_SHOOTABLE)) and found.type ~= MT_PLAYER
		
			if not PSO.info.enemies[found.type] or not PSO.info.enemies[found.type].notarget
			
				-- check angle
				local angle_diff = (p.cmd.angleturn*FRACUNIT) - R_PointToAngle2(mo.x, mo.y, found.x, found.y)
				
				if (PSO.lib.angToInt(angle_diff) > 180)
					angle_diff = InvAngle(angle_diff)
				end
							
				if (PSO.lib.angToInt(angle_diff) < ang)
					enemies[#enemies+1] = found
				end
			end
		end
	end
	
	searchBlockmap("objects", search, mo, x1, x2, y1, y2)
	
	-- sort enemies by distance / angle... give them some sort of score.
	local scores = {}
	for i = 1, #enemies
		
		scores[i] = {i, 0}
		scores[i][2] = $ + FixedDiv(R_PointToDist2(mo.x, mo.y, enemies[i].x, enemies[i].y), mo.scale) / FRACUNIT / 4	-- angle matters more
		scores[i][2] = $ + abs(mo.z - enemies[i].z) / FRACUNIT	-- z diff matters too
		scores[i][2] = $ + abs((mo.angle - R_PointToAngle2(mo.x, mo.y, enemies[i].x, enemies[i].y)) / ANG1)
		
		-- cheat and make it the best enemy if we're locked on to it.
		if enemies[i] == p.pso.lockon
			scores[i][2] = -FRACUNIT
		end
		
		if not (enemies[i].flags & MF_ENEMY)	-- another damageable thing? crate? It's very low priority.
			scores[i][2] = $*4
		end
		--print("enemy "..i.." score: "..scores[i])
	end
	
	table.sort(scores, function(a, b) return a[2] < b[2] end)
	
	if scores and scores[1]
		local final_target = enemies[scores[1][1]]
		if final_target and final_target.valid
		
			p.pso.target = final_target
		end
	end
end

local function playerPPRegen(p)
	
	local mo = p.mo
	if not mo or not mo.valid return end
	
	local ppwellup
	
	if p.pso.action and p.pso.taction
	and PSO.lib.hasSkill(p, "fo_ppwellup")
		ppwellup = true
	end
	
	if p.pso.action
	and not ppwellup
	or p.pso.photondash
		p.pso.ppregentimer = 0
	else
		p.pso.ppregentimer = $+1
		
		if p.pso.ppregentimer >= p.pso.ppregen*(ppwellup and 2 or 1)
			p.pso.ppregentimer = 0
			p.pso.pp = min(p.pso.maxpp, $+1)
		end
	end
	
	if not server.questdata
	or gamemap == PSO.info.LOBBY_MAP
	or p.pflags & PF_GODMODE
		p.pso.pp = p.pso.maxpp
		p.pso.hp = p.pso.maxhp
	end
	
end

local function playerHideHUD(p)
	if p.pso.hidehud
		p.pso.hidehudtimer = min(TICRATE/2, $+1)
	else
		p.pso.hidehudtimer = max(0, $-1)
	end
end

-- builds the quick palette
local function playerBuildQuickPalette(p)
	local vispal = {}
	local pal = p.pso.palette.layout
	-- build what's on the palette...
	local start = {p.pso.palette.itemnum, p.pso.palette.weaponnum, p.pso.palette.armornum}
	-- we need to go back by 2 indices (loop back to PALSIZE if we go up)
		
	for col = 1, #pal
			
		local cur = start[col]
		vispal[col] = {}
			
		vispal[col][3] = {pal[col][cur], cur}
		-- make it go back...
		local filled = 0
		local jackshit = 0
		
		while filled < 2 and jackshit < PSO.info.MAXPALSIZE
			cur = $-1
				
			if cur < 1
				cur = PSO.info.MAXPALSIZE
			end
			
			if pal[col][cur]
				filled = $+1
				jackshit = 0
				vispal[col][3-filled] = {pal[col][cur], cur}
			else
				jackshit = $+1
			end
		end
		
		-- and forward
		cur = start[col]
		filled = 0
		jackshit = 0
		
		while filled < 2 and jackshit < PSO.info.MAXPALSIZE
			cur = $+1
				
			if cur > PSO.info.MAXPALSIZE
				cur = 1
			end
			
			if pal[col][cur]
				filled = $+1
				jackshit = 0
				vispal[col][3+filled] = {pal[col][cur], cur}
			else
				jackshit = $+1
			end
		end
	end
	
	p.pso.palette.qlayout = vispal
end

local function playerHandlePaletteSwitch(p)
	
	--print(leveltime..": "..tostring(PSO.lib.playerInputDown(p, IPT_PALETTE)))
	
	if p.pso.toggle_apalette
		if PSO.lib.playerInputDown(p, IPT_PALETTE) == 1
			p.pso.openpal = not p.pso.openpal and true or false
		end
	else
		p.pso.openpal = PSO.lib.playerInputDown(p, IPT_PALETTE)
	end
	
	local openpal = p.pso.openpal
	--print(openpal)
	if p.pso.menu.ref
		openpal = 0
	end
	
	p.pso.palette.scroll = $*3/4
	
	if openpal
		p.pso.palette.open = min($+1, TICRATE/4)
		
		if p.pso.palette.open == 1
			-- sets qlayout (not saved)
			p.pso.palette.column = 2	-- default to weapons
			playerBuildQuickPalette(p)
		end
		
		local inputs = {IPT_ATTACK2, IPT_ATTACK, IPT_SECONDPALETTE}
		local vals = {"itemnum", "weaponnum", "armornum"}
		
		for i = 1, #inputs
			if PSO.lib.playerInputDown(p, inputs[i]) == 1
			and not p.pso.palette.scroll
				
				if p.pso.palette.column ~= i
					p.pso.palette.column = i
				else
					--print(p.pso.palette.itemnum)
					if p.pso.palette.qlayout[i][4] ~= nil
						p.pso.palette[vals[i]] = p.pso.palette.qlayout[i][4][2]	-- 4 is always the next one with how things are laid out.
					end
					--print(p.pso.palette.itemnum)
					--print("---")
					p.pso.palette.scroll = 16
					
					-- rebuild the palette layout
					playerBuildQuickPalette(p)
				end
			end
		end
		
		--print(p.pso.palette.open)
	else
		
		if p.pso.palette.open == (TICRATE/4 -1)
			if p.pso.palette.column == 1	-- consumables
				-- use em!
				local it = p.pso.palette.layout[1][p.pso.palette.itemnum]
				
				if it
					local itinfo = PSO.lib.itemInfo(it)
					
					-- either there's no check or it returns true
					if not itinfo.canuse or itinfo.canuse(p)
						local invslot = PSO.lib.findPlayerItem(p, it)
						-- if invslot is 0 then we don't have any of these.
						if not invslot
							PSO.lib.dprint("No "..it.." in inventory!")
						else
							-- if we made it here then we can use it!
							PSO.lib.playerLoseItem(p, invslot, 1)	-- use the item up
							
							if itinfo.use
								itinfo.use(p)
							end
							PSO.lib.dprint("Used "..it)
						end
					else
						PSO.lib.dprint("Cannot use item "..it)
					end
				end
			
			elseif p.pso.palette.column == 3	-- armours
				-- recalculate stats for units!
				PSO.lib.playerCalcStats(p, p.pso.level)
			end
		end
		
		p.pso.palette.open = max(0, $-1)
	end
end

-- pickup drops, keycards and interact w/ other items...
local function playerPickupDrops(p)
	
	-- note: the target can only be set if the drop can be SEEN by the player anyway
	-- there are sanity checks preventing getting it again regardless.
	if p.pso.target and p.pso.target.valid
	and PSO.lib.playerInputDown(p, IPT_INTERACT) == 1
	and not p.pso.action
	and not p.pso.menu.ref	
	and not p.pso.stagger
	and not p.pso.knockback
	and not p.pso.blowdown	-- this fixes a bunch of bugs, really.
		if p.pso.target.type == MT_PSO_DROP
			if PSO.lib.playerPickUpDrop(p, p.pso.target)
				p.pso.target = nil	-- remove that
			end		
			return true
		elseif p.pso.target.type == MT_PSO_UNLOCKCARD
			PSO.lib.playerGetKeyCard(p, p.pso.target)
			p.pso.target = nil	-- remove that
			return true
			
		elseif p.pso.target.type == MT_PSO_CARDREADER
			if PSO.lib.activateCardReader(p, p.pso.target)
				p.pso.target = nil
			end
			return true
		
		elseif p.pso.target.type == MT_PSO_JUMPPAD
		and not playerActionFreeze(p)
			PSO.lib.playerUseJumpPad(p, p.pso.target)
			p.pso.target = nil
			return true
		
		elseif p.pso.target.type == MT_PSO_BOSSTELEPORTER
			PSO.lib.openMenu(p, "bosstp")
			p.pso.teletarget = p.pso.target
			p.pso.target = nil
			return true
		
		elseif p.pso.target.type == MT_PSO_TELEPORTER
		and not playerActionFreeze(p)
			if p.pso.target.target
				PSO.lib.softResetPlayer(p)
				p.pso.teleport = 1
				p.pso.teleporttarget = p.pso.target.target
				
				if p.pso.target.lobbytele
					p.pso.teleporttype = 1
				elseif p.pso.target.fromlobbytele
					
					if p.pso.usecheckpoint
						-- hack when respawning after dying to make walking back less of a slaugh
						p.pso.usecheckpoint = nil
						
						local ref = server.keycheckpoint
						if ref and ref.valid
							
							local tempx = ref.x + 128*cos(ref.angle)
							local tempy = ref.y + 128*sin(ref.angle)
							local temp = P_SpawnMobj(tempx, tempy, ref.z, MT_THOK)
							temp.state = S_INVISIBLE
							temp.tics = TICRATE
							
							p.pso.teleporttarget = temp
						end
					end
					
					p.pso.teleporttype = 2
				else
					p.pso.teleporttype = 0
				end
			end
			return true
		
		elseif p.pso.target.type == MT_PSO_NPC
			PSO.lib.NPCInteract(p, p.pso.target)
			return true
		end
	end
end

local function playerHandleDamage(p)
	
	local mo = p.mo
	if p.pso.stagger
		mo.rollangle = 0

		
		if p.pso.stagger == 1
			local saveang = p.drawangle
			P_DoPlayerPain(p)
			p.drawangle = saveang
			P_InstaThrust(p.mo, 0, 0)
			p.mo.momz = p.mo.scale*3
			p.powers[pw_flashing] = TICRATE*5/2
			p.pso.stagger = nil
			return
		end
		
		p.pso.stagger = $+1
	end
	
	if p.pso.knockback
		mo.rollangle = 0
		
		p.powers[pw_flashing] = TICRATE/2
		p.normalspeed = 0
		
		if not P_IsObjectOnGround(mo)
			mo.momx = FixedMul(mo.momx, mo.friction)
			mo.momy = FixedMul(mo.momy, mo.friction)
			mo.state = S_PLAY_PAIN
		else
			local rad = mo.radius/FRACUNIT
			local d = P_SpawnMobj(mo.x + P_RandomRange(-rad, rad)*FRACUNIT, mo.y + P_RandomRange(-rad, rad)*FRACUNIT, mo.z, MT_SPINDUST)
			d.momz = P_RandomRange(1, 4)*mo.scale
			
			mo.state = S_PLAY_WALK
			mo.frame = A			
		end
		
		mo.momx = FixedMul(mo.momx, mo.friction)
		mo.momy = FixedMul(mo.momy, mo.friction)
		
		if FixedHypot(mo.momx, mo.momy) < mo.scale
			p.pso.knockback = 0
		end
	end
	
	if p.pso.blowdown
	
		p.pso.blowdown = $+1
		mo.state = S_PLAY_PAIN
		p.powers[pw_flashing] = TICRATE
	
		if not P_IsObjectOnGround(mo)
			mo.blowdownmomz = mo.momz
			mo.rollangle = $ + ANG1*32
			mo.z = $ + mo.momz
		elseif p.pso.blowdown > 4
			
			if FixedHypot(mo.momx, mo.momy) < mo.scale*2
				p.pso.blowdown = nil
				mo.rollangle = 0
			else
			
				mo.momz = -mo.blowdownmomz/3
				mo.momx = $/2
				mo.momy = $/2
				
				for i = 1, 8
					local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
					d.angle = (360/8)*i*ANG1
					P_InstaThrust(d, d.angle, mo.blowdownmomz*3)
					d.momz = -mo.blowdownmomz/4
				end	
			end
		end
	end
end		

local sizes = {
	{FRACUNIT, FRACUNIT},
	{FRACUNIT/2, FRACUNIT*2},
	{FRACUNIT/4, FRACUNIT*4},
	{FRACUNIT/8, FRACUNIT*8},
}

local function playerHandleTeleport(p)
	if p.pso.teleport
		p.normalspeed = 0
		
		p.pso.teleport = $ + 1
		
		local index = p.pso.teleport
		if p.pso.teleport > 5
			index = 5-(p.pso.teleport-5)
		end
		
		if sizes[index]
			p.mo.spritexscale = sizes[index][1]
			p.mo.spriteyscale = sizes[index][2]
		end
		
		if p.pso.teleport == 5
			P_SetOrigin(p.mo, p.pso.teleporttarget.x, p.pso.teleporttarget.y, p.pso.teleporttarget.z)
			S_StartSound(nil, sfx_mixup, p)
			
			-- change sky/weather/music:
			if p.pso.teleporttype == 1			-- level -> lobby
				P_SetupLevelSky(22, p)
				P_SwitchWeather(0, p)
				S_ChangeMusic("LOBBY", true, p)
				
			elseif p.pso.teleporttype == 2		-- lobby -> level
				local header = mapheaderinfo[gamemap]
				
				P_SetupLevelSky(header.skynum, p)
				P_SwitchWeather(header.weather, p)
				S_ChangeMusic(header.musname, true, p)
				
				if PSO.lib.questDifficulty() >= DIFF_HARD
					if header.hardsky
						P_SetupLevelSky(header.hardsky, p)
					end
					
					if header.hardweather
						P_SwitchWeather(_G[header.hardweather] or 0, p)
					end	
					
					if header.hardmusic
						S_ChangeMusic(header.hardmusic, true, p)
					end
				end
			end
		end
		
		if p.pso.teleport == 10
			p.pso.teleport = nil
		end
	end
end

local function playerBackPalette(p)
	
	if p.pso.toggle_bpalette
		
		if PSO.lib.playerInputDown(p, IPT_SECONDPALETTE) == 1
			p.pso.backpalette = not p.pso.backpalette and true or false
		end
	else
		p.pso.backpalette = PSO.lib.playerInputDown(p, IPT_SECONDPALETTE)
	end
end

local function playerHandleDeath(p)
	local mo = p.mo
	
	mo.state = S_PLAY_PAIN
	p.normalspeed = 0
	p.jumpfactor = 0
	
	p.pso.deadtimer = $+1
	
	if p.pso.deadtimer == TICRATE
		PSO.lib.openMenu(p, "skillissue")
	end
	
end

-- mostly counts up the time if we're in the actual quest map lol
local function playerHandleQuestTrackers(p)
	
	if server.questdata
	and gamemap == server.questdata.map
	and p.pso.quest
		p.pso.quest.time = $+1
		
		if p.pso.quest.completetimer
			
			p.normalspeed = 0
			
			p.pso.quest.completetimer = $+1
			
			if PSO.lib.playerInputDown(p, IPT_CONFIRM) == 1
				
				if p.pso.quest.completetimer < TICRATE*3
					p.pso.quest.completetimer = TICRATE*3
				else
					p.pso.quest.completetimer = nil
					p.pso.quest.turnin = 1	-- can be turned in
				end
			end
			
			if p.pso.quest.completetimer == TICRATE*3/2
				S_StartSound(nil, sfx_rnkup2, p)
			end
			
			if p.pso.quest.completetimer == TICRATE*3
				S_StartSound(nil, sfx_rank1 + (p.pso.quest.ranking -1), p)
			end
		end
	end
end

local function playerJumpPad(p)
	if p.pso.jumppad
		
		local mo = p.mo
		
		--print("flags: "..(mo.flags & (MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING)))
					
		p.normalspeed = 0
		p.thrustfactor = 0
		mo.flags = $|MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT
		p.pflags = $|PF_NOCLIP
		
		p.pso.jumppadtimer = $+1
		if p.pso.jumppadtimer > TICRATE*10	-- this isn't normal

			P_MoveOrigin(p.mo, p.pso.jumppad.x, p.pso.jumppad.y, p.pso.jumppad.z)
			p.pso.jumppad = nil
			p.pso.jumppadtimer = nil
			p.thrustfactor = skins[p.mo.skin].thrustfactor
			mo.momx = 0
			mo.momy = 0
			mo.jumppad_next = nil

			mo.flags = $ & ~(MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT)
			p.pflags = $ & ~(PF_NOCLIP)
			
			return
		end
		
		-- emergency cancel
		if (mo.momz < 0 and mo.z <= p.pso.jumppad.z)
			-- !?
			mo.z = p.pso.jumppad.z
			p.pso.jumppad = nil

			mo.flags = $ & ~(MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT)
			p.pflags = $ & ~(PF_NOCLIP)
			
			mo.momx = 0
			mo.momy = 0
			
			mo.jumppad_next = nil
			
			p.thrustfactor = skins[p.mo.skin].thrustfactor
			p.pso.jumppadtimer = nil
			return
		end

		if p.pso.jumppad.valid
			local dist = R_PointToDist2(mo.x, mo.y, p.pso.jumppad.x, p.pso.jumppad.y)
			local spd = FixedHypot(mo.momx, mo.momy)
			
			if dist < spd
			or p.pso.jumppadstop
				mo.momx = $*90/100
				mo.momy = $*90/100
				
				mo.flags = $ & ~(MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT)
				p.pflags = $ & ~(PF_NOCLIP)
				
				p.pso.jumppadstop = true
				
				if FixedHypot(mo.momx, mo.momy) < mo.scale*2
				or P_IsObjectOnGround(mo)
				
					p.pso.jumppad = nil
					p.thrustfactor = skins[p.mo.skin].thrustfactor
					mo.momx = 0
					mo.momy = 0
					mo.jumppad_next = nil
					p.pso.jumppadtimer = nil
				end
			else
				-- noclip for players fucking sucks so we're gonna have to hack things.
				if mo.jumppad_next
					P_MoveOrigin(mo, mo.jumppad_next[1], mo.jumppad_next[2], mo.jumppad_next[3])
				end
				
				if not P_IsObjectOnGround(mo)
					mo.jumppad_next = {
						mo.x + mo.momx,
						mo.y + mo.momy,
						mo.z + mo.momz,
					}
				else
					mo.jumppad_next = nil
				end

				if leveltime & 1
					local g = P_SpawnGhostMobj(mo)
					g.flags = $|MF_NOCLIPHEIGHT|MF_NOCLIP
					g.colorized = true
						
					P_MoveOrigin(g, mo.x, mo.y, mo.z)
				end
					
				if mo.momz > -mo.scale*4
				and mo.momz < mo.scale*4
					mo.state = S_PLAY_ROLL
				elseif mo.momz < -mo.scale*4
					mo.state = S_PLAY_FALL
				end
				
			end
		end
	end
end

local function playerDodgeRoll(p, nododge)
	
	if not p.pso.dodgeroll
	and PSO.lib.playerInputDown(p, IPT_INTERACT) == 1
	and P_IsObjectOnGround(p.mo)
	and not nododge
	and not p.pso.action
	and p.normalspeed	-- menus etc
	--and not p.pso.jumppad
	--and not p.powers[pw_nocontrol]
	and not playerActionFreeze(p)
		
		if p.pso.pp < PSO.info.dodgeroll_cost
			PSO.lib.noPPNotif(p)
			return
		end
		
		p.pso.pp = $ - PSO.info.dodgeroll_cost
		
		p.pso.dodgeroll = TICRATE/3
		
		if PSO.lib.hasSkill(p, "all_dodgeextension")
			p.pso.dodgeroll = $*3/2
		end
		
		S_StartSound(p.mo, sfx_jump)
		
		local an = p.mo.angle
		-- angle depends on player input
		if FixedHypot(p.cmd.forwardmove*FRACUNIT, p.cmd.sidemove*FRACUNIT)
			an = $+R_PointToAngle2(0, 0, p.cmd.forwardmove*FRACUNIT, -p.cmd.sidemove*FRACUNIT)
		else
			an = p.drawangle
		end
			
		
		/*if not FixedHypot(p.mo.momx, p.mo.momy)
			an = p.mo.angle
		else
			an = R_PointToAngle2(0, 0, p.mo.momx, p.mo.momy)
		end*/
		
		P_InstaThrust(p.mo, an, p.mo.scale*20)
		p.mo.momz = p.mo.scale*7
		
		p.mo.state = S_PLAY_ROLL
	end
	
	if p.pso.dodgeroll
		
		p.pso.dodgeroll = $-1
		if not p.pso.dodgeroll
			
			if PSO.lib.playerInputDown(p, IPT_INTERACT) > 1
				p.pso.photondashbuffer = true
			end
			
			if not P_IsObjectOnGround(p.mo)
				p.mo.state = S_PLAY_FALL
			end
		else
			if leveltime & 1
				local g = P_SpawnGhostMobj(p.mo)
				g.colorized = true
			end	
		end
	elseif p.pso.photondashbuffer
		
		if not PSO.lib.playerInputDown(p, IPT_INTERACT)
		or p.pso.action
		or p.pso.jumppad
		or not (p.cmd.forwardmove or p.cmd.sidemove)
		or not p.pso.pp
			p.pso.photondash = nil
			p.pso.photondashbuffer = nil
		else
			if P_IsObjectOnGround(p.mo)
			and p.pso.photondashbuffer
				p.pso.photondash = true
				--p.pso.photondashbuffer = nil
			end
			
			if p.pso.photondash
				p.normalspeed = FRACUNIT*24
				p.runspeed = FRACUNIT*15
				
				if leveltime%3 == 0
					p.pso.pp = max(0, $-1)
				end
				
				if leveltime & 1
					local g = P_SpawnGhostMobj(p.mo)
					g.colorized = true
					g.tics = 2
				end
			end
		end	
	end
end

local function noPPNotif(p)
	if p.pso.nopp
		p.pso.nopp = $-1
	end
end

local function playerTerrainDamage(p)
	local mo = p.mo
	
	-- sector damage
	if P_PlayerTouchingSectorSpecial(p, 1, 2)
	and P_IsObjectOnGround(mo)
		if leveltime%(TICRATE/2) == 0
		and not p.powers[pw_flashing]
			PSO.lib.damageObject(mo, nil, p.pso.maxhp/20)
		end
	end	
end

local function playerLowHP(p)
	if p.pso.hp <= p.pso.maxhp*30/100
		p.pso.lowhp = $ and $+1 or 1
	else
		p.pso.lowhp = 0
	end
end

local function playerHandleNotification(p)
	if p.pso.notification
		p.pso.notification.timer = $-1
		
		if not p.pso.notification.timer
			p.pso.notification = nil
		end
	end
end

local function playerSetQuest(p)
	p.pso.quest = {
		time = 0,
		damagetaken = 0,
		kills = {},	-- tracks kill per object type
		points = 0,	-- quest points (specific to some quests)
		incapacitations = 0
	}
end

local function playerThinker(p)
	
	if not p.pso.quest
		playerSetQuest(p)
	end
	
	if not p.pso.deadtimer
		playerTargetting(p)
		local nododge = playerPickupDrops(p)
		playerMovement(p)
		playerDodgeRoll(p, nododge)
		playerHandleDamage(p)
		playerBackPalette(p)
		playerWeaponActions(p)
		playerPPRegen(p)
		playerHandlePaletteSwitch(p)
		playerTerrainDamage(p)
		PSO.lib.handleStatus(p.mo)
	else
		playerHandleDeath(p)
	end
	
	playerHandleQuestTrackers(p)
	playerHandleTeleport(p)
	playerJumpPad(p)
	playerHideHUD(p)
	noPPNotif(p)
	playerLowHP(p)
	playerHandleNotification(p)
	
	PSO.lib.handleEmergency(p)
	PSO.lib.playerHandleRareDropAnim(p)
	
	p.lives = 1		-- only for singleplayer so that the retry command doesn't work
	p.dashmode = 0	-- dashmode is purely visual but we still don't want it.
	
	-- singleplayer hack, kill tailsbot
	if not netgame
	and players[1]
	and players[1].valid
	and players[1].bot
		G_RemovePlayer(1)
	end
end

PSO.lib.registerPlayerFunc(playerThinker)

-- player: reset quest specific trackers:
addHook("MapChange", function()
	for p in players.iterate do
		
		PSO.lib.softResetPlayer(p)
		playerSetQuest(p)
	end
end)

-- if we have actionspeed
addHook("PlayerCmd", function(p, cmd)
	if p.pso and p.pso.actionspeed ~= nil and not p.pso.menu.ref	-- player controls are frozen during menu anyway
	
		local maxmove = 50 * p.pso.actionspeed / 100
	
		if cmd.sidemove < -maxmove
			cmd.sidemove = -maxmove
		elseif cmd.sidemove > maxmove
			cmd.sidemove = maxmove
		end
		
		if cmd.forwardmove < -maxmove
			cmd.forwardmove = -maxmove
		elseif cmd.forwardmove > maxmove
			cmd.forwardmove = maxmove
		end
	end
end)